/*
 * DRFPORT.H - Internal include file for the Resource File API that encapsulates all
 * source code portability concerns.
 *
 * Copyright (C) 2001 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 *
 * 29-Sept-2008 BG  Adding POSIX name macros for continued portability support - see below for more
 */

#ifndef _DRFPORT

#define _DRFPORT

#ifdef _WINDOWS

// Windows platforms, Microsoft C/C++ compiler, little endian (Intel)

/* Standard include files */
#include <ctype.h>
#include <float.h>
#include <io.h>
#include <malloc.h>
#include <share.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <windows.h>
#include <winreg.h>
#include <wtypes.h>


/* Directory separator character */

#define DIR_SEPARATOR_CHAR    '\\'
#define DIR_SEPARATOR_CHAR2   '/'
#define DIR_SEPARATOR_STRING  "\\"
#define DIR_SEPARATOR_STRING2 "/"


/* Microsoft's "_fsopen" function is an extended "fopen" */

#define FSOPEN(name, mode, flags) _fsopen(name, mode, flags)


/* Microsoft doesn't have an "ftruncate" function, but the more general
 * MS-specific "_chsize" has the same arguments, and works instead. */
#define ftruncate _chsize


/* fopen() argument */

#define F_ACCESS_APPEND_TEXT        "at"
#define F_ACCESS_READ_TEXT          "rt"
#define F_ACCESS_WRITE_TEXT         "wt"
#define F_ACCESS_READ_BINARY        "rb"
#define F_ACCESS_UPDATE_EXIS_BINARY "r+b"
#define F_ACCESS_UPDATE_NEW_BINARY  "w+b"
#define F_ACCESS_WRITE_BINARY       "wb"

/*
 * The following macros refer to POSIX names of runtime library functions. With the 
 * event of Visual Studio 2008, those became deprecated by Microsoft (eventually, in 
 * a future release, to be discontinued). Microsoft's warning C4996 says "The POSIX 
 * name for this item is deprecated. Instead, use the ISO C++ conformant name: xxx",
 * with old/new name pairs such as stricmp -> _stricmp. 
 * The following macros map the POSIX name to the ISO C++ name. Other ports may use
 * other names for these functions, or continue using the POSIX names.
 */
#define STRICMP _stricmp
#define MEMCCPY _memccpy
#define FILENO  _fileno

/* This definition enables the default implementation of the functions
 * LdrfAllocFn() and LdrfFreeFn().  The default implementation uses the
 * malloc() and free() memory allocation interface.  If a different
 * implementation is desired, the platform-specific ported implementation
 * should remove this definition and define its own implementations.
 */

#define DEFAULT_MEMORY_ALLOC_IMPLEMENTATION


/* This definition should be present for non "big-endian" processors only */

#define LCADRF_BYTE_SWAP


/* These definitions are used for memory models */

//#define NEAR - defined in windows.h
//#define FAR - defined in windows.h


/* These definitions are used for function prototypes */

#define STATIC_FN(type) static type NEAR
#define SPROTO(retType, fnName, fnParms) static retType NEAR fnName fnParms
#define XPROTO(retType, fnName, fnParms) extern retType fnName fnParms


/* These sizes for the LdrfStringService... functions determine the
 * maximum size of the cache.  The numbers are good for Windows.  Other
 * platforms may want to tune these values differently. */

#define STRING_SERVICE_MAX_CACHED_FILES 10
#define STRING_SERVICE_MAX_CACHED_STRINGS 50


// End of "Windows, Microsoft C/C++, Intel"

#else


/* These types are used in DRF API but known to be Microsoft-specific */

#ifndef WINDOWS_TYPES
#define WINDOWS_TYPES
typedef char* LPSTR;
typedef const char* LPCSTR;

#endif


/* This definition enables the default implementation of the functions
 * LdrfAllocFn() and LdrfFreeFn().  The default implementation uses the
 * malloc() and free() memory allocation interface.  If a different
 * implementation is desired, the platform-specific ported implementation
 * should remove this definition and define its own implementations.
 */

#define DEFAULT_MEMORY_ALLOC_IMPLEMENTATION


// Define your favorite platform below, using the above "Wintel/Microsoft"
// definition as a model

#error "Requires platform definitions"

#endif

#endif
