/*
 * CACHEUTL.H - Internal include file for the API that contains all
 * definitions for the file I/O cache performance-improvement module
 *
 * Copyright (C) 1997 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#ifndef CACHEUTL_H

#define CACHEUTL_H


/* C runtime */

#ifndef TRUE
#define TRUE    1

#endif

#ifndef FALSE
#define FALSE   0

#endif


/* Some type definitions */

typedef TULong TFileAddr; /* DOS file address */

typedef unsigned int TErrCode;

#ifndef TBOOLBYTE_DEFINED

typedef TUByte TBoolByte; /* Boolean	- stored in a byte */

#define TBOOLBYTE_DEFINED

#endif

#define MAX_CACHE_SIZE  4096

#define SENTINEL_SIZE   0

#ifdef DOS
#define CACHE_START_DEACTIVATE  0xFFFFFFFFUL

#else
#define CACHE_START_DEACTIVATE  0xFFFFFFFF

#endif


/* This data structure is the control structure for the cache */

typedef struct sCacheControl {
    FILE* pFile;
    TULong fileSize;
    TFileAddr curFileAddrOffset;
    TFileAddr nextRWAddr;
    TFileAddr cacheStartAddr;
    TFileAddr cacheEndAddr;
    TErrCode cacheErrStatus;
    TErrCode cacheErrSeek;
    TErrCode cacheErrRead;
    TErrCode cacheErrWrite;
    TUShort firstDirtyByteOffset;
    TUShort dirtyByteCount;
    TUShort cacheSize;
    TBoolByte ftruncateAllowed;
    TBoolByte writeAccess;
    TBoolByte callFnOnErr;
#if SENTINEL_SIZE

    char preSentinel[SENTINEL_SIZE];

#endif

    char theCache[MAX_CACHE_SIZE];
#if SENTINEL_SIZE

    char postSentinel[SENTINEL_SIZE];

#endif

} TCacheControl;

typedef TCacheControl* PCacheControl;

#define CACHE_SEEK_EOF ((TFileAddr)(0xFFFFFFFFL))

XPROTO(PCacheControl,   CacheOpen,      (LPCSTR pFileName,
                         LPCSTR mode,
                         TUShort    cacheSize,
                         TErrCode   errSeek,
                         TErrCode   errRead,
                         TErrCode   errWrite,
                         TErrCode   errOpen,
                         TBoolByte  callFnOnErr));

XPROTO(TFileAddr, CacheWhere, (PCacheControl pCache));

XPROTO(TFileAddr,   CacheSeek,      (PCacheControl  pCache,
                         TFileAddr  fileAddr));

XPROTO(TULong,      CacheRead,      (PCacheControl  pCache,
                         void *     pData,
                         TULong size));

XPROTO(TULong,      CacheWrite,     (PCacheControl  pCache,
                         const void *   pData,
                         TULong size));

XPROTO(TFileAddr,   CacheSeekRead,      (PCacheControl  pCache,
                         TFileAddr  fileAddr,
                         void *     pData,
                         TULong size));

XPROTO(TFileAddr,   CacheSeekWrite,     (PCacheControl  pCache,
                         TFileAddr  fileAddr,
                         const void *   pData,
                         TULong size));

XPROTO(TBoolByte,   CacheReadString,    (PCacheControl  pCache,
                         char *     pString,
                         TUShort    maxChars));

XPROTO(void,    CacheWriteString,   (PCacheControl  pCache,
                         const char *   pString,
                         TUShort    numChars));

XPROTO(void,    CacheWriteChar,     (PCacheControl  pCache,
                         char   c));

XPROTO(void,    CacheFileSizeSet,   (PCacheControl  pCache,
                         TULong newFileSize));

XPROTO(TBoolByte,   CacheClose,     (PCacheControl  pCache,
                         char   closeChar,
                         TErrCode   closeErrCode));

XPROTO(void, CacheError, (TErrCode errCode));

#endif
