/*
 * DRFCAT.C - This is the Resource File API file responsible for catalog access methods
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif

LDRF_FN LdrfOpenCatalog(LPCSTR directory, TBool readOnly, PLdrfFileInfo* ppInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfFileInfo pInfo = NULL;
    // FILE * pFile = NULL;
    PCacheControl pCache = NULL;

    LPSTR pPath = NULL, q;

    TCHAR redirectedPath[MAX_PATH + 1];

    if ((!ppInfo) || (!directory)) 
    {
        errCode = LDRF_ERR_PARAM;
    }
    else
    {
        errCode = LdrfRedirectCatalogPath(directory, redirectedPath, MAX_PATH);
    }

    if (!errCode)
    {
        pPath = NEW_STRING(strlen(redirectedPath) + strlen(DIR_SEPARATOR_STRING) + strlen(CATALOG_NAME));

        if (!pPath) {
            errCode = LDRF_ERR_SYS;
        } else {
            strcpy(pPath, redirectedPath);
            q = strchr(pPath, EOS);

            if ((q) && (q > pPath)) {
                --q;

                if (((* q) != DIR_SEPARATOR_CHAR) && ((* q) != DIR_SEPARATOR_CHAR2)) {
                    *(++q) = DIR_SEPARATOR_CHAR;
                }
                strcpy(++q, CATALOG_NAME);
            } else {
                errCode = LDRF_ERR_PARAM;
            }
        }
    }

    if (!errCode) {
        errCode = LdrfNewFileInfo(&pInfo, LDRF_CATALOG);
    }

    if (!errCode) {
        // pFile = _fsopen(pPath, F_ACCESS_READ_BINARY, _SH_DENYWR);
        pCache = CacheOpen(pPath, F_ACCESS_READ_BINARY, 0, 1, 1, 1, 1, F);

        // if (!pFile) {
        if (!pCache) {
            if (readOnly) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else {
                /* Creating new catalog */
                // pFile = _fsopen(pPath, F_ACCESS_UPDATE_NEW_BINARY, _SH_DENYWR);
                pCache = CacheOpen(pPath, F_ACCESS_UPDATE_NEW_BINARY, 0, 1, 1, 1, 1, F);

                // if (!pFile) {
                if ((!pCache) || (pCache->cacheErrStatus)) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    // pInfo->pFile = pFile;
                    pInfo->pCache = pCache;
                    errCode = LdrfInitCatalogHeader(pInfo);
                    pInfo->userFlag = F;

                    if (!errCode) {
                        pInfo->editAllowed = T;
                        errCode =
                            LdrfSetFileHdrInfo(pInfo, "software from Echelon Corporation, USA", "1-888-LonWorks (+1-888-566-9675)", "www.echelon.com",
                                "http://www.echelon.com", 0, 2, 0, 1);

                        if (!errCode) {
                            TUByte zeroArray[NUM_PROG_ID_BYTES];
                            memset(zeroArray, 0, NUM_PROG_ID_BYTES);
                            errCode = LdrfSetFileVersion(pInfo, 1, 0, 0, zeroArray);
                        }
                    }

                    if (!errCode) {
						// Add the original input directory to the catalog, as
						// that is the one that the user will expect to be added
						// based on legacy functionality
                        errCode = LdrfCatalogAddDir(pInfo, directory);
                    }
                }
            }
        } else {
            if (readOnly) {
                pInfo->editAllowed = F;
            } else {
                /* Edit existing file */
                pInfo->editAllowed = T;
                // fclose(pFile);
                CacheClose(pCache, 0, 1);
                // pFile = _fsopen(pPath, F_ACCESS_UPDATE_EXIS_BINARY, _SH_DENYWR);
                pCache = CacheOpen(pPath, F_ACCESS_UPDATE_EXIS_BINARY, 0, 1, 1, 1, 1, F);
            }

            // if (!pFile) {
            if ((!pCache) || (pCache->cacheErrStatus)) {
                errCode = LDRF_ERR_SYS;
            } else {
                // pInfo->pFile = pFile;
                pInfo->pCache = pCache;
                pCache->ftruncateAllowed = T;
                errCode = LdrfReadCatalogHeader(pInfo);
            }
        }
    }

    if ((!errCode) && (pInfo->editAllowed) && (pInfo->u.cat.stale)) {
        errCode = LdrfCatalogRefresh(pInfo);
    }

    if (pPath) {
        FREE(pPath);
    }

    if (errCode) {
        // if (pFile) {
        if (pCache) {
            // fclose(pFile);
            CacheClose(pCache, 0, 1);
        }

        if (pInfo) {
            LdrfFreeFileInfo(&pInfo);
        }
    } else {
        // pInfo->pFile = pFile;  NOT NEEDED?
        * ppInfo = pInfo;
    }

    return errCode;
} /* end of: LdrfOpenCatalog */

TLdrfErrCodes LdrfReadCatalogHeader(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfReadFileHeader(pInfo, T);

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &(pInfo->u.cat.stale), sizeof(pInfo->u.cat.stale));
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.cat.numDirec));
        pInfo->u.cat.capacityDirec = pInfo->u.cat.numDirec;
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.cat.numLangFiles));
        pInfo->u.cat.capacityLangFiles = pInfo->u.cat.numLangFiles;
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.cat.numTypeFiles));
        pInfo->u.cat.capacityTypeFiles = pInfo->u.cat.numTypeFiles;
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.cat.numFPTFiles));
        pInfo->u.cat.capacityFPTFiles = pInfo->u.cat.numFPTFiles;
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.cat.numFormatFiles));
        pInfo->u.cat.capacityFormatFiles = pInfo->u.cat.numFormatFiles;
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.cat.numDirec, pInfo->u.cat.numDirec, &(pInfo->u.cat.pDirDirec));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.cat.numLangFiles, pInfo->u.cat.numLangFiles, &(pInfo->u.cat.pDirLangFiles));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.cat.numTypeFiles, pInfo->u.cat.numTypeFiles, &(pInfo->u.cat.pDirTypeFiles));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.cat.numFPTFiles, pInfo->u.cat.numFPTFiles, &(pInfo->u.cat.pDirFPTFiles));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.cat.numFormatFiles, pInfo->u.cat.numFormatFiles, &(pInfo->u.cat.pDirFormatFiles));
    }

    return errCode;
} /* end of: LdrfReadCatalogHeader */

TLdrfErrCodes LdrfInitCatalogHeader(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfInitFileHeader(pInfo, 2);

    if (!errCode) {
        pInfo->headerSize += SIZE_CATALOG_FILE_HDR;
        errCode = LdrfIncreaseCapacityCatDirec(pInfo);

        if (!errCode) {
            errCode = LdrfIncreaseCapacityCatLangFiles(pInfo);
        }

        if (!errCode) {
            errCode = LdrfIncreaseCapacityCatTypeFiles(pInfo);
        }

        if (!errCode) {
            errCode = LdrfIncreaseCapacityCatFPTFiles(pInfo);
        }

        if (!errCode) {
            errCode = LdrfIncreaseCapacityCatFormatFiles(pInfo);
        }
    }

    return errCode;
} /* end of: LdrfInitCatalogHeader */

TLdrfErrCodes LdrfWriteCatalogHeader(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfWriteFileHeader(pInfo);

    if (!errCode) {
        errCode = LdrfFileWriteBytes(pInfo, &(pInfo->u.cat.stale), sizeof(pInfo->u.cat.stale));
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &(pInfo->u.cat.numDirec));
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &(pInfo->u.cat.numLangFiles));
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &(pInfo->u.cat.numTypeFiles));
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &(pInfo->u.cat.numFPTFiles));
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &(pInfo->u.cat.numFormatFiles));
    }

    if (!errCode) {
        errCode = LdrfWriteDirectory(pInfo, pInfo->u.cat.numDirec, pInfo->u.cat.pDirDirec);
    }

    if (!errCode) {
        errCode = LdrfWriteDirectory(pInfo, pInfo->u.cat.numLangFiles, pInfo->u.cat.pDirLangFiles);
    }

    if (!errCode) {
        errCode = LdrfWriteDirectory(pInfo, pInfo->u.cat.numTypeFiles, pInfo->u.cat.pDirTypeFiles);
    }

    if (!errCode) {
        errCode = LdrfWriteDirectory(pInfo, pInfo->u.cat.numFPTFiles, pInfo->u.cat.pDirFPTFiles);
    }

    if (!errCode) {
        errCode = LdrfWriteDirectory(pInfo, pInfo->u.cat.numFormatFiles, pInfo->u.cat.pDirFormatFiles);
    }

    if (!errCode) {
        errCode = LdrfUpdateHeaderCRC(pInfo);
    }

    return errCode;
} /* end of: LdrfWriteCatalogHeader */

TLdrfErrCodes LdrfFreeCatalogFileInfo(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    FREE(pInfo->u.cat.pDirDirec);
    FREE(pInfo->u.cat.pDirLangFiles);
    FREE(pInfo->u.cat.pDirTypeFiles);
    FREE(pInfo->u.cat.pDirFPTFiles);
    FREE(pInfo->u.cat.pDirFormatFiles);

    return errCode;
} /* end of: LdrfFreeCatalogFileInfo */

TLdrfErrCodes LdrfExpandCatalogFileHeader(PLdrfFileInfo pInfo, TULong headerSizeIncrement) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    if (headerSizeIncrement) {
        pInfo->headerSize += headerSizeIncrement;

        if ((pInfo->dataAddress) && (pInfo->dataSize > SIZE_DATA_CRC)) {
            while ((!errCode) && (pInfo->dataAddress < pInfo->headerSize)) {
                errCode = LdrfRelocateCatalogFileData(pInfo);
            }
        } else if (pInfo->dataAddress < pInfo->headerSize) {
            pInfo->dataAddress = pInfo->headerSize;

            if (pInfo->dataSize < SIZE_DATA_CRC) {
                pInfo->dataSize = SIZE_DATA_CRC;
            }
        }
    }

    return errCode;
} /* end of: LdrfExpandCatalogFileHeader */

TLdrfErrCodes LdrfRelocateCatalogFileData(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TULong newRecordAddress = pInfo->dataAddress + pInfo->dataSize;
    TULong lowestRecordAddress = newRecordAddress;
    TULong lowestDirecIndex, lowestLangFileIndex, lowestTypeFileIndex, lowestFPTFileIndex, lowestFormatFileIndex;


    /* Search directories for the location of the lowest addressed data record */
    errCode = LdrfSearchDirForMinEntry(pInfo->u.cat.pDirDirec, pInfo->u.cat.numDirec, &lowestDirecIndex, &lowestRecordAddress);

    if (!errCode) {
        errCode = LdrfSearchDirForMinEntry(pInfo->u.cat.pDirLangFiles, pInfo->u.cat.numLangFiles, &lowestLangFileIndex, &lowestRecordAddress);
    }

    if (!errCode) {
        errCode = LdrfSearchDirForMinEntry(pInfo->u.cat.pDirTypeFiles, pInfo->u.cat.numTypeFiles, &lowestTypeFileIndex, &lowestRecordAddress);
    }

    if (!errCode) {
        errCode = LdrfSearchDirForMinEntry(pInfo->u.cat.pDirFPTFiles, pInfo->u.cat.numFPTFiles, &lowestFPTFileIndex, &lowestRecordAddress);
    }

    if (!errCode) {
        errCode = LdrfSearchDirForMinEntry(pInfo->u.cat.pDirFormatFiles, pInfo->u.cat.numFormatFiles, &lowestFormatFileIndex, &lowestRecordAddress);
    }


    /* Analyze results - look for some error conditions */
    if (!errCode) {
        if (lowestRecordAddress < pInfo->dataAddress) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (lowestRecordAddress > pInfo->dataAddress + SIZE_DATA_CRC) {
            pInfo->dataAddress = lowestRecordAddress - SIZE_DATA_CRC;
            pInfo->updateData = T;
            pInfo->updateHeader = T;
        } else {
            if ((lowestDirecIndex) && (pInfo->u.cat.pDirDirec[lowestDirecIndex] > lowestRecordAddress)) {
                lowestDirecIndex = 0;
            }

            if ((lowestLangFileIndex) && (pInfo->u.cat.pDirLangFiles[lowestLangFileIndex] > lowestRecordAddress)) {
                lowestLangFileIndex = 0;
            }

            if ((lowestTypeFileIndex) && (pInfo->u.cat.pDirTypeFiles[lowestTypeFileIndex] > lowestRecordAddress)) {
                lowestTypeFileIndex = 0;
            }

            if ((lowestFPTFileIndex) && (pInfo->u.cat.pDirFPTFiles[lowestFPTFileIndex] > lowestRecordAddress)) {
                lowestFPTFileIndex = 0;
            }

            if ((lowestFormatFileIndex) && (pInfo->u.cat.pDirFormatFiles[lowestFormatFileIndex] > lowestRecordAddress)) {
                lowestFormatFileIndex = 0;
            }

            if (!errCode) {
                errCode = LdrfMoveFileRec(pInfo, lowestRecordAddress, newRecordAddress, OFFSET_CATALOG_FILE_REC_SIZE);
            }

            if (!errCode) {
                if (lowestDirecIndex) {
                    if (pInfo->u.cat.pDirDirec[lowestDirecIndex] != lowestRecordAddress) {
                        errCode = LDRF_ERR_INTERNAL;
                    } else {
                        pInfo->u.cat.pDirDirec[lowestDirecIndex] = newRecordAddress;
                    }
                } else if (lowestLangFileIndex) {
                    if (pInfo->u.cat.pDirLangFiles[lowestLangFileIndex] != lowestRecordAddress) {
                        errCode = LDRF_ERR_INTERNAL;
                    } else {
                        pInfo->u.cat.pDirLangFiles[lowestLangFileIndex] = newRecordAddress;
                    }
                } else if (lowestTypeFileIndex) {
                    if (pInfo->u.cat.pDirTypeFiles[lowestTypeFileIndex] != lowestRecordAddress) {
                        errCode = LDRF_ERR_INTERNAL;
                    } else {
                        pInfo->u.cat.pDirTypeFiles[lowestTypeFileIndex] = newRecordAddress;
                    }
                } else if (lowestFPTFileIndex) {
                    if (pInfo->u.cat.pDirFPTFiles[lowestFPTFileIndex] != lowestRecordAddress) {
                        errCode = LDRF_ERR_INTERNAL;
                    } else {
                        pInfo->u.cat.pDirFPTFiles[lowestFPTFileIndex] = newRecordAddress;
                    }
                } else if (lowestFormatFileIndex) {
                    if (pInfo->u.cat.pDirFormatFiles[lowestFormatFileIndex] != lowestRecordAddress) {
                        errCode = LDRF_ERR_INTERNAL;
                    } else {
                        pInfo->u.cat.pDirFormatFiles[lowestFormatFileIndex] = newRecordAddress;
                    }
                } else {
                    errCode = LDRF_ERR_INTERNAL;
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfRelocateCatalogFileData */

TLdrfErrCodes LdrfIncreaseCapacityCatDirec(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode;
    TULong headerSizeIncrement, capacityDummy = pInfo->u.cat.capacityDirec;

    errCode = LdrfIncreaseCapacity(&capacityDummy, MAX_DIREC_CAPACITY, &headerSizeIncrement, &(pInfo->u.cat.pDirDirec), NULL, NULL);

    if (!errCode) {
        pInfo->u.cat.capacityDirec = (TUShort)capacityDummy;
        errCode = LdrfExpandCatalogFileHeader(pInfo, headerSizeIncrement);
    }

    return errCode;
} /* end of: LdrfIncreaseCapacityCatDirec */

TLdrfErrCodes LdrfIncreaseCapacityCatLangFiles(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode;
    TULong headerSizeIncrement, capacityDummy = pInfo->u.cat.capacityLangFiles;

    errCode = LdrfIncreaseCapacity(&capacityDummy, MAX_FILE_CAPACITY, &headerSizeIncrement, &(pInfo->u.cat.pDirLangFiles), NULL, NULL);

    if (!errCode) {
        pInfo->u.cat.capacityLangFiles = (TUShort)capacityDummy;
        errCode = LdrfExpandCatalogFileHeader(pInfo, headerSizeIncrement);
    }

    return errCode;
} /* end of: LdrfIncreaseCapacityCatLangFiles */

TLdrfErrCodes LdrfIncreaseCapacityCatTypeFiles(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode;
    TULong headerSizeIncrement, capacityDummy = pInfo->u.cat.capacityTypeFiles;

    errCode = LdrfIncreaseCapacity(&capacityDummy, MAX_FILE_CAPACITY, &headerSizeIncrement, &(pInfo->u.cat.pDirTypeFiles), NULL, NULL);

    if (!errCode) {
        pInfo->u.cat.capacityTypeFiles = (TUShort)capacityDummy;
        errCode = LdrfExpandCatalogFileHeader(pInfo, headerSizeIncrement);
    }

    return errCode;
} /* end of: LdrfIncreaseCapacityCatTypeFiles */

TLdrfErrCodes LdrfIncreaseCapacityCatFPTFiles(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode;
    TULong headerSizeIncrement, capacityDummy = pInfo->u.cat.capacityFPTFiles;

    errCode = LdrfIncreaseCapacity(&capacityDummy, MAX_FILE_CAPACITY, &headerSizeIncrement, &(pInfo->u.cat.pDirFPTFiles), NULL, NULL);

    if (!errCode) {
        pInfo->u.cat.capacityFPTFiles = (TUShort)capacityDummy;
        errCode = LdrfExpandCatalogFileHeader(pInfo, headerSizeIncrement);
    }

    return errCode;
} /* end of: LdrfIncreaseCapacityCatFPTFiles */

TLdrfErrCodes LdrfIncreaseCapacityCatFormatFiles(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode;
    TULong headerSizeIncrement, capacityDummy = pInfo->u.cat.capacityFormatFiles;

    errCode = LdrfIncreaseCapacity(&capacityDummy, MAX_FILE_CAPACITY, &headerSizeIncrement, &(pInfo->u.cat.pDirFormatFiles), NULL, NULL);

    if (!errCode) {
        pInfo->u.cat.capacityFormatFiles = (TUShort)capacityDummy;
        errCode = LdrfExpandCatalogFileHeader(pInfo, headerSizeIncrement);
    }

    return errCode;
} /* end of: LdrfIncreaseCapacityCatFormatFiles */

LDRF_FN LdrfGetCatalogInfo(PLdrfFileInfo pInfo, PBool pStale, PUShort pNumDirec, PUShort pNumLangFiles, PUShort pNumTypeFiles, PUShort pNumFPTFiles,
    PUShort pNumFormatFiles)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);

    if (!errCode) {
        if (pStale) {
            * pStale = pInfo->u.cat.stale;
        }

        if (pNumDirec) {
            * pNumDirec = pInfo->u.cat.numDirec;
        }

        if (pNumLangFiles) {
            * pNumLangFiles = pInfo->u.cat.numLangFiles;
        }

        if (pNumTypeFiles) {
            * pNumTypeFiles = pInfo->u.cat.numTypeFiles;
        }

        if (pNumFPTFiles) {
            * pNumFPTFiles = pInfo->u.cat.numFPTFiles;
        }

        if (pNumFormatFiles) {
            * pNumFormatFiles = pInfo->u.cat.numFormatFiles;
        }
    }

    return errCode;
} /* end of: LdrfGetCatalogInfo */

LDRF_FN LdrfCloseCatalog(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);

    if (!errCode) {
        errCode = LdrfCloseFile(pInfo);
    }

    return errCode;
} /* end of: LdrfCloseCatalog */

LDRF_FN LdrfCatalogAddDir(PLdrfFileInfo pInfo, LPCSTR newDir) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);
    TUShort index = 0;
    LPSTR pDirecName, q;

    if (!errCode) {
        if (!(pInfo->editAllowed)) {
            errCode = LDRF_ERR_NO_ACCESS;
        } else if (!newDir) {
            errCode = LDRF_ERR_PARAM;
        } else {
            pDirecName = NEW_STRING(strlen(newDir) + 1);

            if (!pDirecName) {
                errCode = LDRF_ERR_SYS;
            } else {
                strcpy(pDirecName, newDir);
                q = strchr(pDirecName, EOS);

                if ((q) && (q > pDirecName)) {
                    --q;

                    if (((* q) != DIR_SEPARATOR_CHAR) && ((* q) != DIR_SEPARATOR_CHAR2)) {
                        *(++q) = DIR_SEPARATOR_CHAR;
                        *(++q) = EOS;
                    }
                }
                errCode = LdrfFindInUnsortedDir(pInfo, pInfo->u.cat.pDirDirec, pInfo->u.cat.numDirec, pDirecName, NULL, LdrfDirecCompare);
            }

            if (!errCode) {
                errCode = LDRF_ERR_DUPLICATE;
            } else if (errCode == LDRF_ERR_NOT_FOUND) {
                errCode = LDRF_ERR_NONE;
                index = pInfo->u.cat.numDirec + 1;

                if (index > pInfo->u.cat.capacityDirec) {
                    errCode = LdrfIncreaseCapacityCatDirec(pInfo);
                }
            }

            if (!errCode) {
                TULong startAddress = pInfo->dataAddress + pInfo->dataSize;
                errCode = LdrfDirecWrite(pInfo, startAddress, index, pDirecName);

                if (!errCode) {
                    ++(pInfo->u.cat.numDirec);
                    pInfo->u.cat.pDirDirec[index] = startAddress;
                }
            }
            FREE(pDirecName);
        }
    }

    if (!errCode) {
        pInfo->u.cat.stale = T;
    }

    return errCode;
} /* end of: LdrfCatalogAddDir */

LDRF_FN LdrfCatalogGetDir(PLdrfFileInfo pInfo, TUShort index, LPSTR pDirName, PUShort pLength) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);

    if (!errCode) {
        if ((index == 0) || (index > pInfo->u.cat.numDirec)) {
            errCode = LDRF_ERR_PARAM;
        } else if ((!(pInfo->u.cat.pDirDirec)) || (!(pInfo->u.cat.pDirDirec[index]))) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            errCode = LdrfFileSeek(pInfo, pInfo->u.cat.pDirDirec[index] + OFFSET_CATALOG_FILE_REC_SIZE + sizeof(TUShort));
        }
    }

    if (!errCode) {
        char* p = NULL;
        errCode = LdrfFileReadLString(pInfo, &p);

        if (!errCode) {
            errCode = LdrfTransferStringOut(pDirName, pLength, p);
        }
        FREE(p);
    }

    return errCode;
} /* end of: LdrfCatalogGetDir */

LDRF_FN LdrfCatalogRmvDir(PLdrfFileInfo pInfo, LPCSTR oldDir) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);
    TULong index;
    LPSTR pDirecName, q;

    if (!errCode) {
        if (!(pInfo->editAllowed)) {
            errCode = LDRF_ERR_NO_ACCESS;
        } else if (!oldDir) {
            errCode = LDRF_ERR_PARAM;
        } else {
            pDirecName = NEW_STRING(strlen(oldDir) + 1);

            if (!pDirecName) {
                errCode = LDRF_ERR_SYS;
            } else {
                strcpy(pDirecName, oldDir);
                q = strchr(pDirecName, EOS);

                if ((q) && (q > pDirecName)) {
                    --q;

                    if (((* q) != DIR_SEPARATOR_CHAR) && ((* q) != DIR_SEPARATOR_CHAR2)) {
                        *(++q) = DIR_SEPARATOR_CHAR;
                        *(++q) = EOS;
                    }
                }
                errCode = LdrfFindInUnsortedDir(pInfo, pInfo->u.cat.pDirDirec, pInfo->u.cat.numDirec, pDirecName, &index, LdrfDirecCompare);
            }
        }
    }

    if (!errCode) {
        if (index == 1) {
            errCode = LDRF_ERR_PARAM;
        } else {
            if (index != pInfo->u.cat.numDirec) {
                pInfo->u.cat.pDirDirec[index] = pInfo->u.cat.pDirDirec[pInfo->u.cat.numDirec];
            }
            --(pInfo->u.cat.numDirec);
        }
    }

    if (!errCode) {
        TLdrfErrCodes ec2, ec3, ec4, ec5;
        ec2 = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirLangFiles, &(pInfo->u.cat.numLangFiles), T, (TUShort)index);
        ec3 = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirTypeFiles, &(pInfo->u.cat.numTypeFiles), T, (TUShort)index);
        ec4 = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirFPTFiles, &(pInfo->u.cat.numFPTFiles), T, (TUShort)index);
        ec5 = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirFormatFiles, &(pInfo->u.cat.numFormatFiles), T, (TUShort)index);

        if (ec2) {
            errCode = ec2;
        } else if (ec3) {
            errCode = ec3;
        } else if (ec4) {
            errCode = ec4;
        } else {
            errCode = ec5;
        }
    }

    return errCode;
} /* end of: LdrfCatalogRmvDir */

TLdrfErrCodes LdrfDirecWrite(PLdrfFileInfo pInfo, TULong startAddress, TUShort index, LPSTR pDirecName) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, startAddress);

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &index);
    }

    if (!errCode) {
        /* The next field is the size, but don't know that value yet,
         * so just write index again as a convenient way to advance two bytes */
        errCode = LdrfFileWriteShort(pInfo, &index);
    }

    if (!errCode) {
        TUShort len = strlen(pDirecName);

        if (len > 255) {
            errCode = LDRF_ERR_PARAM;
        } else {
            errCode = LdrfFileWriteLString(pInfo, pDirecName, (TUByte)len);
        }
    }

    if (!errCode) {
        errCode = LdrfUpdateDataSize(pInfo);
    }

    if (!errCode) {
        errCode = LdrfUpdateRecSize(pInfo, startAddress, OFFSET_CATALOG_FILE_REC_SIZE);
    }

    return errCode;
} /* end of: LdrfDirecWrite */

TLdrfErrCodes LdrfDirecCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    TLdrfErrCodes errCode;
    LPSTR pEntry2 = NULL;

    /* Skip to the name string */
    errCode = LdrfFileSeek(pInfo, entry2Addr + OFFSET_CATALOG_FILE_REC_SIZE + sizeof(TUShort));

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pEntry2);
    }

    if (!errCode) {
        * pRC = STRICMP(pEntry1, pEntry2);
    }
    FREE(pEntry2);

    return errCode;
} /* end of: LdrfDirecCompare */

TLdrfErrCodes LdrfSelectDirForFileType(PLdrfFileInfo pInfo, TLdrfFileType fileType, TULong* * ppDir, PUShort pNum) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    if ((!ppDir) || (!pNum)) {
        errCode = LDRF_ERR_PARAM;
    } else {
        switch (fileType) {
            case LDRF_LANG_RESOURCE:
                * ppDir = pInfo->u.cat.pDirLangFiles;
                * pNum = pInfo->u.cat.numLangFiles;
                break;

            case LDRF_TYPE:
                * ppDir = pInfo->u.cat.pDirTypeFiles;
                * pNum = pInfo->u.cat.numTypeFiles;
                break;

            case LDRF_FPT:
                * ppDir = pInfo->u.cat.pDirFPTFiles;
                * pNum = pInfo->u.cat.numFPTFiles;
                break;

            case LDRF_FORMAT:
                * ppDir = pInfo->u.cat.pDirFormatFiles;
                * pNum = pInfo->u.cat.numFormatFiles;
                break;

            default:
                errCode = LDRF_ERR_FILE_TYPE;
        }
    }

    return errCode;
} /* end of: LdrfSelectDirForFileType */

LDRF_FN LdrfCatalogRefresh(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);

    if ((!errCode) && (!(pInfo->editAllowed))) {
        errCode = LDRF_ERR_NO_ACCESS;
    }

    /* Delete all file records in the catalog with a non-zero directory index.
     * Files which have a zero directory index were added specifically through the
     * AddFile interface, rather than via a directory.	These files must be verified,
     * but not removed blindly.  All files with non-zero directory entries are purged. */
    if (!errCode) {
        errCode = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirLangFiles, &(pInfo->u.cat.numLangFiles), F, 0);
    }

    if (!errCode) {
        errCode = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirTypeFiles, &(pInfo->u.cat.numTypeFiles), F, 0);
    }

    if (!errCode) {
        errCode = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirFPTFiles, &(pInfo->u.cat.numFPTFiles), F, 0);
    }

    if (!errCode) {
        errCode = LdrfCatalogPurgeFiles(pInfo, pInfo->u.cat.pDirFormatFiles, &(pInfo->u.cat.numFormatFiles), F, 0);
    }


    /* Step through each file still present, remove it, and re-add it. */
    if ((!errCode) && ((pInfo->u.cat.numLangFiles) || (pInfo->u.cat.numTypeFiles) || (pInfo->u.cat.numFPTFiles) || (pInfo->u.cat.numFormatFiles))) {
        errCode = LDRF_ERR_UNIMPL;
    }


    /* AT THIS POINT, THE CATALOG FILE ONLY CONTAINS DIRECTORY RECORDS.
     * THE FOLLOWING CODE FINDS THE LARGEST IRG (Inter-Record-Gap) AND
     * MOVES THE HIGHEST DIRECTORY RECORD DOWN TO FILL THE GAP. */
    if (!errCode) {
        TUShort index, index2;
        TULong largestIRGSize = 0, largestIRGAddress = 0;
        TULong thisRecStartAddress, thisRecLastDataAddress;
        TUShort thisRecSize;
        TULong thisIRGAddress, thisIRGSize;
        TULong highestRecStartAddress = 0;
        TUShort highestRecSize = 0, highestRecIndex = 0;

        for ( index = 1; index <= pInfo->u.cat.numDirec; ++index ) {
            /* Get stats on this record */
            thisRecStartAddress = pInfo->u.cat.pDirDirec[index];
            errCode = LdrfFileSeek(pInfo, thisRecStartAddress + OFFSET_CATALOG_FILE_REC_SIZE);

            if (!errCode) {
                errCode = LdrfFileReadShort(pInfo, &thisRecSize);
            }

            if (!errCode) {
                errCode = LdrfFileWhere(pInfo);
            }

            if (!errCode) {
                thisRecLastDataAddress = pInfo->currentAddress + thisRecSize - 1;
            }

            if (!errCode) {
                if (thisRecStartAddress > highestRecStartAddress) {
                    highestRecStartAddress = thisRecStartAddress;
                    highestRecSize = thisRecSize + sizeof(thisRecSize) + OFFSET_CATALOG_FILE_REC_SIZE;
                    highestRecIndex = index;
                }
            }

            if (errCode) {
                break;
            }

            /* Attempt to find an IRG after this record */
            thisIRGAddress = thisRecLastDataAddress + 1;
            thisIRGSize = 0;

            for ( index2 = 1; index2 <= pInfo->u.cat.numDirec; ++index2 ) {
                TULong otherRecStartAddress = pInfo->u.cat.pDirDirec[index2];

                if (otherRecStartAddress == thisIRGAddress) {
                    /* Consecutive records - no IRG! */
                    thisIRGSize = 0;
                    break;
                } else if (otherRecStartAddress > thisIRGAddress) {
                    /* Possible IRG - but may contain data too */
                    TULong newIRGSize = otherRecStartAddress - thisIRGAddress;

                    if ((!thisIRGSize) || (newIRGSize < thisIRGSize)) {
                        thisIRGSize = newIRGSize;
                    }
                /* else if other rec is lower, then not part of IRG */
                }
            }

            /* Update the IRG stats */
            if ((thisIRGAddress != 0) && (thisIRGSize > largestIRGSize)) {
                largestIRGSize = thisIRGSize;
                largestIRGAddress = thisIRGAddress;
            }
        }

        /* Now, move the record, and fix the directory */
        if ((!errCode) && (highestRecIndex) && (largestIRGSize) && (highestRecSize) && (largestIRGSize >= highestRecSize)) {
            errCode = LdrfMoveFileRec(pInfo, highestRecStartAddress, largestIRGAddress, OFFSET_CATALOG_FILE_REC_SIZE);

            if (!errCode) {
                pInfo->u.cat.pDirDirec[highestRecIndex] = largestIRGAddress;
            }
        }
    }


    /* Find the upper active data address, and shorten the file to that size */
    if (!errCode) {
        TULong upperDataAddress = pInfo->dataAddress;
        errCode = LdrfCatalogFindMaxDataAddress(pInfo, pInfo->u.cat.pDirLangFiles, pInfo->u.cat.numLangFiles, &upperDataAddress);

        if (!errCode) {
            errCode = LdrfCatalogFindMaxDataAddress(pInfo, pInfo->u.cat.pDirTypeFiles, pInfo->u.cat.numTypeFiles, &upperDataAddress);
        }

        if (!errCode) {
            errCode = LdrfCatalogFindMaxDataAddress(pInfo, pInfo->u.cat.pDirFPTFiles, pInfo->u.cat.numFPTFiles, &upperDataAddress);
        }

        if (!errCode) {
            errCode = LdrfCatalogFindMaxDataAddress(pInfo, pInfo->u.cat.pDirFormatFiles, pInfo->u.cat.numFormatFiles, &upperDataAddress);
        }

        if (!errCode) {
            errCode = LdrfCatalogFindMaxDataAddress(pInfo, pInfo->u.cat.pDirDirec, pInfo->u.cat.numDirec, &upperDataAddress);
        }

        if (!errCode) {
            PCacheControl pCache = pInfo->pCache;

            pInfo->dataSize = upperDataAddress - pInfo->dataAddress + 1;
            CacheFileSizeSet(pCache, upperDataAddress + 1);

            if (pCache->cacheErrStatus) {
                errCode = LDRF_ERR_SYS;
            } else {
                pInfo->updateData = T;
                pInfo->updateHeader = T;
            }
        }
    }


    /* Step through each directory.  For each directory, attempt to
     * add each file in the directory. */
    if (!errCode) {
        TULong count, handle;
        char path[_MAX_PATH];
#ifdef _WINDOWS
        /* The following declarations are only needed for the "splitpath" call
         * in the code below. */

        char basename[_MAX_PATH];
        char ext[_MAX_PATH];
#endif

        char* p, * q;
        struct _finddata_t fileinfo;

        for ( count = 1; count <= pInfo->u.cat.numDirec; ++count ) {
            p = NULL;

            if (!pInfo->u.cat.pDirDirec[count]) {
                errCode = LDRF_ERR_INTERNAL;
            } else {
                errCode = LdrfFileSeek(pInfo, pInfo->u.cat.pDirDirec[count] + (OFFSET_CATALOG_FILE_REC_SIZE + sizeof(TUShort)));
            }

            if (errCode) break;
            errCode = LdrfFileReadLString(pInfo, &p);

            if (errCode) break;

            if (strlen(p) > (_MAX_PATH - 4)) {
                errCode = LDRF_ERR_INTERNAL;
            } else {
                strcpy(path, p);
                q = strchr(path, '\0');

                if (q) {
                    strcpy(q, "*.*");
                    handle = _findfirst(path, &fileinfo);

                    if (handle != (TULong)(-1)) {
                        do {
                            if (!(fileinfo.attrib &(_A_HIDDEN + _A_SUBDIR + _A_SYSTEM))) {
#ifdef _WINDOWS
                                /* Only on windows, split the path and filter out some basenames that get
                                 * created as backup files by the Resource Editor (which is a Windows-only
                                 * tool).  Presumably the backup (and marked-deleted) files won't be copied
                                 * to any non-Windows platforms, but if they do, we perform similar filtering
                                 * using the entire pathname.  The "_splitpath" call isn't available on non-
                                 * Windows platforms.
                                 */

                                _splitpath(fileinfo.name, NULL, NULL, basename, ext);

                                if ((!strchr(basename, '~')) && (!strchr(ext, '~')) && (!strchr(ext, '#')) && (!strchr(ext, '$')))

#else

                                if ((!strchr(fileinfo.name, '~')) && (!strchr(fileinfo.name, '#')) && (!strchr(fileinfo.name, '$')))

#endif

                                {
                                    * q = '\0';

                                    if (strlen(fileinfo.name) + strlen(path) <= (_MAX_PATH - 1)) {
                                        strcpy(q, fileinfo.name);
                                        LdrfCatalogAddFileInternal(pInfo, path, (TUShort)count);
                                    }
                                }
                            }
                        } while (!(_findnext(handle, &fileinfo)));
                        _findclose(handle);
                    }
                }
            }
            FREE(p);

            if (errCode) break;
        }
    }

    if (!errCode) {
        pInfo->u.cat.stale = F;
    }

    return errCode;
} /* end of: LdrfCatalogRefresh */

TLdrfErrCodes LdrfCatalogFindMaxDataAddress(PLdrfFileInfo pInfo, PULong pDir, TUShort numEntries, PULong pUpperDataAddress) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TUShort index, recSize;
    TULong upperDataAddress, thisRecLastDataAddress, recStartAddress;

    if (pUpperDataAddress) {
        upperDataAddress = *pUpperDataAddress;

        for ( index = 1; index <= numEntries; ++index ) {
            recStartAddress = pDir[index];
            errCode = LdrfFileSeek(pInfo, recStartAddress + OFFSET_CATALOG_FILE_REC_SIZE);

            if (!errCode) {
                errCode = LdrfFileReadShort(pInfo, &recSize);
            }

            if (!errCode) {
                errCode = LdrfFileWhere(pInfo);
            }

            if (!errCode) {
                thisRecLastDataAddress = pInfo->currentAddress + recSize - 1;
            }

            if (!errCode) {
                if (thisRecLastDataAddress > upperDataAddress) {
                    upperDataAddress = thisRecLastDataAddress;
                }
            }

            if (errCode) {
                break;
            }
        }

        if (!errCode) {
            * pUpperDataAddress = upperDataAddress;
        }
    }

    return errCode;
} /* end of: LdrfCatalogFindMaxDataAddress */

TLdrfErrCodes LdrfCatalogPurgeFiles(PLdrfFileInfo pInfo, PULong pDir, PUShort pNumEntries, TBool match, TUShort matchDirecIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    TUShort index, oldNum, newNum, direcIndex;
    TBool purgeIt;

    oldNum = *pNumEntries;
    newNum = 0;

    for ( index = 1; index <= oldNum; ++index ) {
        errCode = LdrfGetCatalogFileInfo(pInfo, pDir[index], NULL, NULL, NULL, NULL, NULL, &direcIndex, NULL, NULL, NULL);

        if (errCode) break;

        purgeIt = F;

        if (match) {
            if (matchDirecIndex == direcIndex) {
                purgeIt = T;
            }
        } else if (matchDirecIndex != direcIndex) {
            purgeIt = T;
        }

        if (!purgeIt) {
            ++newNum;

            if (index != newNum) {
                pDir[newNum] = pDir[index];
            }
        }
    }
    * pNumEntries = newNum;

    return errCode;
} /* end of: LdrfCatalogPurgeFiles */

LDRF_FN LdrfCatalogAddFile(PLdrfFileInfo pInfo, LPCSTR newFile) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);

    if (!errCode) {
        if (!(pInfo->editAllowed)) {
            errCode = LDRF_ERR_NO_ACCESS;
        } else {
            errCode = LdrfCatalogAddFileInternal(pInfo, newFile, 0);
        }
    }

    return errCode;
} /* end of: LdrfCatalogAddFile */

TLdrfErrCodes LdrfCatalogAddFileInternal(PLdrfFileInfo pInfo, LPCSTR newFile, TUShort direcIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NOT_FOUND;
    PLdrfFileInfo pTempInfo = NULL;
    TLdrfFileType fileType;
    TULong locale = 0;
    TUShort majorDataVer, index;
    TUByte minorDataVer, scope, refID[NUM_PROG_ID_BYTES];
    PULong pDir = NULL;
    PUShort pNum = NULL;

    if (!newFile) {
        errCode = LDRF_ERR_PARAM;
    } else {
        errCode = LdrfGetFileType(newFile, &fileType);
    }

    if ((errCode == LDRF_ERR_FILE_TYPE) && (fileType == LDRF_WILD_CARD)) {
        if (strlen(newFile) > 4) {
            const char* p = strrchr(newFile, '.');

            if ((p) && (!STRICMP(p, ".FMT"))) {
                errCode = LDRF_ERR_NONE;
                fileType = LDRF_FORMAT;
            }
        }
    }

    if (!errCode) {
        if (fileType == LDRF_FORMAT) {
            errCode = LdrfScanFormatFile(newFile, &majorDataVer, &minorDataVer, &scope, refID);
        } else {
            errCode = LdrfOpenFile(newFile, fileType, 1, T, &pTempInfo);

            if (!errCode) {
                PUByteArray pRefID = NULL;
                errCode = LdrfGetFileVersion(pTempInfo, NULL, NULL, &majorDataVer, &minorDataVer, NULL, NULL, NULL, NULL, NULL, NULL, &scope, &pRefID);

                if (!errCode) {
                    memcpy(refID, pRefID, NUM_PROG_ID_BYTES);
                }
                LdrfFreeByteArray(pRefID);
            }

            if ((!errCode) && (fileType == LDRF_LANG_RESOURCE)) {
                errCode = LdrfGetLangFileInfo(pTempInfo, &locale, NULL);
            }

            if (pTempInfo) {
                TLdrfErrCodes ec2 = LdrfCloseFile(pTempInfo);

                if (!errCode) {
                    errCode = ec2;
                }
            }
        }
    }

    if (!errCode) {
        switch (fileType) {
            case LDRF_LANG_RESOURCE:
                pNum = &(pInfo->u.cat.numLangFiles);

                if (*pNum >= pInfo->u.cat.capacityLangFiles) {
                    errCode = LdrfIncreaseCapacityCatLangFiles(pInfo);
                }
                pDir = pInfo->u.cat.pDirLangFiles;
                break;

            case LDRF_TYPE:
                pNum = &(pInfo->u.cat.numTypeFiles);

                if (*pNum >= pInfo->u.cat.capacityTypeFiles) {
                    errCode = LdrfIncreaseCapacityCatTypeFiles(pInfo);
                }
                pDir = pInfo->u.cat.pDirTypeFiles;
                break;

            case LDRF_FPT:
                pNum = &(pInfo->u.cat.numFPTFiles);

                if (*pNum >= pInfo->u.cat.capacityFPTFiles) {
                    errCode = LdrfIncreaseCapacityCatFPTFiles(pInfo);
                }
                pDir = pInfo->u.cat.pDirFPTFiles;
                break;

            case LDRF_FORMAT:
                pNum = &(pInfo->u.cat.numFormatFiles);

                if (*pNum >= pInfo->u.cat.capacityFormatFiles) {
                    errCode = LdrfIncreaseCapacityCatFormatFiles(pInfo);
                }
                pDir = pInfo->u.cat.pDirFormatFiles;
                break;

            default:
                errCode = LDRF_ERR_INTERNAL;
        }
    }

    if (!errCode) {
        errCode = LdrfFindInUnsortedDir(pInfo, pDir, *pNum, newFile, NULL, LdrfCatalogFileCompare);

        if (!errCode) {
            errCode = LDRF_ERR_DUPLICATE;
        } else if (errCode == LDRF_ERR_NOT_FOUND) {
            errCode = LDRF_ERR_NONE;
        }
    }

    if (!errCode) {
        TULong startAddress = pInfo->dataAddress + pInfo->dataSize;
        index = *pNum + 1;
        errCode = LdrfCatalogFileWrite(pInfo, startAddress, index, fileType, scope, refID, direcIndex, majorDataVer, minorDataVer, locale, newFile);

        if (!errCode) {
            ++(* pNum);
            pDir[index] = startAddress;
        }
    }

    return errCode;
} /* end of: LdrfCatalogAddFileInternal */

static void LdrfGetToken(char* start, char* * nextStart, char* token) {
    TUShort limit = 39;

    while ((*start != '\0') && ((*start == ' ') || (*start == '\t') || (*start == '\n') || (*start == '\r'))) {
        ++start;
    }

    if (*start == '!') {
        while (*start != '\0') {
            ++start;
        }
    } else if ((*start == ';') || (*start == ':')) {
        *(token++) = *(start++);
    } else {
        while ((*start != '\0') && ((*start != ' ') && (*start != '\t') && (*start != '\n') && (*start != ':') && (*start != ';'))) {
            *(token++) = *(start++);

            if (!(--limit)) break;
        }
    }

    * nextStart = start;
    * token = '\0';
} /* end of: LdrfGetToken */

TLdrfErrCodes LdrfScanFormatFile(LPCSTR newFile, PUShort pMajorDataVer, PUByte pMinorDataVer, PUByte pScope, PUByteArray refID) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NOT_FORMAT;
    FILE* pFile = fopen(newFile, F_ACCESS_READ_TEXT);
#define MAX_LINEBUF 255

    char lineBuf[MAX_LINEBUF + 1];
    char token[40];
    char* next = lineBuf;
    TUShort state = 1, refIDIndex = 0, lastState = 0;
    TBool unexp = F;

    static struct {
        char* token;
        TUShort next;
        TBool alternate;
    } expectedTokenTable [] =
        {
        /* Index */ /* Token */ /* Next */ /* Alternate */
        /*   0 */
        { "", 0, F },
        /*   1 */
        { "set", 2, F },
        /*   2 */
        { "program_id", 3, F },
        /*   3 */
        { NULL, 4, F },
        /*   4 */
        { ":", 5, F },
        /*   5 */
        { NULL, 6, F },
        /*   6 */
        { ":", 7, F },
        /*   7 */
        { NULL, 8, F },
        /*   8 */
        { ":", 9, F },
        /*   9 */
        { NULL, 10, F },
        /*  10 */
        { ":", 11, F },
        /*  11 */
        { NULL, 12, F },
        /*  12 */
        { ":", 13, F },
        /*  13 */
        { NULL, 14, F },
        /*  14 */
        { ":", 15, F },
        /*  15 */
        { NULL, 16, F },
        /*  16 */
        { ":", 17, F },
        /*  17 */
        { NULL, 18, F },
        /*  18 */
        { ";", 19, F },
        /*  19 */
        { "set", 20, F },
        /*  20 */
        { "selector", 21, F },
        /*  21 */
        { "standard", 24, T },
        /*  22 */
        { "device_class", 25, T },
        /*  23 */
        { "manufacturer", 29, F },
        /*  24 */
        { ";", 0, F },
        /*  25 */
        { ";", 0, T },
        /*  26 */
        { "and", 27, F },
        /*  27 */
        { "sub_class", 28, F },
        /*  28 */
        { ";", 0, F },
        /*  29 */
        { ";", 0, T },
        /*  30 */
        { "and", 31, F },
        /*  31 */
        { "device_class", 32, F },
        /*  32 */
        { ";", 0, T },
        /*  33 */
        { "and", 34, F },
        /*  34 */
        { "sub_class", 35, F },
        /*  35 */
        { ";", 0, T },
        /*  36 */
        { "and", 37, F },
        /*  37 */
        { "model", 38, F },
        /*  38 */
        { ";", 0, F }
        };

    if (pFile) {
        * token = '\0';

        while (T) {
            while ((*token == '\0')) {
                if (!fgets(lineBuf, MAX_LINEBUF + 1, pFile)) {
                    unexp = T;
                    break;
                } else {
                    LdrfGetToken(lineBuf, &next, token);
                }
            }

            if (expectedTokenTable[state].token == NULL) {
                /* Read token as a hex byte */
                TUByte byte = 0;
                char c = token[0];

                if (('0' <= c) && (c <= '9')) {
                    byte = c - '0';
                } else if (('a' <= c) && (c <= 'f')) {
                    byte = c - 'a' + 10;
                } else if (('A' <= c) && (c <= 'F')) {
                    byte = c - 'A' + 10;
                } else {
                    unexp = T;
                }

                if (!unexp) {
                    c = token[1];

                    if (('0' <= c) && (c <= '9')) {
                        byte = (byte << 4) + (c - '0');
                    } else if (('a' <= c) && (c <= 'f')) {
                        byte = (byte << 4) + (c - 'a' + 10);
                    } else if (('A' <= c) && (c <= 'F')) {
                        byte = (byte << 4) + (c - 'A' + 10);
                    } else {
                        unexp = T;
                    }
                }

                if ((!unexp) && (token[2] != '\0')) {
                    unexp = T;
                }

                if (!unexp) {
                    refID[refIDIndex++] = byte;
                    lastState = state;
                    state = expectedTokenTable[state].next;
                }
            } else {
                do {
                    if (!STRICMP(expectedTokenTable[state].token, token)) {
                        /* Match - go to next state ... */
                        unexp = F;
                        lastState = state;
                        state = expectedTokenTable[state].next;
                    } else {
                        /* Token string mismatch */
                        unexp = T;
                    }
                } while ((unexp) && (expectedTokenTable[state++].alternate));
            }

            if ((!state) || (unexp)) {
                break;
            }
            LdrfGetToken(next, &next, token);
        }

        if (!unexp) {
            errCode = LDRF_ERR_NONE;
            * pMajorDataVer = 0;
            * pMinorDataVer = 0;

            if (lastState == 24) {
                * pScope = 0;
            } else if (lastState == 25) {
                * pScope = 1;
            } else if (lastState == 28) {
                * pScope = 2;
            } else if (lastState == 29) {
                * pScope = 3;
            } else if (lastState == 32) {
                * pScope = 4;
            } else if (lastState == 35) {
                * pScope = 5;
            } else if (lastState == 38) {
                * pScope = 6;
            } else {
                errCode = LDRF_ERR_NOT_FOUND;
            }
        }

        fclose(pFile);
    }

    return errCode;
} /* end of: LdrfScanFormatFile */

LDRF_FN LdrfCatalogGetFile(PLdrfFileInfo pInfo, TLdrfFileType fileType, TUShort index, PUByte pMatchScope, PUByteArray* ppProgID, PUShort pDirIndex,
    PUShort pMajorVersion, PUByte pMinorVersion, PULong pLocale, LPSTR pFileName, PUShort pLength)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);
    TULong* pDir;
    TUShort num;

    if (ppProgID) {
        * ppProgID = NULL;
    }

    if (!errCode) {
        errCode = LdrfSelectDirForFileType(pInfo, fileType, &pDir, &num);
    }

    if (!errCode) {
        if ((index == 0) || (index > num)) {
            errCode = LDRF_ERR_PARAM;
        } else if ((!pDir) || (!(pDir[index]))) {
            errCode = LDRF_ERR_INTERNAL;
        }
    }

    if (!errCode) {
        PUByteArray pProgID = NEW_ARRAY(TUByte, NUM_PROG_ID_BYTES);

        if (!pProgID) {
            errCode = LDRF_ERR_SYS;
        } else {
            errCode = LdrfGetCatalogFileInfo(pInfo, pDir[index], NULL, NULL, NULL, pMatchScope, pProgID, pDirIndex, pMajorVersion, pMinorVersion, pLocale);

            if ((!errCode) && (ppProgID)) {
                * ppProgID = pProgID;
            } else {
                FREE(pProgID);
            }
        }
    }

    if (!errCode) {
        char* p = NULL;
        errCode = LdrfFileReadLString(pInfo, &p);

        if (!errCode) {
            errCode = LdrfTransferStringOut(pFileName, pLength, p);
        }
        FREE(p);
    }

    return errCode;
} /* end of: LdrfCatalogGetFile */

TLdrfErrCodes LdrfGetCatalogFileInfo(PLdrfFileInfo pInfo, TULong startAddress, PUShort pIndex, PUShort pRecSize, TLdrfFileType* pFileType, PUByte pScope,
    PUByteArray refId, PUShort pDirecIndex, PUShort pMajorDataVer, PUByte pMinorDataVer, PULong pLocale)
    {
    /* Local variables */
    TLdrfErrCodes errCode;
    TUShort index, recSize, direcIndex, majorDataVer;
    TUByte fileType, scope, minorDataVer;
    TULong locale;

    if ((pIndex) || (pRecSize)) {
        errCode = LdrfFileSeek(pInfo, startAddress);

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &index);
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &recSize);
        }

        if (!errCode) {
            if (pIndex) {
                * pIndex = index;
            }

            if (pRecSize) {
                * pRecSize = recSize;
            }
        }
    } else {
        errCode = LdrfFileSeek(pInfo, startAddress + OFFSET_CATALOG_FILE_REC_SIZE + sizeof(TUShort));
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &fileType, sizeof(fileType));

        if ((!errCode) && (pFileType)) {
            * pFileType = fileType;
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &scope, sizeof(scope));

        if ((!errCode) && (pScope)) {
            * pScope = scope;
        }
    }

    if (!errCode) {
        TUByte dummy[NUM_PROG_ID_BYTES];
        errCode = LdrfFileReadBytes(pInfo, dummy, NUM_PROG_ID_BYTES);

        if ((!errCode) && (refId)) {
            memcpy(refId, dummy, NUM_PROG_ID_BYTES);
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &direcIndex);

        if ((!errCode) && (pDirecIndex)) {
            * pDirecIndex = direcIndex;
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &majorDataVer);

        if ((!errCode) && (pMajorDataVer)) {
            * pMajorDataVer = majorDataVer;
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &minorDataVer, sizeof(minorDataVer));

        if ((!errCode) && (pMinorDataVer)) {
            * pMinorDataVer = minorDataVer;
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadLong(pInfo, &locale);

        if ((!errCode) && (pLocale)) {
            * pLocale = locale;
        }
    }

    return errCode;
} /* end of: LdrfGetCatalogFileInfo */

TLdrfErrCodes LdrfCatalogFileWrite(PLdrfFileInfo pInfo, TULong startAddress, TUShort index, TLdrfFileType fileType, TUByte scope, PUByteArray refID,
    TUShort direcIndex, TUShort majorDataVer, TUByte minorDataVer, TULong locale, LPCSTR pFileName)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, startAddress);

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &index);
    }

    if (!errCode) {
        /* The next field is the size, but don't know that value yet,
         * so just write index again as a convenient way to advance two bytes */
        errCode = LdrfFileWriteShort(pInfo, &index);
    }

    if (!errCode) {
        TUByte fileTypeByte = (TUByte)fileType;
        errCode = LdrfFileWriteBytes(pInfo, &fileTypeByte, sizeof(fileTypeByte));
    }

    if (!errCode) {
        errCode = LdrfFileWriteBytes(pInfo, &scope, sizeof(scope));
    }

    if (!errCode) {
        errCode = LdrfFileWriteBytes(pInfo, refID, NUM_PROG_ID_BYTES);
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &direcIndex);
    }

    if (!errCode) {
        errCode = LdrfFileWriteShort(pInfo, &majorDataVer);
    }

    if (!errCode) {
        errCode = LdrfFileWriteBytes(pInfo, &minorDataVer, sizeof(minorDataVer));
    }

    if (!errCode) {
        errCode = LdrfFileWriteLong(pInfo, &locale);
    }

    if (!errCode) {
        TUShort len = strlen(pFileName);

        if (len > 255) {
            errCode = LDRF_ERR_PARAM;
        } else {
            errCode = LdrfFileWriteLString(pInfo, pFileName, (TUByte)len);
        }
    }

    if (!errCode) {
        errCode = LdrfUpdateDataSize(pInfo);
    }

    if (!errCode) {
        errCode = LdrfUpdateRecSize(pInfo, startAddress, OFFSET_CATALOG_FILE_REC_SIZE);
    }

    return errCode;
} /* end of: LdrfCatalogFileWrite */

TLdrfErrCodes LdrfCatalogFileCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    TLdrfErrCodes errCode;
    LPSTR pEntry2 = NULL;

    errCode =
        LdrfFileSeek(pInfo,
            entry2Addr
                + (OFFSET_CATALOG_FILE_REC_SIZE + sizeof(TUShort) + 2 * sizeof(TUByte) + NUM_PROG_ID_BYTES + 2 * sizeof(TUShort) + sizeof(TUByte)
                    + sizeof(TULong)));

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pEntry2);
    }

    if (!errCode) {
        * pRC = STRICMP(pEntry1, pEntry2);
    }
    FREE(pEntry2);

    return errCode;
} /* end of: LdrfCatalogFileCompare */

LDRF_FN LdrfCatalogRmvFile(PLdrfFileInfo pInfo, LPCSTR oldFile) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);

    if (!errCode) {
        if (!(pInfo->editAllowed)) {
            errCode = LDRF_ERR_NO_ACCESS;
        } else {
            errCode = LDRF_ERR_UNIMPL;
        }
    }

    return errCode;
} /* end of: LdrfCatalogRmvFile */

LDRF_FN LdrfSearchCatalog(PLdrfFileInfo pInfo, PUByteArray pProgID, TUByte matchScope, TLdrfFileType fileType, TULong locale, LPSTR pFile, PUShort pLength,
    PUShort pMajorVersion, PUByte pMinorVersion)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);
    TULong* pDir;
    TULong index, fileLocale;
    TUShort num, majorDataVer;
    TUByte fileScope, fileRefID[NUM_PROG_ID_BYTES];
    TUByte minorDataVer, firstScope = 6, lastScope = 0, stepScope;
    TBool foundAMatch = F;

    if (!errCode) {
        if ((fileType != LDRF_WILD_CARD) && ((fileType < LDRF_LANG_RESOURCE) || (fileType > LDRF_FORMAT))) {
            errCode = LDRF_ERR_PARAM;
        } else if ((matchScope > NUM_SCOPES) && (matchScope != 0xFF)) {
            errCode = LDRF_ERR_PARAM;
        } else if ((!pProgID) && (matchScope)) {
            errCode = LDRF_ERR_PARAM;
        } else if ((!pFile) || (!pLength)) {
            errCode = LDRF_ERR_PARAM;
        } else if (pInfo->u.cat.stale) {
            errCode = LDRF_ERR_STALE;
        }
    }

    if (!errCode) {
        TLdrfFileType firstFileType = fileType, lastFileType = fileType, stepFileType;

        if (fileType == LDRF_WILD_CARD) {
            firstFileType = LDRF_LANG_RESOURCE;
            lastFileType = LDRF_FORMAT;
        }

        for ( stepFileType = firstFileType; stepFileType <= lastFileType; ++stepFileType ) {
            errCode = LdrfSelectDirForFileType(pInfo, stepFileType, &pDir, &num);

            if (matchScope != 0xFF) {
                firstScope = lastScope = matchScope;
            }

            for ( stepScope = firstScope; ((stepScope <= firstScope) && (stepScope >= lastScope)); --stepScope ) {
                for ( index = 1; ((index <= num) && (!foundAMatch)); ++index ) {
                    errCode =
                        LdrfGetCatalogFileInfo(pInfo, pDir[index], NULL, NULL, NULL, &fileScope, fileRefID, NULL, &majorDataVer, &minorDataVer, &fileLocale);

                    if (errCode) break;

                    foundAMatch = T;
                    /* Use fileType, not stepFileType, in line below, because the locale test
                     * is only done if we were looking for a language file to begin with. */
                    if ((fileType == LDRF_LANG_RESOURCE) && (fileLocale != locale)) {
                        foundAMatch = F;
                    } else if ((stepScope != fileScope) || (LdrfMatchProgID(fileScope, fileRefID, pProgID))) {
                        foundAMatch = F;
                    }
                }

                if ((errCode) || (foundAMatch)) break;
            }

            if ((errCode) || (foundAMatch)) break;
        }
    }

    if (!errCode) {
        if (!foundAMatch) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else {
            char* p = NULL;
            errCode = LdrfFileReadLString(pInfo, &p);

            if (!errCode) {
                errCode = LdrfTransferStringOut(pFile, pLength, p);
            }
            FREE(p);

            if (!errCode) {
                if (pMajorVersion) {
                    * pMajorVersion = majorDataVer;
                }

                if (pMinorVersion) {
                    * pMinorVersion = minorDataVer;
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfSearchCatalog */

LDRF_FN LdrfMatchProgID(TUByte fileScope, PUByteArray pFileRefID, PUByteArray pProgID) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NOT_FOUND;

    if (fileScope == 0) {
        if ((!(pFileRefID[0])) && (!(pFileRefID[1])) && (!(pFileRefID[2])) && (!(pFileRefID[3])) && (!(pFileRefID[4])) && (!(pFileRefID[5]))
            && (!(pFileRefID[6])) && (!(pFileRefID[7])))
            {
            errCode = LDRF_ERR_NONE;
        }
    } else if (fileScope == 6) {
        if (((((pFileRefID[0] & 0x80) == 0) && ((pFileRefID[0] == pProgID[0])))
            || (((pFileRefID[0] & 0xE0) == 0x80) && ((pProgID[0] & 0xE0) == 0x80) && ((pFileRefID[0] & 0x0F) == (pProgID[0] & 0x0F))))
            && (!memcmp(&(pFileRefID[1]), &(pProgID[1]), 7)))
            {
            errCode = LDRF_ERR_NONE;
        }
    } else if (fileScope > 5) {
        errCode = LDRF_ERR_PARAM;
    } else if (((pFileRefID[0] & 0xE0) == 0x80) && ((pProgID[0] & 0xE0) == 0x80)) {
        switch (fileScope) {
            case 1:
                if (!memcmp(&(pFileRefID[3]), &(pProgID[3]), 2)) {
                    errCode = LDRF_ERR_NONE;
                }
                break;

            case 2:
                if (!memcmp(&(pFileRefID[3]), &(pProgID[3]), 4)) {
                    errCode = LDRF_ERR_NONE;
                }
                break;

            case 3:
                if (((pFileRefID[0] & 0x0F) == (pProgID[0] & 0x0F)) && (!memcmp(&(pFileRefID[1]), &(pProgID[1]), 2))) {
                    errCode = LDRF_ERR_NONE;
                }
                break;

            case 4:
                if (((pFileRefID[0] & 0x0F) == (pProgID[0] & 0x0F)) && (!memcmp(&(pFileRefID[1]), &(pProgID[1]), 4))) {
                    errCode = LDRF_ERR_NONE;
                }
                break;

            case 5:
                if (((pFileRefID[0] & 0x0F) == (pProgID[0] & 0x0F)) && (!memcmp(&(pFileRefID[1]), &(pProgID[1]), 6))) {
                    errCode = LDRF_ERR_NONE;
                }
                break;
        } /* end of: switch */
    }

    return errCode;
} /* end of: LdrfMatchProgID */

LDRF_FN LdrfCatalogDependencyCode(PLdrfFileInfo pInfo, PUByteArray pProgID, PULong pDepCode) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_CATALOG);
    TUShort fnDepCode = 0, statsDepCode = 0;
    TUShort numLangFiles, numTypeFiles, numFPTFiles, i;
    struct _stat statbuf;

    if (!errCode) {
        TBool stale;
        errCode = LdrfGetCatalogInfo(pInfo, &stale, NULL, &numLangFiles, &numTypeFiles, &numFPTFiles, NULL);

        if ((!errCode) && (stale)) {
            errCode = LDRF_ERR_STALE;
        }
    }

    for ( i = 1; ((!errCode) && (i <= numLangFiles)); ++i ) {
        TUByte matchScope, minorVersion;
        PUByteArray pFileProgID = NULL;
        TUShort majorVersion, fnLength;
        TULong locale;
        char fileName[260];
        fnLength = 258;
        errCode = LdrfCatalogGetFile(pInfo, LDRF_LANG_RESOURCE, i, &matchScope, &pFileProgID, NULL, &majorVersion, &minorVersion, &locale, fileName, &fnLength);

        if (!errCode) {
            errCode = LdrfMatchProgID(matchScope, pFileProgID, pProgID);

            if (!errCode) {
                /* Add this file's information into the "dependency code" */
                fnDepCode = LdrfComputeCRC(fileName, fnLength, fnDepCode);
                statsDepCode = LdrfComputeCRC((PUByteArray) &majorVersion, sizeof(majorVersion), statsDepCode);
                statsDepCode = LdrfComputeCRC((PUByteArray) &minorVersion, sizeof(minorVersion), statsDepCode);

                if (_stat(fileName, &statbuf)) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_size), sizeof(statbuf.st_size), statsDepCode);
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_mtime), sizeof(statbuf.st_mtime), statsDepCode);
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_ctime), sizeof(statbuf.st_ctime), statsDepCode);
                }
            } else if (errCode == LDRF_ERR_NOT_FOUND) {
                errCode = LDRF_ERR_NONE;
            }
        }

        if (pFileProgID) {
            TLdrfErrCodes ec2 = LdrfFreeByteArray(pFileProgID);

            if (!errCode) {
                errCode = ec2;
            }
        }
    }

    for ( i = 1; ((!errCode) && (i <= numTypeFiles)); ++i ) {
        TUByte matchScope, minorVersion;
        PUByteArray pFileProgID = NULL;
        TUShort majorVersion, fnLength;
        TULong locale;
        char fileName[260];
        fnLength = 258;
        errCode = LdrfCatalogGetFile(pInfo, LDRF_TYPE, i, &matchScope, &pFileProgID, NULL, &majorVersion, &minorVersion, &locale, fileName, &fnLength);

        if (!errCode) {
            errCode = LdrfMatchProgID(matchScope, pFileProgID, pProgID);

            if (!errCode) {
                /* Add this file's information into the "dependency code" */
                fnDepCode = LdrfComputeCRC(fileName, fnLength, fnDepCode);
                statsDepCode = LdrfComputeCRC((PUByteArray) &majorVersion, sizeof(majorVersion), statsDepCode);
                statsDepCode = LdrfComputeCRC((PUByteArray) &minorVersion, sizeof(minorVersion), statsDepCode);

                if (_stat(fileName, &statbuf)) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_size), sizeof(statbuf.st_size), statsDepCode);
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_mtime), sizeof(statbuf.st_mtime), statsDepCode);
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_ctime), sizeof(statbuf.st_ctime), statsDepCode);
                }
            } else if (errCode == LDRF_ERR_NOT_FOUND) {
                errCode = LDRF_ERR_NONE;
            }
        }

        if (pFileProgID) {
            TLdrfErrCodes ec2 = LdrfFreeByteArray(pFileProgID);

            if (!errCode) {
                errCode = ec2;
            }
        }
    }

    for ( i = 1; ((!errCode) && (i <= numFPTFiles)); ++i ) {
        TUByte matchScope, minorVersion;
        PUByteArray pFileProgID = NULL;
        TUShort majorVersion, fnLength;
        TULong locale;
        char fileName[260];
        fnLength = 258;
        errCode = LdrfCatalogGetFile(pInfo, LDRF_FPT, i, &matchScope, &pFileProgID, NULL, &majorVersion, &minorVersion, &locale, fileName, &fnLength);

        if (!errCode) {
            errCode = LdrfMatchProgID(matchScope, pFileProgID, pProgID);

            if (!errCode) {
                /* Add this file's information into the "dependency code" */
                fnDepCode = LdrfComputeCRC(fileName, fnLength, fnDepCode);
                statsDepCode = LdrfComputeCRC((PUByteArray) &majorVersion, sizeof(majorVersion), statsDepCode);
                statsDepCode = LdrfComputeCRC((PUByteArray) &minorVersion, sizeof(minorVersion), statsDepCode);

                if (_stat(fileName, &statbuf)) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_size), sizeof(statbuf.st_size), statsDepCode);
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_mtime), sizeof(statbuf.st_mtime), statsDepCode);
                    statsDepCode = LdrfComputeCRC((PUByteArray) &(statbuf.st_ctime), sizeof(statbuf.st_ctime), statsDepCode);
                }
            } else if (errCode == LDRF_ERR_NOT_FOUND) {
                errCode = LDRF_ERR_NONE;
            }
        }

        if (pFileProgID) {
            TLdrfErrCodes ec2 = LdrfFreeByteArray(pFileProgID);

            if (!errCode) {
                errCode = ec2;
            }
        }
    }

    if (!errCode) {
        if (pDepCode) {
            * pDepCode = (fnDepCode << 16) + statsDepCode;
        }
    }

    return errCode;
} /* end of: LdrfCatalogDependencyCode */
