/*
 * DRFFPT.C - This is the Resource File API file responsible for functional profile access methods
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif

TLdrfErrCodes LdrfReadFPTFileHeader(PLdrfFileInfo pInfo, TBool checkCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfReadFileHeader(pInfo, checkCRC);
    TUShort count;

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, pInfo->refID, sizeof(pInfo->refID));
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &(pInfo->scope), sizeof(pInfo->scope));
    }

    for ( count = 0; ((count < NUM_SCOPES) && (!errCode)); ++count ) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.fpt.resDeps[count]));
    }

    for ( count = 0; ((count < NUM_SCOPES) && (!errCode)); ++count ) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.fpt.typDeps[count]));
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.fpt.numFPTs));
        pInfo->u.fpt.capacityFPT = pInfo->u.fpt.numFPTs;
    }

    if ((!errCode) && (pInfo->majorFmtVer > 3)) {
        // Format 4 & later of FPT file also has a "numEmpty" count
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.fpt.numEmptyFPTs));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.fpt.numFPTs, pInfo->u.fpt.numFPTs, &(pInfo->u.fpt.pDirFPTIndex));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.fpt.numFPTs, pInfo->u.fpt.numFPTs - pInfo->u.fpt.numEmptyFPTs, &(pInfo->u.fpt.pDirFPTName));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.fpt.numFPTs, pInfo->u.fpt.numFPTs - pInfo->u.fpt.numEmptyFPTs, &(pInfo->u.fpt.pDirFPTKey));
    }

    if (!errCode) {
        pInfo->u.fpt.fptSetInProgress = F;
        pInfo->u.fpt.needsFPTFileInfo = F;
    }

    return errCode;
} /* end of: LdrfReadFPTFileHeader */

TLdrfErrCodes LdrfFreeFPTFileInfo(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    FREE(pInfo->u.fpt.pDirFPTIndex);
    FREE(pInfo->u.fpt.pDirFPTName);
    FREE(pInfo->u.fpt.pDirFPTKey);

    return errCode;
} /* end of: LdrfFreeFPTFileInfo */

LDRF_FN LdrfGetFPTFileInfo(PLdrfFileInfo pInfo, PUShort pResDep0, PUShort pResDep1, PUShort pResDep2, PUShort pResDep3, PUShort pResDep4, PUShort pResDep5,
    PUShort pResDep6, PUShort pTypDep0, PUShort pTypDep1, PUShort pTypDep2, PUShort pTypDep3, PUShort pTypDep4, PUShort pTypDep5, PUShort pTypDep6,
    PUShort pNumFPTs)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);

    if (!errCode) {
        if (pResDep0) {
            * pResDep0 = pInfo->u.fpt.resDeps[0];
        }

        if (pResDep1) {
            * pResDep1 = pInfo->u.fpt.resDeps[1];
        }

        if (pResDep2) {
            * pResDep2 = pInfo->u.fpt.resDeps[2];
        }

        if (pResDep3) {
            * pResDep3 = pInfo->u.fpt.resDeps[3];
        }

        if (pResDep4) {
            * pResDep4 = pInfo->u.fpt.resDeps[4];
        }

        if (pResDep5) {
            * pResDep5 = pInfo->u.fpt.resDeps[5];
        }

        if (pResDep6) {
            * pResDep6 = pInfo->u.fpt.resDeps[6];
        }

        if (pTypDep0) {
            * pTypDep0 = pInfo->u.fpt.typDeps[0];
        }

        if (pTypDep1) {
            * pTypDep1 = pInfo->u.fpt.typDeps[1];
        }

        if (pTypDep2) {
            * pTypDep2 = pInfo->u.fpt.typDeps[2];
        }

        if (pTypDep3) {
            * pTypDep3 = pInfo->u.fpt.typDeps[3];
        }

        if (pTypDep4) {
            * pTypDep4 = pInfo->u.fpt.typDeps[4];
        }

        if (pTypDep5) {
            * pTypDep5 = pInfo->u.fpt.typDeps[5];
        }

        if (pTypDep6) {
            * pTypDep6 = pInfo->u.fpt.typDeps[6];
        }

        if (pNumFPTs) {
            * pNumFPTs = pInfo->u.fpt.numFPTs;
        }
    }

    return errCode;
} /* end of: LdrfGetFPTFileInfo */

LDRF_FN LdrfGetFPT(PLdrfFileInfo pInfo, TUShort index, PUShort pKey, LPSTR pName, PUShort pLength, PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope,
    PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV)
{
    return LdrfGetFPTEx(pInfo, index, pKey, pName, pLength, pResNmScope, pResNmIndex, pResCmtScope,
                        pResCmtIndex, pManNVs, pOptNVs, pManCPs, pOptCPs, pPrincipalNV, NULL);
}

LDRF_FN LdrfGetFPTEx(PLdrfFileInfo pInfo, TUShort index, PUShort pKey, LPSTR pName, PUShort pLength, PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope,
    PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV, PUShort pFlags)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFPT(pInfo, index);

    if ((!errCode) || (errCode == LDRF_ERR_EMPTY_RECORD)) {
        if (errCode == LDRF_ERR_EMPTY_RECORD) {
            if (!(pInfo->emptyRecordAware)) {
                char nameBuf[20];
                strcpy(nameBuf, (pInfo->scope) ? "UFPT" : "SFPT");
                strcpy(nameBuf + 4, "empty");
                _itoa(index, nameBuf + 9, 10);
                strcat(nameBuf, "~");

                if (pKey) {
                    * pKey = 65535;
                }

                if ((pName) && (pLength)) {
                    strncpy(pName, nameBuf, (* pLength)-1);
                    pName[(* pLength)-1] = '\0';
                    * pLength = strlen(pName);
                }

                if (pResNmScope) {
                    * pResNmScope = 0;
                }

                if (pResNmIndex) {
                    * pResNmIndex = 0;
                }

                if (pResCmtScope) {
                    * pResCmtScope = 0;
                }

                if (pResCmtIndex) {
                    * pResCmtIndex = 0;
                }

                if (pManNVs) {
                    * pManNVs = 0;
                }

                if (pOptNVs) {
                    * pOptNVs = 0;
                }

                if (pManCPs) {
                    * pManCPs = 0;
                }

                if (pOptCPs) {
                    * pOptCPs = 0;
                }

                if (pPrincipalNV) {
                    * pPrincipalNV = 0;
                }
                if (pFlags) {
                    * pFlags = 0;
                }
                errCode = LDRF_ERR_NONE;
            }
        } else {
            errCode =
                LdrfFPTRead(pInfo, pInfo->u.fpt.pDirFPTIndex[index], NULL, pKey, pName, pLength, pResNmScope, pResNmIndex, pResCmtScope, pResCmtIndex, pManNVs,
                    pOptNVs, pManCPs, pOptCPs, pPrincipalNV, pFlags);
        }
    }

    return errCode;
} /* end of: LdrfGetFPT */

LDRF_FN LdrfValidateFPT(PLdrfFileInfo pInfo, TUShort index) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);

    if (!errCode) {
        if (index < 1) {
            errCode = LDRF_ERR_PARAM;
        } else if (index > pInfo->u.fpt.numFPTs) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else if (!(pInfo->u.fpt.pDirFPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.fpt.pDirFPTIndex[index])) {
            errCode = LDRF_ERR_EMPTY_RECORD;
        }
    }

    return errCode;
} /* end of: LdrfValidateFPT */

LDRF_FN LdrfFindEmptyFPT(PLdrfFileInfo pInfo, PUShort pIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);
    TUShort index = 0;

    if (!errCode) {
        if (!(pInfo->u.fpt.numFPTs)) {
            index = 1;
        } else if (!(pInfo->u.fpt.pDirFPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            errCode = LDRF_ERR_NOT_FOUND;

            for ( index = 1; index <= pInfo->u.fpt.numFPTs; ++index ) {
                if (!(pInfo->u.fpt.pDirFPTIndex[index])) {
                    errCode = LDRF_ERR_NONE;
                    break;
                }
            }

            if (errCode) {
                if (pInfo->u.fpt.numFPTs < 65535UL) {
                    index = pInfo->u.fpt.numFPTs + 1;
                    errCode = LDRF_ERR_NONE;
                }
            }
        }
    }

    if (pIndex) {
        * pIndex = index;
    }

    return errCode;
} /* end of: LdrfFindEmptyFPT */

LDRF_FN LdrfGetFPTByName(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PUShort pKey, PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope,
    PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV)
{
    return LdrfGetFPTByNameEx(pInfo, name, pIndex, pKey, pResNmScope, pResNmIndex, pResCmtScope,
                              pResCmtIndex, pManNVs, pOptNVs, pManCPs, pOptCPs, pPrincipalNV, NULL);
}

LDRF_FN LdrfGetFPTByNameEx(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PUShort pKey, PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope,
    PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV,
    PUShort pFlags)
{
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);
    TULong entry;

    if (!errCode) {
        errCode = LdrfFindInSortedDir(pInfo, pInfo->u.fpt.pDirFPTName, pInfo->u.fpt.numFPTs - pInfo->u.fpt.numEmptyFPTs, name, &entry, LdrfFPTNameCompare);
    }

    if (!errCode) {
        errCode =
            LdrfFPTRead(pInfo, pInfo->u.fpt.pDirFPTName[entry], pIndex, pKey, NULL, NULL, pResNmScope, pResNmIndex, pResCmtScope, pResCmtIndex, pManNVs,
                pOptNVs, pManCPs, pOptCPs, pPrincipalNV, pFlags);
    }

    return errCode;
} /* end of: LdrfGetFPTByName */

LDRF_FN LdrfGetFPTByKey(PLdrfFileInfo pInfo, TUShort key, PUShort pIndex, LPSTR pName, PUShort pLength, PUByte pResNmScope, PULong pResNmIndex,
    PUByte pResCmtScope, PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV)
{
    return LdrfGetFPTByKeyEx(pInfo, key, pIndex, pName, pLength, pResNmScope, pResNmIndex,
                             pResCmtScope, pResCmtIndex, pManNVs, pOptNVs, pManCPs, pOptCPs, pPrincipalNV, NULL);
}

LDRF_FN LdrfGetFPTByKeyEx(PLdrfFileInfo pInfo, TUShort key, PUShort pIndex, LPSTR pName, PUShort pLength, PUByte pResNmScope, PULong pResNmIndex,
    PUByte pResCmtScope, PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV, PUShort pFlags)
{
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);
    TULong entry, dummy;

    if (!errCode) {
        dummy = key;
        errCode =
            LdrfFindInSortedDir(pInfo, pInfo->u.fpt.pDirFPTKey, pInfo->u.fpt.numFPTs - pInfo->u.fpt.numEmptyFPTs, (LPCSTR)dummy, &entry, LdrfFPTKeyCompare);
    }

    if (!errCode) {
        errCode =
            LdrfFPTRead(pInfo, pInfo->u.fpt.pDirFPTKey[entry], pIndex, NULL, pName, pLength, pResNmScope, pResNmIndex, pResCmtScope, pResCmtIndex, pManNVs,
                pOptNVs, pManCPs, pOptCPs, pPrincipalNV, pFlags);
    }

    return errCode;
} /* end of: LdrfGetFPTByKey */

LDRF_FN LdrfGetFPTObsolete(PLdrfFileInfo pInfo, TUShort index, PBoolByte pObsolete) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);

    if (!errCode) {
        if (((index == 0) || (index > pInfo->u.fpt.numFPTs)) || (!pObsolete)) {
            errCode = LDRF_ERR_PARAM;
        } else if (pInfo->majorFmtVer < 2) {
            * pObsolete = F;
        } else if (!(pInfo->u.fpt.pDirFPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.fpt.pDirFPTIndex[index])) {
            if (pInfo->emptyRecordAware) {
                errCode = LDRF_ERR_EMPTY_RECORD;
            } else {
                * pObsolete = T;
            }
        } else {
            /* Skip index, rec-size, and key */
            errCode = LdrfFileSeek(pInfo, pInfo->u.fpt.pDirFPTIndex[index] + OFFSET_FPT_FILE_REC_SIZE + sizeof(TUShort) /* sizeof rec-size */
            + sizeof(TUShort));                                                                                         /* sizeof key */

            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, pObsolete, sizeof(*pObsolete));
            }
        }
    }

    return errCode;
} /* end of: LdrfGetFPTObsolete */

LDRF_FN LdrfGetFPTInherit(PLdrfFileInfo pInfo, TUShort index, PBoolByte pInherit) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);

    if (!errCode) {
        if (((index == 0) || (index > pInfo->u.fpt.numFPTs)) || (!pInherit)) {
            errCode = LDRF_ERR_PARAM;
        } else if (pInfo->majorFmtVer < 3) {
            * pInherit = F;
        } else if (!(pInfo->u.fpt.pDirFPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.fpt.pDirFPTIndex[index])) {
            if (pInfo->emptyRecordAware) {
                errCode = LDRF_ERR_EMPTY_RECORD;
            } else {
                * pInherit = T;
            }
        } else {
            /* Skip index, rec-size, key, and obsolete */
            errCode = LdrfFileSeek(pInfo, pInfo->u.fpt.pDirFPTIndex[index] + OFFSET_FPT_FILE_REC_SIZE + sizeof(TUShort) /* sizeof rec-size */
            + sizeof(TUShort)                                                                                           /* sizeof key */
            + sizeof(TBoolByte));                                                                                       /* sizeof obsolete */

            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, pInherit, sizeof(*pInherit));
            }
        }
    }

    return errCode;
} /* end of: LdrfGetFPTInherit */

TLdrfErrCodes LdrfFPTRead(PLdrfFileInfo pInfo, TULong startAddress, PUShort pIndex, PUShort pKey, LPSTR pName, PUShort pLength, PUByte pResNmScope,
    PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV, PUShort pFlags)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    LPSTR pFPTName = NULL;

    if ((!errCode) && (pIndex)) {
        errCode = LdrfFileSeek(pInfo, startAddress);

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, pIndex);
        }
    }

    if (!errCode) {
        /* Skip rec-size */
        errCode = LdrfFileSeek(pInfo, startAddress + OFFSET_FPT_FILE_REC_SIZE + sizeof(TUShort));
    }

    if (!errCode) {
        /* Read the key */
        TUShort key;
        errCode = LdrfFileReadShort(pInfo, &key);

        if ((!errCode) && (pKey)) {
            * pKey = key;
        }
    }

    if (!errCode) {
        if (pInfo->majorFmtVer > 1) {
            /* Major Fmt Ver 2 or higher permits the FPT to be marked as
             * obsolete.  This must be obtained from a separate function.
             * In this read we just skip over the flag if present. */
            TBoolByte obsoleteMark;
            errCode = LdrfFileReadBytes(pInfo, &obsoleteMark, sizeof(obsoleteMark));
        }
    }

    if (!errCode) {
        if (pInfo->majorFmtVer > 2) {
            /* Major Fmt Ver 3 or higher permits the FPT to be marked as
             * inheriting.  This must be obtained from a separate function.
             * In this read we just skip over the flag if present. */
            TBoolByte inheritMark;
            errCode = LdrfFileReadBytes(pInfo, &inheritMark, sizeof(inheritMark));
        }
    }

    if (!errCode) {
        /* Read the name */
        errCode = LdrfFileReadLString(pInfo, &pFPTName);

        if (!errCode) {
            errCode = LdrfTransferStringOut(pName, pLength, pFPTName);
        }
        FREE(pFPTName);
    }

    if (!errCode) {
        /* Read the resource info */
        TUByte resNmScope, resCmtScope;
        TULong resNmIndex, resCmtIndex;
        TUShort flags = 0;
        errCode = LdrfFileReadBytes(pInfo, &resNmScope, sizeof(resNmScope));

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &resNmIndex);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &resCmtScope, sizeof(resCmtScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &resCmtIndex);
        }

        if (!errCode && (pInfo->majorFmtVer >= 5)) {
            errCode = LdrfFileReadShort(pInfo, &flags);
        }
        
        if (!errCode) {
            if (pResNmScope) {
                * pResNmScope = resNmScope;
            }

            if (pResNmIndex) {
                * pResNmIndex = resNmIndex;
            }

            if (pResCmtScope) {
                * pResCmtScope = resCmtScope;
            }

            if (pResCmtIndex) {
                * pResCmtIndex = resCmtIndex;
            }

            if (pFlags) {
                * pFlags = flags;
            }
        }
    }

    if (!errCode) {
        /* Read the counts and principal NV */
        TUByte manNVs, optNVs, manCPs, optCPs, principalNV;
        errCode = LdrfFileReadBytes(pInfo, &manNVs, sizeof(manNVs));

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &optNVs, sizeof(optNVs));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &manCPs, sizeof(manCPs));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &optCPs, sizeof(optCPs));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &principalNV, sizeof(principalNV));
        }

        if (!errCode) {
            if (pManNVs) {
                * pManNVs = manNVs;
            }

            if (pOptNVs) {
                * pOptNVs = optNVs;
            }

            if (pManCPs) {
                * pManCPs = manCPs;
            }

            if (pOptCPs) {
                * pOptCPs = optCPs;
            }

            if (pPrincipalNV) {
                * pPrincipalNV = principalNV;
            }
        }
    }

    return errCode;
} /* end of: LdrfFPTRead */

TLdrfErrCodes LdrfFPTNameCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    TLdrfErrCodes errCode;
    LPSTR pEntry2 = NULL;

    unsigned int amountToSkip = sizeof(TUShort) /* skip rec-size */
    + sizeof(TUShort);                          /* skip key */

    if (pInfo->majorFmtVer > 1) {
        /* Major Fmt Ver 2 or higher permits the FPT to be
         * marked as obsolete. In those format-versions, the obsolete mark
         * must be skipped (it comes after the rec-size & the key,
         * but before the FPT name. */
        amountToSkip += sizeof(TBoolByte);
    }

    if (pInfo->majorFmtVer > 2) {
        /* Major Fmt Ver 3 or higher permits the FPT to be
         * marked as inheriting. In those format-versions, the inherit mark
         * must be skipped (it comes after the rec-size, key, and obsolete mark,
         * but before the FPT name. */
        amountToSkip += sizeof(TBoolByte);
    }

    /* Skip to the name string */
    errCode = LdrfFileSeek(pInfo, entry2Addr + OFFSET_FPT_FILE_REC_SIZE + amountToSkip);

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pEntry2);
    }

    if (!errCode) {
        * pRC = strcmp(pEntry1, pEntry2);
    }
    FREE(pEntry2);

    return errCode;
} /* end of: LdrfFPTNameCompare */

TLdrfErrCodes LdrfFPTKeyCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, /* Name string (skip index, recsize) */
    entry2Addr + OFFSET_FPT_FILE_REC_SIZE + sizeof(TUShort));
    TULong dummy = (TULong)pEntry1;
    TUShort keyValue1, keyValue2;

    keyValue1 = (TUShort)dummy;

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &keyValue2);
    }

    if (!errCode) {
        * pRC = keyValue1 - keyValue2;
    }

    return errCode;
} /* end of: LdrfFPTKeyCompare */

TLdrfErrCodes LdrfGetFPTCommonSetup(PLdrfFileInfo pInfo, TUShort FPTIndex, PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs, PUByte pPrincipalNV)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_FPT);
    TUByte manNVs, optNVs, manCPs, optCPs, principalNV;

    if (!errCode) {
        if ((FPTIndex == 0) || (FPTIndex > pInfo->u.fpt.numFPTs)) {
            errCode = LDRF_ERR_PARAM;
        } else if ((!(pInfo->u.fpt.pDirFPTIndex)) || (!(pInfo->u.fpt.pDirFPTIndex[FPTIndex]))) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            errCode =
                LdrfFPTRead(pInfo, pInfo->u.fpt.pDirFPTIndex[FPTIndex], NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &manNVs, &optNVs, &manCPs, &optCPs,
                    &principalNV, NULL);
        }
    }

    if (!errCode) {
        if (pManNVs) {
            * pManNVs = manNVs;
        }

        if (pOptNVs) {
            * pOptNVs = optNVs;
        }

        if (pManCPs) {
            * pManCPs = manCPs;
        }

        if (pOptCPs) {
            * pOptCPs = optCPs;
        }

        if (pPrincipalNV) {
            * pPrincipalNV = principalNV;
        }
    }

    return errCode;
} /* end of: LdrfGetFPTCommonSetup */

TLdrfErrCodes LdrfGetFPTNVCommonSetup(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVIndex) {
    /* Local variables */
    TUShort totalNVs;
    TUByte manNVs, optNVs;
    TLdrfErrCodes errCode = LdrfGetFPTCommonSetup(pInfo, FPTIndex, &manNVs, &optNVs, NULL, NULL, NULL);

    if (!errCode) {
        totalNVs = manNVs + optNVs;

        if ((NVIndex == 0) || (NVIndex > totalNVs)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            /* Skip through the FPTNV sub-records, until we get to the one
             * requested by NVIndex. */
            TUShort index, recSize;

            for ( index = 1; index < NVIndex; ++index ) {
                errCode = LdrfFileReadShort(pInfo, &recSize);

                if (errCode) break;

                errCode = LdrfFileWhere(pInfo);

                if (errCode) break;
                errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + recSize);

                if (errCode) break;
            }
        }
    }

    return errCode;
} /* end of: LdrfGetFPTNVCommonSetup */

LDRF_FN LdrfGetFPTNV(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVIndex, LPSTR pNVName, PUShort pLength, PBool pMandatory, PUByte pResNmScope,
    PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pNVTScope, PUShort pNVTIndex, PUByte pDirPollServ, PUShort pByteArrayLen,
    PUByteArray* ppValMin, PUByteArray* ppValMax)
{
    return LdrfGetFPTNVEx(pInfo, FPTIndex, NVIndex, pNVName, pLength, pMandatory, pResNmScope, pResNmIndex, 
        pResCmtScope, pResCmtIndex, pNVTScope, pNVTIndex, pDirPollServ, pByteArrayLen, 
        ppValMin, ppValMax, NULL);
}

LDRF_FN LdrfGetFPTNVEx(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVIndex, LPSTR pNVName, PUShort pLength, PBool pMandatory, PUByte pResNmScope,
    PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pNVTScope, PUShort pNVTIndex, PUByte pDirPollServ, PUShort pByteArrayLen,
    PUByteArray* ppValMin, PUByteArray* ppValMax, PUByteArray* ppValInvalid)
{
    return LdrfGetFPTNVEx2(pInfo, FPTIndex, NVIndex, pNVName, pLength, pMandatory, pResNmScope, pResNmIndex, 
        pResCmtScope, pResCmtIndex, pNVTScope, pNVTIndex, pDirPollServ, pByteArrayLen, 
        ppValMin, ppValMax, ppValInvalid, NULL);
}

LDRF_FN LdrfGetFPTNVEx2(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVIndex, LPSTR pNVName, PUShort pLength, PBool pMandatory, PUByte pResNmScope,
    PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pNVTScope, PUShort pNVTIndex, PUByte pDirPollServ, PUShort pByteArrayLen,
    PUByteArray* ppValMin, PUByteArray* ppValMax, PUByteArray* ppValInvalid, PUShort pFlags)
{
    /* Local variables */
    TLdrfErrCodes errCode = LdrfGetFPTNVCommonSetup(pInfo, FPTIndex, NVIndex);
    LPSTR pNVNameInternal = NULL;
    TUByte NVTScope;
	TUShort NVTIndex;

    if (!errCode) {
        TUShort recSize;
        /* Read record size (but just throw it away) */
        errCode = LdrfFileReadShort(pInfo, &recSize);
    }

    if ((!errCode) && (pInfo->majorFmtVer > 2)) {
        TUShort NVMemberNumber;
        /* For FPT file format-version 3 and above, the FPTNV sub-record
         * contains an explicit member number that may be different
         * than the index.  It follows the recSize, but precedes the name.
         * We are going to just throw it away, because this interface
         * does not use it.  For performance reasons, someday we
         * may cache it for reading later. */
        errCode = LdrfFileReadShort(pInfo, &NVMemberNumber);
    }

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pNVNameInternal);

        if (!errCode) {
            errCode = LdrfTransferStringOut(pNVName, pLength, pNVNameInternal);
        }
        FREE(pNVNameInternal);
    }

    if (!errCode) {
        TUByte mandatory;
        TUByte resNmScope, resCmtScope;
        TULong resNmIndex, resCmtIndex;
        TUShort flags = 0;
        errCode = LdrfFileReadBytes(pInfo, &mandatory, sizeof(mandatory));

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &resNmScope, sizeof(resNmScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &resNmIndex);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &resCmtScope, sizeof(resCmtScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &resCmtIndex);
        }
        if (!errCode && (pInfo->majorFmtVer >= 5)) {
            errCode = LdrfFileReadShort(pInfo, &flags);
        }

        if (!errCode) {
            if (pMandatory) {
                * pMandatory = (TBool)mandatory;
            }

            if (pResNmScope) {
                * pResNmScope = resNmScope;
            }

            if (pResNmIndex) {
                * pResNmIndex = resNmIndex;
            }

            if (pResCmtScope) {
                * pResCmtScope = resCmtScope;
            }

            if (pResCmtIndex) {
                * pResCmtIndex = resCmtIndex;
            }
            
            if (pFlags) {
                * pFlags = flags;
            }
        }
    }

    if (!errCode) {
        TUByte dirPollServ, byteArrayLen;
        PUByteArray pMinArray = NULL, pMaxArray = NULL, pInvalidArray = NULL;
        errCode = LdrfFileReadBytes(pInfo, &NVTScope, sizeof(NVTScope));

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &NVTIndex);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &dirPollServ, sizeof(dirPollServ));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &byteArrayLen, sizeof(byteArrayLen));
        }

        if (!errCode) {
            if (pNVTScope) {
                * pNVTScope = NVTScope;
            }

            if (pNVTIndex) {
                * pNVTIndex = NVTIndex;
            }

            if (pDirPollServ) {
                * pDirPollServ = dirPollServ &(~(FPTNV_MIN_VAL_RG + FPTNV_MAX_VAL_RG + FPTNV_INVALID_VAL_RG));
            }

            if (pByteArrayLen) {
                * pByteArrayLen = byteArrayLen;
            }
        }

        if (!errCode && byteArrayLen) {

            // 27-Jan-2009  Bernd  2.31.16  
            // Temporaryfix (or permanent precaution): some types manage to report byteArrayLen of zero (correct for those cases that we saw
            // in debugging), yet manage to raise the FTPNV_INVALID_VAL_RG flag. The subsequent call to LdrfFielReadBytes fails with LDRF_ERR_PARAM,
            // as pInvalidArray is NULL as a result of the zero-length byteArray. 
            //
            // Once the root cause is understood (how come this flag is raised in these cases), this comment should be updated accordingly. 

            if (dirPollServ & FPTNV_MIN_VAL_RG) {
                pMinArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pMinArray, byteArrayLen);
            }

            if ((!errCode) && (dirPollServ &FPTNV_MAX_VAL_RG)) {
                pMaxArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pMaxArray, byteArrayLen);
            }

            if ((!errCode) && (dirPollServ & FPTNV_INVALID_VAL_RG)) {
                pInvalidArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pInvalidArray, byteArrayLen);
            }

            if (!errCode) {
                if (ppValMin) {
                    * ppValMin = pMinArray;
                    pMinArray = NULL;
                }

                if (ppValMax) {
                    * ppValMax = pMaxArray;
                    pMaxArray = NULL;
                }
                if (ppValInvalid) {
                    * ppValInvalid = pInvalidArray;
                    pInvalidArray = NULL;
                }
            }
            FREE(pMinArray);
            FREE(pMaxArray);
            FREE(pInvalidArray);
        }
    }

	if (!errCode
	&& !pInfo->extendedSNVTIDaware
	&& NVTScope == 0
	&& NVTIndex > LEGACY_NVT_CAPACITY_SCOPE0) {
		errCode = LDRF_ERR_NOT_FOUND;
	}

    return errCode;
} /* end of: LdrfGetFPTNVEx */

LDRF_FN LdrfGetFPTNVMemberNumber(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVIndex, PUShort pNVMemberNumber) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfGetFPTNVCommonSetup(pInfo, FPTIndex, NVIndex);
    TUShort memberNumber, recSize;

    if (!errCode) {
        if (pInfo->majorFmtVer < 3) {
            /* In versions before '3', the NVIndex and the NVMemberNumber
             * were the same.  The Member Number was not explicitly saved. */
            memberNumber = NVIndex;
        } else {
            if (!errCode) {
                /* Read record size (but just throw it away) */
                errCode = LdrfFileReadShort(pInfo, &recSize);
            }

            if (!errCode) {
                /* For FPT file format-version 3 and above, the FPTNV sub-record
                 * contains an explicit member number that may be different
                 * than the index.  It follows the recSize, but precedes the name. */
                errCode = LdrfFileReadShort(pInfo, &memberNumber);
            }
        }
    }

    if ((!errCode) && (pNVMemberNumber)) {
        * pNVMemberNumber = memberNumber;
    }

    return errCode;
} /* end of: LdrfGetFPTNVMemberNumber */

LDRF_FN LdrfGetFPTNVIndex(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVMemberNumber, PUShort pNVIndex) {
    /* Local variables */
    TUShort totalNVs, NVIndex = 0;
    TUByte manNVs, optNVs;
    TLdrfErrCodes errCode = LdrfGetFPTCommonSetup(pInfo, FPTIndex, &manNVs, &optNVs, NULL, NULL, NULL);

    if (!errCode) {
        totalNVs = manNVs + optNVs;

        if (pInfo->majorFmtVer < 3) {
            /* The member number and the index are the same for ver < 3.
             * This code just checks that the member number is valid. */
            if ((NVMemberNumber == 0) || (NVMemberNumber > totalNVs)) {
                errCode = LDRF_ERR_PARAM;
            } else {
                NVIndex = NVMemberNumber;
            }
        } else {
            /* Skip through the FPTNV sub-records, until we get to one
             * whose member number matches the one passed in. */
            TUShort index, recSize, memberNumber;
            TULong fptNVRecStart;

            for ( index = 1; index <= totalNVs; ++index ) {
                errCode = LdrfFileReadShort(pInfo, &recSize);

                if (errCode) break;

                errCode = LdrfFileWhere(pInfo);

                if (errCode) break;
                fptNVRecStart = pInfo->currentAddress;
                errCode = LdrfFileReadShort(pInfo, &memberNumber);

                if (errCode) break;

                if (memberNumber == NVMemberNumber) {
                    NVIndex = index;
                    break;
                }
                errCode = LdrfFileSeek(pInfo, fptNVRecStart + recSize);

                if (errCode) break;
            }

            if (!NVIndex) {
                errCode = LDRF_ERR_NOT_FOUND;
            }
        }
    }

    if (!errCode) {
        if (pNVIndex) {
            * pNVIndex = NVIndex;
        }
    }

    return errCode;
} /* end of: LdrfGetFPTNVIndex */

TLdrfErrCodes LdrfGetFPTCPCommonSetup(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex) {
    /* Local variables */
    TUShort totalNVs, totalCPs;
    TUByte manNVs, optNVs, manCPs, optCPs;
    TLdrfErrCodes errCode = LdrfGetFPTCommonSetup(pInfo, FPTIndex, &manNVs, &optNVs, &manCPs, &optCPs, NULL);

    if (!errCode) {
        totalNVs = manNVs + optNVs;
        totalCPs = manCPs + optCPs;

        if ((CPIndex == 0) || (CPIndex > totalCPs)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            TUShort index, recSize;
            /* Skip through all the FPTNV sub-records. */
            for ( index = 1; index <= totalNVs; ++index ) {
                errCode = LdrfFileReadShort(pInfo, &recSize);

                if (errCode) break;

                errCode = LdrfFileWhere(pInfo);

                if (errCode) break;
                errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + recSize);

                if (errCode) break;
            }

            if (!errCode) {
                /* Skip through the FPTCP sub-records, until we get to the one
                 * requested by CPIndex. */
                for ( index = 1; index < CPIndex; ++index ) {
                    errCode = LdrfFileReadShort(pInfo, &recSize);

                    if (errCode) break;

                    errCode = LdrfFileWhere(pInfo);

                    if (errCode) break;
                    errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + recSize);

                    if (errCode) break;
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfGetFPTCPCommonSetup */

LDRF_FN LdrfGetFPTCP(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex, PUShort pAppliesTo, LPSTR pCPName, PUShort pLength, PBool pMandatory,
    PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pCPTScope, PUShort pCPTIndex, PUByte pModifyArray,
    PUShort pByteArrayLen, PUByteArray* ppDefault, PUByteArray* ppValMin, PUByteArray* ppValMax)
{
    return LdrfGetFPTCPEx(pInfo, FPTIndex, CPIndex, pAppliesTo, pCPName, pLength, pMandatory,
    pResNmScope, pResNmIndex, pResCmtScope, pResCmtIndex, pCPTScope, pCPTIndex, pModifyArray,
    pByteArrayLen, ppDefault, ppValMin, ppValMax, NULL);
}

LDRF_FN LdrfGetFPTCPEx(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex, PUShort pAppliesTo, LPSTR pCPName, PUShort pLength, PBool pMandatory,
    PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pCPTScope, PUShort pCPTIndex, PUByte pModifyArray,
    PUShort pByteArrayLen, PUByteArray* ppDefault, PUByteArray* ppValMin, PUByteArray* ppValMax, PUByteArray* ppValInvalid)
{
    return LdrfGetFPTCPEx2(pInfo, FPTIndex, CPIndex, pAppliesTo, pCPName, pLength, pMandatory,
    pResNmScope, pResNmIndex, pResCmtScope, pResCmtIndex, pCPTScope, pCPTIndex, pModifyArray,
    pByteArrayLen, ppDefault, ppValMin, ppValMax, ppValInvalid, NULL);
}


LDRF_FN LdrfGetFPTCPEx2(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex, PUShort pAppliesTo, LPSTR pCPName, PUShort pLength, PBool pMandatory,
    PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pCPTScope, PUShort pCPTIndex, PUByte pModifyArray,
    PUShort pByteArrayLen, PUByteArray* ppDefault, PUByteArray* ppValMin, PUByteArray* ppValMax, PUByteArray* ppValInvalid, 
    PUShort pFlags)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfGetFPTCPCommonSetup(pInfo, FPTIndex, CPIndex);
    LPSTR pCPNameInternal = NULL;

    if (!errCode) {
        TUShort recSize;
        /* Read record size (but just throw it away) */
        errCode = LdrfFileReadShort(pInfo, &recSize);
    }

    if (!errCode) {
        /* Read the "appliesTo" value */
        TUShort appliesTo;
        errCode = LdrfFileReadShort(pInfo, &appliesTo);

        if ((!errCode) && (pAppliesTo)) {
            * pAppliesTo = appliesTo;
        }
    }

    if ((!errCode) && (pInfo->majorFmtVer > 2)) {
        TUShort CPMemberNumber;
        /* For FPT file format-version 3 and above, the FPTCP sub-record
         * contains an explicit member number that may be different
         * than the index.  It follows the appliesTo field, but
         * precedes the name.
         * We are going to just throw it away, because this interface
         * does not use it.  For performance reasons, someday we
         * may cache it for reading later. */
        errCode = LdrfFileReadShort(pInfo, &CPMemberNumber);
    }

    if ((!errCode) && (pInfo->majorFmtVer > 3)) {
        TUShort CPMinArraySize, CPMaxArraySize;
        /* For FPT file format-version 4 and above, the FPTCP sub-record
         * contains two size values.  These follow the member-number
         * field (see above), but precede the name.
         * We are going to just throw these away, because this
         * interface does not use them.  For performance reasons,
         * someday we may cache them for reading later. */
        errCode = LdrfFileReadShort(pInfo, &CPMinArraySize);

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &CPMaxArraySize);
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pCPNameInternal);

        if (!errCode) {
            errCode = LdrfTransferStringOut(pCPName, pLength, pCPNameInternal);
        }
        FREE(pCPNameInternal);
    }

    if (!errCode) {
        TUByte mandatory;
        TUByte resNmScope, resCmtScope;
        TULong resNmIndex, resCmtIndex;
        TUShort flags = 0;
        errCode = LdrfFileReadBytes(pInfo, &mandatory, sizeof(mandatory));

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &resNmScope, sizeof(resNmScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &resNmIndex);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &resCmtScope, sizeof(resCmtScope));
        }

        if (!errCode) {
            errCode = LdrfFileRead24(pInfo, &resCmtIndex);
        }

        if (!errCode && (pInfo->majorFmtVer >= 5)) {
            errCode = LdrfFileReadShort(pInfo, &flags);
        }

        if (!errCode) {
            if (pMandatory) {
                * pMandatory = (TBool)mandatory;
            }

            if (pResNmScope) {
                * pResNmScope = resNmScope;
            }

            if (pResNmIndex) {
                * pResNmIndex = resNmIndex;
            }

            if (pResCmtScope) {
                * pResCmtScope = resCmtScope;
            }

            if (pResCmtIndex) {
                * pResCmtIndex = resCmtIndex;
            }

            if (pFlags) {
                * pFlags = flags;
            }
        }
    }

    if (!errCode) {
        TUByte CPTScope, modifyArray, controlBits;
        TUShort CPTIndex, byteArrayLen;
        PUByteArray pDefArray = NULL, pMinArray = NULL, pMaxArray = NULL, pInvalidArray = NULL;
        errCode = LdrfFileReadBytes(pInfo, &CPTScope, sizeof(CPTScope));

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &CPTIndex);
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &modifyArray, sizeof(modifyArray));
        }

        if (!errCode) {
            errCode = LdrfFileReadBytes(pInfo, &controlBits, sizeof(controlBits));
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &byteArrayLen);
        }

        if (!errCode) {
            if (pCPTScope) {
                * pCPTScope = CPTScope;
            }

            if (pCPTIndex) {
                * pCPTIndex = CPTIndex;
            }

            if (pModifyArray) {
                * pModifyArray = modifyArray;
            }

            if (pByteArrayLen) {
                * pByteArrayLen = byteArrayLen;
            }
        }

        if (!errCode && byteArrayLen) {

            // 27-Jan-2009  Bernd  2.31.16  
            // Temporaryfix (or permanent precaution): some types manage to report byteArrayLen of zero (correct for those cases that we saw
            // in debugging), yet manage to raise the FTPNV_INVALID_VAL_RG flag. The subsequent call to LdrfFielReadBytes fails with LDRF_ERR_PARAM,
            // as pInvalidArray is NULL as a result of the zero-length byteArray. 
            //
            // Once the root cause is understood (how come this flag is raised in these cases), this comment should be updated accordingly. 
            // 
            if (controlBits & DF_BIT) {
                pDefArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pDefArray, byteArrayLen);
            }

            if ((!errCode) && (controlBits &MN_BIT)) {
                pMinArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pMinArray, byteArrayLen);
            }

            if ((!errCode) && (controlBits &MX_BIT)) {
                pMaxArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pMaxArray, byteArrayLen);
            }

            if ((!errCode) && (controlBits & IV_BIT)) {
                pInvalidArray = NEW_ARRAY(TUByte, byteArrayLen);
                errCode = LdrfFileReadBytes(pInfo, pInvalidArray, byteArrayLen);
            }

            if (!errCode) {
                if (ppDefault) {
                    * ppDefault = pDefArray;
                    pDefArray = NULL;
                }

                if (ppValMin) {
                    * ppValMin = pMinArray;
                    pMinArray = NULL;
                }

                if (ppValMax) {
                    * ppValMax = pMaxArray;
                    pMaxArray = NULL;
                }

                if (ppValInvalid) {
                    * ppValInvalid = pInvalidArray;
                    pInvalidArray = NULL;
                }
            }
            FREE(pDefArray);
            FREE(pMinArray);
            FREE(pMaxArray);
            FREE(pInvalidArray);
        }
    }

    return errCode;
} /* end of: LdrfGetFPTCPEx */

LDRF_FN LdrfGetFPTCPByAttributes(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort appliesTo, TUByte CPTScope, TUShort CPTIndex, PUShort pCPIndex, LPSTR pCPName,
    PUShort pLength, PBool pMandatory, PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pModifyArray,
    PUShort pByteArrayLen, PUByteArray* ppDefault, PUByteArray* ppValMin, PUByteArray* ppValMax)
{
    return LdrfGetFPTCPByAttributesEx(pInfo, FPTIndex, appliesTo, CPTScope, CPTIndex, pCPIndex, pCPName,
    pLength, pMandatory, pResNmScope, pResNmIndex, pResCmtScope, pResCmtIndex, pModifyArray,
    pByteArrayLen, ppDefault, ppValMin, ppValMax, NULL);
}

LDRF_FN LdrfGetFPTCPByAttributesEx(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort appliesTo, TUByte CPTScope, TUShort CPTIndex, PUShort pCPIndex, LPSTR pCPName,
    PUShort pLength, PBool pMandatory, PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex, PUByte pModifyArray,
    PUShort pByteArrayLen, PUByteArray* ppDefault, PUByteArray* ppValMin, PUByteArray* ppValMax, PUByteArray* ppValInvalid)
    {
    /* Local variables */
    TUShort totalNVs, totalCPs;
    TUByte manNVs, optNVs, manCPs, optCPs, principalNV;
    TLdrfErrCodes errCode = LdrfGetFPTCommonSetup(pInfo, FPTIndex, &manNVs, &optNVs, &manCPs, &optCPs, &principalNV);
    LPSTR pCPNameInternal = NULL;

    if (!errCode) {
        /* Skip through all the FPTNV sub-records. */
        TUShort index, recSize;
        totalNVs = manNVs + optNVs;

        for ( index = 1; index <= totalNVs; ++index ) {
            errCode = LdrfFileReadShort(pInfo, &recSize);

            if (errCode) break;

            errCode = LdrfFileWhere(pInfo);

            if (errCode) break;
            errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + recSize);

            if (errCode) break;
        }
    }

    if (!errCode) {
        /* Read through all the FPTCP sub-records looking for one whose
         * attributes match. */
        TUShort CPIndex, recSize, appliesToField, CPTIndexField;
        TUByte resNmScope, resCmtScope, CPTScopeField, mandatory;
        TULong resNmIndex, resCmtIndex, fptCPRecStart;
        totalCPs = manCPs + optCPs;
        CPIndex = 1;

        while (CPIndex <= totalCPs) {
            errCode = LdrfFileReadShort(pInfo, &recSize);

            if (errCode) break;

            errCode = LdrfFileWhere(pInfo);

            if (errCode) break;
            fptCPRecStart = pInfo->currentAddress;
            errCode = LdrfFileReadShort(pInfo, &appliesToField);

            if (errCode) break;

            if ((!errCode) && (pInfo->majorFmtVer > 2)) {
                TUShort CPMemberNumber;
                /* For FPT file format-version 3 and above, the FPTCP sub-record
                 * contains an explicit member number that may be different
                 * than the index.  It follows the appliesTo field, but
                 * precedes the name.
                 * We are going to just throw it away, because this interface
                 * does not use it.  For performance reasons, someday we
                 * may cache it for reading later. */
                errCode = LdrfFileReadShort(pInfo, &CPMemberNumber);
            }

            if ((!errCode) && (pInfo->majorFmtVer > 3)) {
                TUShort CPMinArraySize, CPMaxArraySize;
                /* For FPT file format-version 4 and above, the FPTCP sub-record
                 * contains two size values.  These follow the member-number
                 * field (see above), but precede the name.
                 * We are going to just throw these away, because this
                 * interface does not use them.  For performance reasons,
                 * someday we may cache them for reading later. */
                errCode = LdrfFileReadShort(pInfo, &CPMinArraySize);

                if (!errCode) {
                    errCode = LdrfFileReadShort(pInfo, &CPMaxArraySize);
                }
            }
            FREE(pCPNameInternal); /* Throw away any old (leftover) name */
            errCode = LdrfFileReadLString(pInfo, &pCPNameInternal);

            if (errCode) break;
            errCode = LdrfFileReadBytes(pInfo, &mandatory, sizeof(mandatory));

            if (errCode) break;
            errCode = LdrfFileReadBytes(pInfo, &resNmScope, sizeof(resNmScope));

            if (errCode) break;
            errCode = LdrfFileRead24(pInfo, &resNmIndex);

            if (errCode) break;
            errCode = LdrfFileReadBytes(pInfo, &resCmtScope, sizeof(resCmtScope));

            if (errCode) break;
            errCode = LdrfFileRead24(pInfo, &resCmtIndex);

            if (errCode) break;
            errCode = LdrfFileReadBytes(pInfo, &CPTScopeField, sizeof(CPTScopeField));

            if (errCode) break;
            errCode = LdrfFileReadShort(pInfo, &CPTIndexField);

            if (errCode) break;


            /* This is the old code prior to the fix for EPR 14130:
                    if	(   (	(appliesTo == appliesToField)
                         || ((appliesToField == 0) && (appliesTo == principalNV))
                         || ((appliesToField == principalNV) && (appliesTo == 0)) )
                     && (CPTScope == CPTScopeField) && (CPTIndex == CPTIndexField) ) {
                    break;
                    }
            */

            /* This is the new code after the fix for EPR 14130: */
            if ((appliesTo == appliesToField) && (CPTScope == CPTScopeField) && (CPTIndex == CPTIndexField)) {
                break;
            }

            ++CPIndex;
            errCode = LdrfFileSeek(pInfo, fptCPRecStart + recSize);

            if (errCode) break;
        }

        if (!errCode) {
            if (CPIndex > totalCPs) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else {
                if (pCPIndex) {
                    * pCPIndex = CPIndex;
                }
                errCode = LdrfTransferStringOut(pCPName, pLength, pCPNameInternal);

                if (!errCode) {
                    if (pMandatory) {
                        * pMandatory = (TBool)mandatory;
                    }

                    if (pResNmScope) {
                        * pResNmScope = resNmScope;
                    }

                    if (pResNmIndex) {
                        * pResNmIndex = resNmIndex;
                    }

                    if (pResCmtScope) {
                        * pResCmtScope = resCmtScope;
                    }

                    if (pResCmtIndex) {
                        * pResCmtIndex = resCmtIndex;
                    }
                }

                if (!errCode) {
                    TUByte modifyArray, controlBits;
                    TUShort byteArrayLen;

                    if (!errCode) {
                        errCode = LdrfFileReadBytes(pInfo, &modifyArray, sizeof(modifyArray));
                    }

                    if (!errCode) {
                        errCode = LdrfFileReadBytes(pInfo, &controlBits, sizeof(controlBits));
                    }

                    if (!errCode) {
                        errCode = LdrfFileReadShort(pInfo, &byteArrayLen);
                    }

                    if (!errCode) {
                        PUByteArray pDefArray = NULL;
                        PUByteArray pMinArray = NULL, pMaxArray = NULL, pInvalidArray = NULL;

                        if (pModifyArray) {
                            * pModifyArray = modifyArray;
                        }

                        if (pByteArrayLen) {
                            * pByteArrayLen = byteArrayLen;
                        }

                        if (controlBits & DF_BIT) {
                            pDefArray = NEW_ARRAY(TUByte, byteArrayLen);
                            errCode = LdrfFileReadBytes(pInfo, pDefArray, byteArrayLen);
                        }

                        if ((!errCode) && (controlBits &MN_BIT)) {
                            pMinArray = NEW_ARRAY(TUByte, byteArrayLen);
                            errCode = LdrfFileReadBytes(pInfo, pMinArray, byteArrayLen);
                        }

                        if ((!errCode) && (controlBits &MX_BIT)) {
                            pMaxArray = NEW_ARRAY(TUByte, byteArrayLen);
                            errCode = LdrfFileReadBytes(pInfo, pMaxArray, byteArrayLen);
                        }

                        if ((!errCode) && (controlBits & IV_BIT)) {
                            pInvalidArray = NEW_ARRAY(TUByte, byteArrayLen);
                            errCode = LdrfFileReadBytes(pInfo, pInvalidArray, byteArrayLen);
                        }

                        if (!errCode) {
                            if (ppDefault) {
                                * ppDefault = pDefArray;
                                pDefArray = NULL;
                            }

                            if (ppValMin) {
                                * ppValMin = pMinArray;
                                pMinArray = NULL;
                            }

                            if (ppValMax) {
                                * ppValMax = pMaxArray;
                                pMaxArray = NULL;
                            }

                            if (ppValInvalid) {
                                * ppValInvalid = pInvalidArray;
                                pInvalidArray = NULL;
                            }
                        }
                        FREE(pDefArray);
                        FREE(pMinArray);
                        FREE(pMaxArray);
                        FREE(pInvalidArray);
                    }
                }
            }
        }
        FREE(pCPNameInternal);
    }

    return errCode;
} /* end of: LdrfGetFPTCPByAttributesEx */

LDRF_FN LdrfGetFPTCPMemberNumber(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex, PUShort pCPMemberNumber) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfGetFPTCPCommonSetup(pInfo, FPTIndex, CPIndex);
    TUShort memberNumber, dummy;

    if (!errCode) {
        if (pInfo->majorFmtVer < 3) {
            /* In versions before '3', the CPIndex and the CPMemberNumber
             * were the same.  The Member Number was not explicitly saved. */
            memberNumber = CPIndex;
        } else {
            if (!errCode) {
                /* Read record size (but just throw it away) */
                errCode = LdrfFileReadShort(pInfo, &dummy);
            }

            if (!errCode) {
                /* Also read the "applies-to" field, and throw it away */
                errCode = LdrfFileReadShort(pInfo, &dummy);
            }

            if (!errCode) {
                /* For FPT file format-version 3 and above, the FPTCP sub-record
                 * contains an explicit member number that may be different
                 * than the index.  It follows the recSize, but precedes the name. */
                errCode = LdrfFileReadShort(pInfo, &memberNumber);
            }
        }
    }

    if ((!errCode) && (pCPMemberNumber)) {
        * pCPMemberNumber = memberNumber;
    }

    return errCode;
} /* end of: LdrfGetFPTCPMemberNumber */

LDRF_FN LdrfGetFPTCPIndex(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPMemberNumber, PUShort pCPIndex) {
    /* Local variables */
    TUShort totalNVs, totalCPs, CPIndex = 0;
    TUByte manNVs, optNVs, manCPs, optCPs;
    TLdrfErrCodes errCode = LdrfGetFPTCommonSetup(pInfo, FPTIndex, &manNVs, &optNVs, &manCPs, &optCPs, NULL);

    if (!errCode) {
        totalNVs = manNVs + optNVs;
        totalCPs = manCPs + optCPs;

        if (pInfo->majorFmtVer < 3) {
            /* The member number and the index are the same for ver < 3.
             * This code just checks that the member number is valid. */
            if ((CPMemberNumber == 0) || (CPMemberNumber > totalCPs)) {
                errCode = LDRF_ERR_PARAM;
            } else {
                CPIndex = CPMemberNumber;
            }
        } else {
            /* Skip through all the FPTNV sub-records. */
            TUShort index, recSize, memberNumber, appliesTo;
            totalNVs = manNVs + optNVs;

            for ( index = 1; index <= totalNVs; ++index ) {
                errCode = LdrfFileReadShort(pInfo, &recSize);

                if (errCode) break;

                errCode = LdrfFileWhere(pInfo);

                if (errCode) break;
                errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + recSize);

                if (errCode) break;
            }

            if (!errCode) {
                TULong fptCPRecStart;
                /* Skip through the FPTCP sub-records, until we get to one
                 * whose member number matches the one passed in. */
                for ( index = 1; index <= totalCPs; ++index ) {
                    errCode = LdrfFileReadShort(pInfo, &recSize);

                    if (errCode) break;

                    errCode = LdrfFileWhere(pInfo);

                    if (errCode) break;
                    fptCPRecStart = pInfo->currentAddress;
                    errCode = LdrfFileReadShort(pInfo, &appliesTo);

                    if (errCode) break;
                    errCode = LdrfFileReadShort(pInfo, &memberNumber);

                    if (errCode) break;

                    if (memberNumber == CPMemberNumber) {
                        CPIndex = index;
                        break;
                    }
                    errCode = LdrfFileSeek(pInfo, fptCPRecStart + recSize);

                    if (errCode) break;
                }

                if (!CPIndex) {
                    errCode = LDRF_ERR_NOT_FOUND;
                }
            }
        }
    }

    if ((!errCode) && (pCPIndex)) {
        * pCPIndex = CPIndex;
    }

    return errCode;
} /* end of: LdrfGetFPTCPIndex */

LDRF_FN LdrfGetFPTCPArrayDetails(PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex, PUShort pMinArraySize, PUShort pMaxArraySize,
    PBool pDetailsAreDefaults)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfGetFPTCPCommonSetup(pInfo, FPTIndex, CPIndex);
    TUShort dummy, minArraySize = 0, maxArraySize = 0;
    TBool detailsAreDefaults = T;

    if (!errCode) {
        if (pInfo->majorFmtVer > 3) {
            /* In versions before '4', there was no array size information.
               In that case, we just default to "non-array" values. */
            if (!errCode) {
                /* Read record size (but just throw it away) */
                errCode = LdrfFileReadShort(pInfo, &dummy);
            }

            if (!errCode) {
                /* Also read the "applies-to" field, and throw it away */
                errCode = LdrfFileReadShort(pInfo, &dummy);
            }

            if (!errCode) {
                /* Also read the explicit member number, and throw it away */
                errCode = LdrfFileReadShort(pInfo, &dummy);
            }

            if (!errCode) {
                /* Read the min Array Size */
                errCode = LdrfFileReadShort(pInfo, &minArraySize);
            }

            if (!errCode) {
                /* Read the max Array Size */
                errCode = LdrfFileReadShort(pInfo, &maxArraySize);
            }

            if (!errCode) {
                detailsAreDefaults = F;
            }
        }
    }

    if (pMinArraySize) {
        * pMinArraySize = minArraySize;
    }

    if (pMaxArraySize) {
        * pMaxArraySize = maxArraySize;
    }

    if (pDetailsAreDefaults) {
        * pDetailsAreDefaults = detailsAreDefaults;
    }

    return errCode;
} /* end of: LdrfGetFPTCPArrayDetails */
