/*
 * DRFTYPE.C - This is the Resource File API file responsible for access methods for type files.
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */

#include "drfport.h"

#ifdef LCADRF_WRITE
#include "lcadrfiw.h"

#else
#include "lcadrfi.h"

#endif

TLdrfErrCodes LdrfReadTypeFileHeader(PLdrfFileInfo pInfo, TBool checkCRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfReadFileHeader(pInfo, checkCRC);
    TUShort count;

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, pInfo->refID, sizeof(pInfo->refID));
    }

    if (!errCode) {
        errCode = LdrfFileReadBytes(pInfo, &(pInfo->scope), sizeof(pInfo->scope));
    }

    for ( count = 0; ((count < NUM_SCOPES) && (!errCode)); ++count ) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.resDeps[count]));
    }

    for ( count = 0; ((count < NUM_SCOPES) && (!errCode)); ++count ) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.typDeps[count]));
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.numNVTs));
        pInfo->u.type.capacityNVT = pInfo->u.type.numNVTs;
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.numCPTs));
        pInfo->u.type.capacityCPT = pInfo->u.type.numCPTs;
    }

    if (!errCode) {
        errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.numEnumSets));
        pInfo->u.type.capacityEnumSets = pInfo->u.type.numEnumSets;
    }

    if (pInfo->majorFmtVer > 3) {
        // Format 4 & later of TYP file also has 3 "numEmpty" counts
        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.numEmptyNVTs));
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.numEmptyCPTs));
        }

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pInfo->u.type.numEmptyEnumSets));
        }
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.type.numNVTs, pInfo->u.type.numNVTs, &(pInfo->u.type.pDirNVTIndex));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.type.numNVTs, pInfo->u.type.numNVTs - pInfo->u.type.numEmptyNVTs, &(pInfo->u.type.pDirNVTName));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.type.numCPTs, pInfo->u.type.numCPTs, &(pInfo->u.type.pDirCPTIndex));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.type.numCPTs, pInfo->u.type.numCPTs - pInfo->u.type.numEmptyCPTs, &(pInfo->u.type.pDirCPTName));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.type.numEnumSets, pInfo->u.type.numEnumSets, &(pInfo->u.type.pDirEnumIndex));
    }

    if (!errCode) {
        errCode = LdrfReadDirectory(pInfo, pInfo->u.type.numEnumSets, pInfo->u.type.numEnumSets - pInfo->u.type.numEmptyEnumSets, &(pInfo->u.type.pDirEnumTag));
    }

    if (!errCode) {
        errCode =
            LdrfReadDirectory(pInfo, pInfo->u.type.numEnumSets, pInfo->u.type.numEnumSets - pInfo->u.type.numEmptyEnumSets, &(pInfo->u.type.pDirEnumFile));
    }

    if (!errCode) {
        pInfo->u.type.needsTypeFileInfo = F;
        pInfo->u.type.pEnumSet = NULL;
    }

    return errCode;
} /* end of: LdrfReadTypeFileHeader */

TLdrfErrCodes LdrfFreeTypeFileInfo(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;

    FREE(pInfo->u.type.pDirNVTIndex);
    FREE(pInfo->u.type.pDirNVTName);
    FREE(pInfo->u.type.pDirCPTIndex);
    FREE(pInfo->u.type.pDirCPTName);
    FREE(pInfo->u.type.pDirEnumIndex);
    FREE(pInfo->u.type.pDirEnumTag);
    FREE(pInfo->u.type.pDirEnumFile);

    if (pInfo->u.type.pEnumSet) {
        errCode = LDRF_ERR_INTERNAL;
    }

    return errCode;
} /* end of: LdrfFreeTypeFileInfo */

LDRF_FN LdrfGetTypeFileInfo(PLdrfFileInfo pInfo, PUShort pResDep0, PUShort pResDep1, PUShort pResDep2, PUShort pResDep3, PUShort pResDep4, PUShort pResDep5,
    PUShort pResDep6, PUShort pTypDep0, PUShort pTypDep1, PUShort pTypDep2, PUShort pTypDep3, PUShort pTypDep4, PUShort pTypDep5, PUShort pTypDep6,
    PUShort pNumNVTs, PUShort pNumCPTs, PUShort pNumEnumSets)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if (pResDep0) {
            * pResDep0 = pInfo->u.type.resDeps[0];
        }

        if (pResDep1) {
            * pResDep1 = pInfo->u.type.resDeps[1];
        }

        if (pResDep2) {
            * pResDep2 = pInfo->u.type.resDeps[2];
        }

        if (pResDep3) {
            * pResDep3 = pInfo->u.type.resDeps[3];
        }

        if (pResDep4) {
            * pResDep4 = pInfo->u.type.resDeps[4];
        }

        if (pResDep5) {
            * pResDep5 = pInfo->u.type.resDeps[5];
        }

        if (pResDep6) {
            * pResDep6 = pInfo->u.type.resDeps[6];
        }

        if (pTypDep0) {
            * pTypDep0 = pInfo->u.type.typDeps[0];
        }

        if (pTypDep1) {
            * pTypDep1 = pInfo->u.type.typDeps[1];
        }

        if (pTypDep2) {
            * pTypDep2 = pInfo->u.type.typDeps[2];
        }

        if (pTypDep3) {
            * pTypDep3 = pInfo->u.type.typDeps[3];
        }

        if (pTypDep4) {
            * pTypDep4 = pInfo->u.type.typDeps[4];
        }

        if (pTypDep5) {
            * pTypDep5 = pInfo->u.type.typDeps[5];
        }

        if (pTypDep6) {
            * pTypDep6 = pInfo->u.type.typDeps[6];
        }

        if (pNumNVTs) {
			if (pInfo->extendedSNVTIDaware
			&&  pInfo->scope == 0) {
				*pNumNVTs = pInfo->u.type.numNVTs;
			} else {
				*pNumNVTs = min(pInfo->u.type.numNVTs, LEGACY_NVT_CAPACITY_SCOPE0);
			}
        }

        if (pNumCPTs) {
            * pNumCPTs = pInfo->u.type.numCPTs;
        }

        if (pNumEnumSets) {
            * pNumEnumSets = pInfo->u.type.numEnumSets;
        }
    }

    return errCode;
} /* end of: LdrfGetTypeFileInfo */

LDRF_FN LdrfSelectEnumSet(PLdrfFileInfo pInfo, TUShort index, LPSTR pTag, PUShort pTagLen, LPSTR pFile, PUShort pFileLen) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateEnumSet(pInfo, index);

    if ((!errCode) || (errCode == LDRF_ERR_NOT_FOUND) || (errCode == LDRF_ERR_EMPTY_RECORD)) {
        if ((!(pInfo->u.type.pEnumSet)) || (pInfo->u.type.pEnumSet->index != index)) {
            /* There is not currently selected enum set, or it does not match */
            TLdrfErrCodes ec2 = LdrfDeselectEnumSet(pInfo);

            if (ec2) {
                errCode = ec2;
            }

            if (errCode == LDRF_ERR_EMPTY_RECORD) {
                if (!(pInfo->emptyRecordAware)) {
                    char nameBuf[20];
                    strcpy(nameBuf, "empty");
                    _itoa(index, nameBuf + 5, 10);
                    strcat(nameBuf, "~");
                    errCode = LdrfAllocateEnumSetStruct(&(pInfo->u.type.pEnumSet), index, 0);

                    if (!errCode) {
                        PEnumSet pEnumSet = pInfo->u.type.pEnumSet;
                        pEnumSet->index = index;
                        pEnumSet->pEnumTag = _strdup(nameBuf);
                        pEnumSet->pEnumFile = _strdup(nameBuf);

                        if ((!(pEnumSet->pEnumTag)) || (!(pEnumSet->pEnumFile))) {
                            errCode = LDRF_ERR_SYS;
                        }
                    }
                }
            } else if (!errCode) {
                errCode = LdrfAllocateEnumSetStruct(&(pInfo->u.type.pEnumSet), index, pInfo->u.type.pDirEnumIndex[index]);

                if (!errCode) {
                    errCode = LdrfReadEnumSet(pInfo);
                }
            }
        }
    }

    if (!errCode) {
        errCode = LdrfTransferStringOut(pTag, pTagLen, pInfo->u.type.pEnumSet->pEnumTag);
    }

    if (!errCode) {
        errCode = LdrfTransferStringOut(pFile, pFileLen, pInfo->u.type.pEnumSet->pEnumFile);
    }

    return errCode;
} /* end of: LdrfSelectEnumSet */

LDRF_FN LdrfValidateEnumSet(PLdrfFileInfo pInfo, TUShort index) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if (index < 1) {
            errCode = LDRF_ERR_PARAM;
        } else if (index > pInfo->u.type.numEnumSets) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else if (!(pInfo->u.type.pDirEnumIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.type.pDirEnumIndex[index])) {
            errCode = LDRF_ERR_EMPTY_RECORD;
        }
    }

    return errCode;
} /* end of: LdrfValidateEnumSet */

LDRF_FN LdrfSelectEnumSetByTag(PLdrfFileInfo pInfo, LPCSTR tag, PUShort pIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    TULong entry;

    if (!errCode) {
        errCode = LdrfDeselectEnumSet(pInfo);
    }

    if (!errCode) {
        errCode =
            LdrfFindInSortedDir(pInfo, pInfo->u.type.pDirEnumTag, pInfo->u.type.numEnumSets - pInfo->u.type.numEmptyEnumSets, tag, &entry, LdrfEnumTagCompare);
    }

    if (!errCode) {
        errCode = LdrfAllocateEnumSetStruct(&(pInfo->u.type.pEnumSet), 0, pInfo->u.type.pDirEnumTag[entry]);
    }

    if (!errCode) {
        pInfo->u.type.pEnumSet->tagDirEntry = (TUShort)entry;
        errCode = LdrfReadEnumSet(pInfo);
    }

    if ((!errCode) && (pIndex)) {
        * pIndex = pInfo->u.type.pEnumSet->index;
    }

    return errCode;
} /* end of: LdrfSelectEnumSetByTag */

LDRF_FN LdrfSelectEnumSetByFile(PLdrfFileInfo pInfo, LPCSTR file, PUShort pIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    TULong entry;

    if (!errCode) {
        errCode = LdrfDeselectEnumSet(pInfo);
    }

    if (!errCode) {
        errCode =
            LdrfFindInSortedDir(pInfo, pInfo->u.type.pDirEnumFile, pInfo->u.type.numEnumSets - pInfo->u.type.numEmptyEnumSets, file, &entry,
                LdrfEnumFileCompare);
    }

    if (!errCode) {
        errCode = LdrfAllocateEnumSetStruct(&(pInfo->u.type.pEnumSet), 0, pInfo->u.type.pDirEnumFile[entry]);
    }

    if (!errCode) {
        pInfo->u.type.pEnumSet->fileDirEntry = (TUShort)entry;
        errCode = LdrfReadEnumSet(pInfo);
    }

    if ((!errCode) && (pIndex)) {
        * pIndex = pInfo->u.type.pEnumSet->index;
    }

    return errCode;
} /* end of: LdrfSelectEnumSetByFile */

LDRF_FN LdrfGetEnumMemberCount(PLdrfFileInfo pInfo, PUShort pNumMembers) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if ((!errCode) && (!(pInfo->u.type.pEnumSet))) {
        errCode = LDRF_ERR_SEQUENCE;
    }

    if (!errCode) {
        if (pNumMembers) {
            * pNumMembers = pInfo->u.type.pEnumSet->numEnums;
        }
    }

    return errCode;
} /* end of: LdrfGetEnumMember */

TLdrfErrCodes LdrfAllocateEnumSetStruct(PEnumSet* ppEnumSet, TUShort index, TULong startAddress) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_SYS;
    PEnumSet pEnumSet = NEW(TEnumSet);

    if (pEnumSet) {
        errCode = LDRF_ERR_NONE;
        memset(pEnumSet, 0, sizeof(TEnumSet));
        pEnumSet->index = index;
        pEnumSet->address = startAddress;
        * ppEnumSet = pEnumSet;
    }

    return errCode;
} /* end of: LdrfAllocateEnumSetStruct */

TLdrfErrCodes LdrfDeselectEnumSet(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PEnumSet pEnumSet = pInfo->u.type.pEnumSet;
    TUShort i;

    if (pEnumSet) {
        if (pEnumSet->needsToBeWritten) {
#ifdef LCADRF_WRITE

            errCode = LdrfOutputEnumSet(pInfo);
#else

            errCode = LDRF_ERR_INTERNAL;

#endif

        }
        FREE(pEnumSet->pEnumTag);
        FREE(pEnumSet->pEnumFile);

        for ( i = 0; i < pEnumSet->numEnums; ++i ) {
            FREE(pEnumSet->e[i].eString);
        }
        pInfo->u.type.pEnumSet = NULL;
        FREE(pEnumSet);
    }

    return errCode;
} /* end of: LdrfDeselectEnumSet */

TLdrfErrCodes LdrfReadEnumSet(PLdrfFileInfo pInfo) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_INTERNAL;

    if ((pInfo) && (pInfo->u.type.pEnumSet) && (pInfo->u.type.pEnumSet->address)) {
        PEnumSet pEnumSet = pInfo->u.type.pEnumSet;
        errCode = LdrfFileSeek(pInfo, pEnumSet->address);

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, &(pEnumSet->index));
        }

        if (!errCode) {
            TUShort dummy; /* Skip over the record size */
            errCode = LdrfFileReadBytes(pInfo, &dummy, sizeof(dummy));
        }

        if (!errCode) {
            errCode = LdrfFileReadLString(pInfo, &(pEnumSet->pEnumTag));
        }

        if (!errCode) {
            errCode = LdrfFileReadLString(pInfo, &(pEnumSet->pEnumFile));
        }

        if (!errCode) {
            if (pInfo->majorFmtVer == 2) {
                /* As a matter of accident rather than design, the older
                 * format version of the type file did not use the proper
                 * routine (for byte-swapping) of the number of Enums.
                 * Thus it was byte-swapped when not supposed to be.
                 * This older code, below has been modified to be platform-
                 * independent. */
                TUShort temp;
                errCode = LdrfFileReadBytes(pInfo, &temp, sizeof(temp));
#ifndef LCADRF_BYTE_SWAP

                {
                    TUByte* p = (TUByte* )( &temp);
                    TUByte swapTemp = p[0];
                    p[0] = p[1];
                    p[1] = swapTemp;
                }

#endif

                pEnumSet->numEnums = temp;
            } else {
                /* The newer code, for format-version > 3 */
                errCode = LdrfFileReadShort(pInfo, &(pEnumSet->numEnums));
            }
        }

        if (!errCode) {
            TUShort count, relative;
            TULong rememberWhere;

            for ( count = 0; count < pEnumSet->numEnums; ++count ) {
                errCode = LdrfFileReadBytes(pInfo, &(pEnumSet->e[count].eValue), sizeof(pEnumSet->e[count].eValue));

                if (errCode) break;

                errCode = LdrfFileReadShort(pInfo, &relative);

                if (errCode) break;
                errCode = LdrfFileReadBytes(pInfo, &(pEnumSet->e[count].eResScope), sizeof(pEnumSet->e[count].eResScope));

                if (errCode) break;
                errCode = LdrfFileRead24(pInfo, &(pEnumSet->e[count].eResIndex));

                if (errCode) break;
                errCode = LdrfFileWhere(pInfo);

                if (errCode) break;
                rememberWhere = pInfo->currentAddress;
                errCode = LdrfFileSeek(pInfo, pEnumSet->address + relative);

                if (errCode) break;
                errCode = LdrfFileReadLString(pInfo, &(pEnumSet->e[count].eString));

                if (errCode) break;
                errCode = LdrfFileSeek(pInfo, rememberWhere);

                if (errCode) break;
            }
        }
    }

    return errCode;
} /* end of: LdrfReadEnumSet */

TLdrfErrCodes LdrfEnumTagCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, entry2Addr + 4); /* Tag string */
    LPSTR pEntry2 = NULL;

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pEntry2);
    }

    if (!errCode) {
        * pRC = strcmp(pEntry1, pEntry2);
    }
    FREE(pEntry2);

    return errCode;
} /* end of: LdrfEnumTagCompare */

TLdrfErrCodes LdrfEnumFileCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfFileSeek(pInfo, entry2Addr + 4); /* Tag string */
    LPSTR pEntry2 = NULL;

    if (!errCode) {
        /* Skip past the tag string to the file string */
        TUByte stringLen;
        errCode = LdrfFileReadBytes(pInfo, &stringLen, sizeof(stringLen));

        if (!errCode) {
            errCode = LdrfFileWhere(pInfo);
        }

        if (!errCode) {
            errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + stringLen);
        }
    }

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pEntry2);
    }

    if (!errCode) {
        * pRC = STRICMP(pEntry1, pEntry2);
    }
    FREE(pEntry2);

    return errCode;
} /* end of: LdrfEnumFileCompare */

LDRF_FN LdrfGetEnumMember(PLdrfFileInfo pInfo, TByte value, LPSTR pString, PUShort pLength, PUByte pResScope, PULong pResIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if ((!errCode) && (!(pInfo->u.type.pEnumSet))) {
        errCode = LDRF_ERR_SEQUENCE;
    }

    if (!errCode) {
        PEnumSet pEnumSet = pInfo->u.type.pEnumSet;
        TUShort index;

        for ( index = 0; (index < pEnumSet->numEnums); ++index ) {
            if (pEnumSet->e[index].eValue == value) {
                break;
            }
        }

        if (index >= pEnumSet->numEnums) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else {
            if (pResScope) {
                * pResScope = pEnumSet->e[index].eResScope;
            }

            if (pResIndex) {
                * pResIndex = pEnumSet->e[index].eResIndex;
            }
            errCode = LdrfTransferStringOut(pString, pLength, pEnumSet->e[index].eString);
        }
    }

    return errCode;
} /* end of: LdrfGetEnumMember */

LDRF_FN LdrfGetEnumValue(PLdrfFileInfo pInfo, LPCSTR string, PByte pValue) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if ((!errCode) && (!(pInfo->u.type.pEnumSet))) {
        errCode = LDRF_ERR_SEQUENCE;
    }

    if (!errCode) {
        PEnumSet pEnumSet = pInfo->u.type.pEnumSet;
        TUShort index;

        for ( index = 0; (index < pEnumSet->numEnums); ++index ) {
            if (!strcmp(pEnumSet->e[index].eString, string)) {
                break;
            }
        }

        if (index > pEnumSet->numEnums) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else if (pValue) {
            * pValue = pEnumSet->e[index].eValue;
        }
    }

    return errCode;
} /* end of: LdrfGetEnumValue */

LDRF_FN LdrfGetEnumMemberByIndex(PLdrfFileInfo pInfo, TUShort index, PByte pValue, LPSTR pString, PUShort pLength, PUByte pResScope, PULong pResIndex) {
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if ((!errCode) && (!(pInfo->u.type.pEnumSet))) {
        errCode = LDRF_ERR_SEQUENCE;
    } else if (index == 0) {
        errCode = LDRF_ERR_PARAM;
    } else {
        --index; // Incoming index is 1-based, but internal is 0-based
    }

    if (!errCode) {
        PEnumSet pEnumSet = pInfo->u.type.pEnumSet;

        if (index >= pEnumSet->numEnums) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else {
            if (pValue) {
                * pValue = pEnumSet->e[index].eValue;
            }

            if (pResScope) {
                * pResScope = pEnumSet->e[index].eResScope;
            }

            if (pResIndex) {
                * pResIndex = pEnumSet->e[index].eResIndex;
            }
            errCode = LdrfTransferStringOut(pString, pLength, pEnumSet->e[index].eString);
        }
    }

    return errCode;
} /* end of: LdrfGetEnumMemberByIndex */

LDRF_FN LdrfGetNVT(PLdrfFileInfo pInfo, TUShort index, PLdrfTypeTree* ppTypeTree) {
    return LdrfGetNVTEx(pInfo, index, ppTypeTree, NULL);
}

LDRF_FN LdrfGetNVTEx(PLdrfFileInfo pInfo, TUShort index, PLdrfTypeTree* ppTypeTree, PUShort pFlags) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateNVT(pInfo, index);
    LPCSTR prefix;

    if ((!errCode) || (errCode == LDRF_ERR_EMPTY_RECORD)) {
        if (pInfo->userFlag) {
            prefix = "UNVT";
        } else {
            prefix = "SNVT";
        }

        if (!ppTypeTree) {
            errCode = LDRF_ERR_PARAM;
        } else if (errCode == LDRF_ERR_EMPTY_RECORD) {
            if (!(pInfo->emptyRecordAware)) {
                PLdrfTypeTree pTypeTree = NULL;
                char nameBuf[20];
                strcpy(nameBuf, prefix);
                strcpy(nameBuf + 4, "empty");
                _itoa(index, nameBuf + 9, 10);
                strcat(nameBuf, "~");
                errCode = LdrfNewTypeTreeNode(NVT_TYPE_UNSIGNED_SHORT, nameBuf, 0, 0, 0, 0, 0, 0, &pTypeTree);

                if (errCode == LDRF_ERR_INCOMPLETE) {
                    errCode = LdrfSetScalarDetails(pTypeTree, 0, 255, 1, 0, 0);
                }

                if (!errCode) {
                    * ppTypeTree = pTypeTree;
                    if (pFlags) {
                        *pFlags = 0;
                    }
                } else {
                    LdrfFreeTypeTree(pTypeTree);
                }
            }
        } else {
            errCode = LdrfNVTRead(pInfo, prefix, pInfo->u.type.pDirNVTIndex[index], NULL, ppTypeTree, pFlags);
        }
    }

    return errCode;
} /* end of: LdrfGetNVT */

LDRF_FN LdrfValidateNVT(PLdrfFileInfo pInfo, TUShort index) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if (index < 1) {
            errCode = LDRF_ERR_PARAM;
        } else if (index > pInfo->u.type.numNVTs) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else if (!(pInfo->u.type.pDirNVTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.type.pDirNVTIndex[index])) {
            errCode = LDRF_ERR_EMPTY_RECORD;
        }
    }

	if (pInfo->scope == 0
	&&  !pInfo->extendedSNVTIDaware
	&&	index > LEGACY_NVT_CAPACITY_SCOPE0) {
		errCode = LDRF_ERR_NOT_FOUND;
	}

    return errCode;
} /* end of: LdrfValidateNVT */

LDRF_FN LdrfFindEmptyNVT(PLdrfFileInfo pInfo, PUShort pIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    TUShort index = 0;

    if (!errCode) {
        if (!(pInfo->u.type.numNVTs)) {
            index = 1;
        } else if (!(pInfo->u.type.pDirNVTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            errCode = LDRF_ERR_NOT_FOUND;

            for ( index = 1; index <= pInfo->u.type.numNVTs; ++index ) {
				// 30-Mar-2012  Bernd  Look for an empty record, but never 
				// accept SNVT-ID 255:
                if (!(pInfo->u.type.pDirNVTIndex[index])
				&& (pInfo->scope != 0 || index != EXCLUDE_NVT_SCOPE0)) {
                    errCode = LDRF_ERR_NONE;
                    break;
                }
            }

            if (errCode) {
                if (pInfo->u.type.numNVTs < 65535UL) {
                    index = pInfo->u.type.numNVTs + 1;
                    errCode = LDRF_ERR_NONE;
                }
            }
        }
    }

	// 30-Mar-2012, Bernd
	// UNVT-ID are limited to MAX_NVT_CAPACITY (was: 65000, now is 4096), 
	// SNVT-ID are limited to MAX_NVT_CAPACITY_SCOPE0 (was: 250, now is 32768 except index 255).
	// The limitation to 4096 UNVT allows encoding the scope identifier within a 16 bit field
	// in the XNVT SI data structure. 255 (EXCLUDE_NVT_SCOPE0) is excluded from the SNVT-ID range
	// because it is used as an escape value, indicating "not a standard NV type."
	if (!errCode) {
		if (pInfo->scope) {		// user scope:
			if (index >= MAX_NVT_CAPACITY) {
				errCode = LDRF_ERR_FULL;
			}
		} else {				// standard scope:
			if (!pInfo->extendedSNVTIDaware) {
				if (index > LEGACY_NVT_CAPACITY_SCOPE0) {
					errCode = LDRF_ERR_FULL;
				}
			} else {
				if (index >= MAX_NVT_CAPACITY_SCOPE0) {
					errCode = LDRF_ERR_FULL;
				} else if (index == EXCLUDE_NVT_SCOPE0) {
					index = EXCLUDE_NVT_SCOPE0 + 1;
				}
			}
		}
	}

    if (pIndex) {
        * pIndex = index;
    }

    return errCode;
} /* end of: LdrfFindEmptyNVT */

LDRF_FN LdrfGetNVTByName(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PLdrfTypeTree* ppTypeTree) {
    return LdrfGetNVTByNameEx(pInfo, name, pIndex, ppTypeTree, 0);
}

LDRF_FN LdrfGetNVTByNameEx(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PLdrfTypeTree* ppTypeTree, PUShort pFlags) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    TULong entry;
    TUShort len = strlen(name);
    LPCSTR prefix;

    if (!errCode) {
        if ((!ppTypeTree) || (len <= PREFIX_LEN)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            LPCSTR nameSuffix = name + PREFIX_LEN;

            if ((name[0] != 'U') && (pInfo->userFlag)) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else if ((name[0] != 'S') && (!(pInfo->userFlag))) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else {
                errCode =
                    LdrfFindInSortedDir(pInfo, pInfo->u.type.pDirNVTName, pInfo->u.type.numNVTs - pInfo->u.type.numEmptyNVTs, nameSuffix, &entry,
                        LdrfNVTCompare);
            }

            if (!errCode) {
                if (pInfo->userFlag) {
                    prefix = "UNVT";
                } else {
                    prefix = "SNVT";
                }
                errCode = LdrfNVTRead(pInfo, prefix, pInfo->u.type.pDirNVTName[entry], pIndex, ppTypeTree, pFlags);
            }
        }
    }

	if (!errCode 
	&& pIndex
	&& *pIndex > LEGACY_NVT_CAPACITY_SCOPE0 
	&& pInfo->scope == 0
	&& !pInfo->extendedSNVTIDaware) {
		errCode = LDRF_ERR_NOT_FOUND;
	}

    return errCode;
} /* end of: LdrfGetNVTByName */

LDRF_FN LdrfGetNVTObsolete(PLdrfFileInfo pInfo, TUShort index, PBoolByte pObsolete) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if ((index == 0) || (index > pInfo->u.type.numNVTs) || (!pObsolete)) {
            errCode = LDRF_ERR_PARAM;
        } else if (pInfo->majorFmtVer < 3) {
            * pObsolete = F;
        } else if (!(pInfo->u.type.pDirNVTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.type.pDirNVTIndex[index])) {
            if (pInfo->emptyRecordAware) {
                errCode = LDRF_ERR_EMPTY_RECORD;
            } else {
                * pObsolete = T;
            }
        } else {
            /* Skip index and rec-size */
            errCode = LdrfFileSeek(pInfo, pInfo->u.type.pDirNVTIndex[index] + OFFSET_TYPE_FILE_REC_SIZE + sizeof(TUShort)); /* sizeof rec-size */

            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, pObsolete, sizeof(*pObsolete));
            }
        }
    }

	if (!errCode 
	&& index > LEGACY_NVT_CAPACITY_SCOPE0 
	&& pInfo->scope == 0
	&& !pInfo->extendedSNVTIDaware) {
		errCode = LDRF_ERR_NOT_FOUND;
	}

    return errCode;
} /* end of: LdrfGetNVTObsolete */

TLdrfErrCodes LdrfNVTRead(PLdrfFileInfo pInfo, LPCSTR prefix, TULong startAddress, PUShort pIndex, PLdrfTypeTree* ppTypeTree, PUShort pFlags) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_NONE;
    PLdrfTypeTree pTypeTree = NULL;
    TBool topLevelFlag, arrayFlag;
    TBool unsignedQuadFakingFlag = FALSE, doubleFakingFlag = FALSE;
    TBool signedInt64FakingFlag = FALSE, unsignedInt64FakingFlag = FALSE;
    TBool structFaking = FALSE, inAggregate = FALSE;
    TBool containsUnsignedQuadType = FALSE;
    TBool containsDoubleFloatType = FALSE;
    TBool containsSignedInt64Type = FALSE;
    TBool containsUnsignedInt64Type = FALSE;
    LPSTR pStructFakingName = NULL;
    TUShort flags = 0;

    if ((!ppTypeTree) || (!startAddress)) {
        errCode = LDRF_ERR_PARAM;
    }

    if ((!errCode) && (pIndex)) {
        errCode = LdrfFileSeek(pInfo, startAddress);

        if (!errCode) {
            errCode = LdrfFileReadShort(pInfo, pIndex);
        }
    }

    if (!errCode) {
        /* Skip rec-size and maybe obsolete mark. */
        unsigned int amountToSkip = sizeof(TUShort); /* skip rec-size */

        if (pInfo->majorFmtVer > 2) {
            /* Major Fmt Ver 3 or higher permits the NVT to be
             * marked as obsolete. In those format-versions, the NVT
             * must be skipped (it comes after the rec-size, but before
             * the type-size. */
            amountToSkip += sizeof(TBoolByte);
        }
        errCode = LdrfFileSeek(pInfo, startAddress + OFFSET_TYPE_FILE_REC_SIZE + amountToSkip);

        if (!errCode) {
            if (pInfo->majorFmtVer >= 6) {
                /* Version 6 and later have a flags byte after the obsolete mark */
                errCode = LdrfFileReadShort(pInfo, &flags);
                amountToSkip += sizeof(flags); // Add in flags for accumulated "amount to skip"
            }
        }

        if (!errCode) {
            // Now skip over type size.
            amountToSkip += sizeof(TUShort); // skip type-size;
            errCode = LdrfFileSeek(pInfo, startAddress + OFFSET_TYPE_FILE_REC_SIZE + amountToSkip);
        }
    }

    topLevelFlag = T;
    arrayFlag = F;

    while (!errCode) {
        TNVTType nodeType = NVT_TYPE_UNKNOWN;
        TUByte resNmScope, resCmtScope, resUntScope;
        TULong resNmIndex, resCmtIndex, resUntIndex;
        LPSTR pName = NULL;

        if ((arrayFlag) || (structFaking)) {
            if (structFaking || pStructFakingName != NULL) {
                pName = NEW_STRING(strlen(pStructFakingName));

                if (pName) {
                    strcpy(pName, pStructFakingName);
                }
            } else {
                pName = "";
            }

            if (doubleFakingFlag || signedInt64FakingFlag || unsignedInt64FakingFlag) {
                resNmIndex = resCmtIndex = resUntIndex = 0;
                resNmScope = resCmtScope = resUntScope = 0;
            }
        } else {
            errCode = LdrfFileReadLString(pInfo, &pName);

            if ((!errCode) && (topLevelFlag)) {
                LPSTR pTempName = pName;
                TUShort len = strlen(pName);
                pName = NEW_STRING(len + PREFIX_LEN);

                if (!pName) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    strncpy(pName, prefix, PREFIX_LEN);
                    strcpy(pName + PREFIX_LEN, pTempName);
                }
                FREE(pTempName);
                topLevelFlag = F;
            }
        } /* end of: if (arrayFlag) */

        if (structFaking) {
            nodeType = NVT_TYPE_ARRAY;
            structFaking = FALSE;
        } else if (doubleFakingFlag || signedInt64FakingFlag || unsignedInt64FakingFlag) {
            nodeType = NVT_TYPE_UNSIGNED_SHORT;
        } else {
            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, &resNmScope, sizeof(resNmScope));
            }

            if (!errCode) {
                errCode = LdrfFileRead24(pInfo, &resNmIndex);
            }

            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, &resCmtScope, sizeof(resCmtScope));
            }

            if (!errCode) {
                errCode = LdrfFileRead24(pInfo, &resCmtIndex);
            }

            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, &resUntScope, sizeof(resUntScope));
            }

            if (!errCode) {
                errCode = LdrfFileRead24(pInfo, &resUntIndex);
            }

            if (!errCode) {
                TUByte bNodeType;
                errCode = LdrfFileReadBytes(pInfo, &bNodeType, sizeof(bNodeType));
                nodeType = bNodeType;
            }

            if (!errCode) {
                if (nodeType == NVT_TYPE_UNSIGNED_QUAD) {
                    if (!(pInfo->u.type.dataTypeUnsignedQuadAware)) {
                        nodeType = NVT_TYPE_SIGNED_QUAD;
                        unsignedQuadFakingFlag = TRUE;
                    } else {
                        containsUnsignedQuadType = TRUE;
                    }
                } else if (nodeType == NVT_TYPE_DOUBLE_FLOAT) {
                    if (!(pInfo->u.type.dataTypeDoubleFloatAware)) {
                        doubleFakingFlag = TRUE;

                        if (!inAggregate || arrayFlag) { // 5-Dec-2003 BG /re EPR 31184
                            structFaking = TRUE;
                            pStructFakingName = "data";
                            nodeType = NVT_TYPE_STRUCT;
                        } else {
                            nodeType = NVT_TYPE_ARRAY;
                        }
                    } else {
                        containsDoubleFloatType = TRUE;
                    }
                }
                else if (nodeType == NVT_TYPE_SIGNED_INT64) {
                    if (!(pInfo->u.type.dataTypeSignedInt64Aware)) {
                        signedInt64FakingFlag = TRUE;
                        if (!inAggregate || arrayFlag) {
                            structFaking = TRUE;
                            pStructFakingName = "drf_int64";
                            nodeType = NVT_TYPE_STRUCT;
                        } else {
                            nodeType = NVT_TYPE_ARRAY;
                        }
                    } else {
                        containsSignedInt64Type = TRUE;
                    }
                }
                else if (nodeType == NVT_TYPE_UNSIGNED_INT64) {
                    if (!(pInfo->u.type.dataTypeUnsignedInt64Aware)) {
                        unsignedInt64FakingFlag = TRUE;
                        if (!inAggregate || arrayFlag) {
                            structFaking = TRUE;
                            pStructFakingName = "drf_uint64";
                            nodeType = NVT_TYPE_STRUCT;
                        } else {
                            nodeType = NVT_TYPE_ARRAY;
                        }
                    } else {
                        containsUnsignedInt64Type = TRUE;
                    }
                }
            }
        }

        if (!errCode) {
            errCode = LdrfNewTypeTreeNode(nodeType, pName, resNmScope, resNmIndex, resCmtScope, resCmtIndex, resUntScope, resUntIndex, &pTypeTree);
        }

        if (arrayFlag) {
            arrayFlag = F;
        } else {
            FREE(pName);
        }

        if (errCode == LDRF_ERR_INCOMPLETE) {
            TUByte rangeScaleControl;
            TLong minValid, maxValid;
            float fMinValid, fMaxValid;
            double dMinValid, dMaxValid;
            __DRF_INT64 i64MinValid, i64MaxValid;

            TShort scaleA = 1, scaleB = 0, scaleC = 0, num;
            TUByte bitfSize = 0, bitfOffset = 0, tScope, typeSzByte, enumScope;
            TUShort tIndex, enumIndex;
            TBool bitfSigned = F;

            if (structFaking) {
                errCode = LdrfSetStructUnionDetails(pTypeTree, 1);
            } else if ((nodeType == NVT_TYPE_ARRAY) && 
                       ((doubleFakingFlag) || (unsignedInt64FakingFlag) || (signedInt64FakingFlag))) {
                errCode = LdrfSetArrayDetails(pTypeTree, 8);
                arrayFlag = T;

            /* Read the specific stuff for ARRAY, STRUCT, UNION, or REFERENCE,
             * and then set the details for the node. */
            } else if ((nodeType >= NVT_TYPE_ARRAY) && (nodeType <= NVT_TYPE_UNION)) {
                errCode = LdrfFileReadShort(pInfo, &num);

                if (nodeType == NVT_TYPE_ARRAY) {
                    errCode = LdrfSetArrayDetails(pTypeTree, num);
                    arrayFlag = T;
                } else {
                    errCode = LdrfSetStructUnionDetails(pTypeTree, num);
                    inAggregate = T;
                }
            } else if (nodeType == NVT_TYPE_REFERENCE) {
                errCode = LdrfFileReadBytes(pInfo, &tScope, sizeof(tScope));

                if (!errCode) {
                    errCode = LdrfFileReadShort(pInfo, &tIndex);
                }

                if (!errCode) {
                    errCode = LdrfFileReadBytes(pInfo, &typeSzByte, sizeof(typeSzByte));
                }

                if (!errCode) {
                    errCode = LdrfSetReferenceDetails(pTypeTree, tScope, tIndex, (TUShort)typeSzByte);
                }
            } else if ((nodeType < NVT_TYPE_SIGNED_CHAR) || (nodeType > NVT_TYPE_UNSIGNED_INT64)) {
                errCode = LDRF_ERR_INTERNAL;
            } else {
                /* Read all the details for the various types of scalar nodes, and
                 * set the details for them. */
                errCode = LDRF_ERR_NONE;

                if (nodeType == NVT_TYPE_ENUM) {
                    errCode = LdrfFileReadBytes(pInfo, &enumScope, sizeof(enumScope));

                    if (!errCode) {
                        errCode = LdrfFileReadShort(pInfo, &enumIndex);
                    }
                } else if (nodeType == NVT_TYPE_BITF) {
                    TUByte bitfStuff;
                    errCode = LdrfFileReadBytes(pInfo, &bitfStuff, sizeof(bitfStuff));

                    if (!errCode) {
                        bitfSigned = (bitfStuff &BITF_SIGNED_MASK) ? T : F;
                        bitfOffset = ((bitfStuff &BITF_OFFSET_MASK) >> BITF_OFFSET_SHIFT);
                        bitfSize = bitfStuff & BITF_SIZE_MASK;
                    }
                }

                /* Read the range/scale control byte */
                if (!errCode) {
                    errCode = LdrfFileReadBytes(pInfo, &rangeScaleControl, sizeof(rangeScaleControl));
                }

                /* Process the range information */
                if (!errCode) {
                    if (unsignedQuadFakingFlag) {
                        nodeType = NVT_TYPE_UNSIGNED_QUAD;
                    } else if (doubleFakingFlag) {
                        nodeType = NVT_TYPE_DOUBLE_FLOAT;
                    } else if (signedInt64FakingFlag) {
                        nodeType = NVT_TYPE_SIGNED_INT64;
                    } else if (unsignedInt64FakingFlag) {
                        nodeType = NVT_TYPE_UNSIGNED_INT64;
                    }

                    switch (nodeType) {
                        case NVT_TYPE_BITF:
                            if (bitfSigned) {
                                minValid = 0xFFFFFFFF << (bitfSize - 1);
                                maxValid = 0;

                                if (bitfSize > 1) {
                                    maxValid = 1 << (bitfSize - 1);
                                }
                                goto SIGNED_BYTE_VALIDATION;
                            } else {
                                minValid = 0;
                                maxValid = (1 << bitfSize) - 1;
                                goto UNSIGNED_BYTE_VALIDATION;
                            }

                        case NVT_TYPE_ENUM:
                        case NVT_TYPE_SIGNED_CHAR:
                        case NVT_TYPE_SIGNED_SHORT:
                            minValid = -128;
                            maxValid = 127;
                            SIGNED_BYTE_VALIDATION:
                            if (!(rangeScaleControl &(LM_BIT + LN_BIT + LZ_BIT))) {
                                TByte b;
                                errCode = LdrfFileReadBytes(pInfo, &b, sizeof(b));
                                minValid = b;
                            }

                            if ((!errCode) && (!(rangeScaleControl &(UM_BIT + UZ_BIT)))) {
                                TByte b;
                                errCode = LdrfFileReadBytes(pInfo, &b, sizeof(b));
                                maxValid = b;
                            }
                            goto VALIDATION_COMMON;

                        case NVT_TYPE_UNSIGNED_CHAR:
                        case NVT_TYPE_UNSIGNED_SHORT:
                            minValid = 0;
                            maxValid = 255;
                            UNSIGNED_BYTE_VALIDATION:
                            if (!(rangeScaleControl &(LM_BIT + LN_BIT + LZ_BIT))) {
                                TUByte b;
                                errCode = LdrfFileReadBytes(pInfo, &b, sizeof(b));
                                minValid = b;
                            }

                            if ((!errCode) && (!(rangeScaleControl &(UM_BIT + UZ_BIT)))) {
                                TUByte b;
                                errCode = LdrfFileReadBytes(pInfo, &b, sizeof(b));
                                maxValid = b;
                            }
                            goto VALIDATION_COMMON;

                        case NVT_TYPE_SIGNED_LONG:
                            minValid = -32768;
                            maxValid = 32767;

                            if (!(rangeScaleControl &(LM_BIT + LN_BIT + LZ_BIT))) {
                                TShort s;
                                errCode = LdrfFileReadShort(pInfo, &s);
                                minValid = s;
                            }

                            if ((!errCode) && (!(rangeScaleControl &(UM_BIT + UZ_BIT)))) {
                                TShort s;
                                errCode = LdrfFileReadShort(pInfo, &s);
                                maxValid = s;
                            }
                            goto VALIDATION_COMMON;

                        case NVT_TYPE_UNSIGNED_LONG:
                            minValid = 0;
                            maxValid = 65535;

                            if (!(rangeScaleControl &(LM_BIT + LN_BIT + LZ_BIT))) {
                                TUShort s;
                                errCode = LdrfFileReadShort(pInfo, &s);
                                minValid = s;
                            }

                            if ((!errCode) && (!(rangeScaleControl &(UM_BIT + UZ_BIT)))) {
                                TUShort s;
                                errCode = LdrfFileReadShort(pInfo, &s);
                                maxValid = s;
                            }
                            goto VALIDATION_COMMON;

                        case NVT_TYPE_SIGNED_QUAD:
                        case NVT_TYPE_UNSIGNED_QUAD:
                            if (nodeType == NVT_TYPE_SIGNED_QUAD) {
                                minValid = 0x80000000;
                                maxValid = 0x7FFFFFFF;
                            } else {
                                minValid = 0;
                                maxValid = 0xFFFFFFFF;
                            }

                            if (!(rangeScaleControl &(LM_BIT + LN_BIT + LZ_BIT))) {
                                errCode = LdrfFileReadLong(pInfo, &minValid);
                            }

                            if ((!errCode) && (!(rangeScaleControl &(UM_BIT + UZ_BIT)))) {
                                errCode = LdrfFileReadLong(pInfo, &maxValid);
                            }
                            goto VALIDATION_COMMON;
                            VALIDATION_COMMON:
                            if (rangeScaleControl & LN_BIT) {
                                minValid = -1;
                            } else if (rangeScaleControl & LZ_BIT) {
                                minValid = 0;
                            }

                            if (rangeScaleControl & UZ_BIT) {
                                maxValid = 0;
                            }
                            break;

                        case NVT_TYPE_FLOAT:
                            if (rangeScaleControl & LM_BIT) {
                                fMinValid = -FLT_MAX;
                            } else if (rangeScaleControl & LN_BIT) {
                                fMinValid = -1.0f;
                            } else if (rangeScaleControl & LZ_BIT) {
                                fMinValid = 0.0f;
                            } else {
                                errCode = LdrfFileReadLong(pInfo, (TULong* ) &fMinValid);
                            }

                            if (rangeScaleControl & UM_BIT) {
                                fMaxValid = FLT_MAX;
                            } else if (rangeScaleControl & UZ_BIT) {
                                fMaxValid = 0.0f;
                            } else {
                                errCode = LdrfFileReadLong(pInfo, (TULong* ) &fMaxValid);
                            }
                            break;

                        case NVT_TYPE_DOUBLE_FLOAT:
                            if (rangeScaleControl & LM_BIT) {
                                dMinValid = -DBL_MAX;
                            } else if (rangeScaleControl & LN_BIT) {
                                dMinValid = -1.0;
                            } else if (rangeScaleControl & LZ_BIT) {
                                dMinValid = 0.0;
                            } else {
                                errCode = LdrfFileReadBigEndianBytes(pInfo, &dMinValid, sizeof(double));
                            }

                            if (rangeScaleControl & UM_BIT) {
                                dMaxValid = DBL_MAX;
                            } else if (rangeScaleControl & UZ_BIT) {
                                dMaxValid = 0.0;
                            } else {
                                errCode = LdrfFileReadBigEndianBytes(pInfo, &dMaxValid, sizeof(double));
                            }
                            break;

                        case NVT_TYPE_SIGNED_INT64:
                        case NVT_TYPE_UNSIGNED_INT64:
                            if (rangeScaleControl & LM_BIT) {
                                LdrfGetScalar64Min(nodeType == NVT_TYPE_SIGNED_INT64, &i64MinValid);
                            } else if (rangeScaleControl & LN_BIT) {
                                LdrfGetScalar64NegativeOne(&i64MinValid);
                            } else if (rangeScaleControl & LZ_BIT) {
                                LdrfGetScalar64Zero(&i64MinValid);
                            } else {
                                errCode = LdrfFileReadInt64(pInfo, &i64MinValid);
                            }

                            if (rangeScaleControl & UM_BIT) {
                                LdrfGetScalar64Max(nodeType == NVT_TYPE_SIGNED_INT64, &i64MaxValid);
                            } else if (rangeScaleControl & UZ_BIT) {
                                LdrfGetScalar64Zero(&i64MaxValid);
                            } else {
                                errCode = LdrfFileReadInt64(pInfo, &i64MaxValid);
                            }
                            break;

                        default:
                            /* Adding this clause here prevents some compilers
                             * from creating warning messages */
                            ;
                    }
                }


                /* Process the scale information */
                if ((!errCode) && (rangeScaleControl &SA_BIT)) {
                    errCode = LdrfFileReadShort(pInfo, &scaleA);
                }

                if ((!errCode) && (rangeScaleControl &SB_BIT)) {
                    errCode = LdrfFileReadShort(pInfo, &scaleB);
                }

                if ((!errCode) && (rangeScaleControl &SC_BIT)) {
                    errCode = LdrfFileReadShort(pInfo, &scaleC);
                }

                /* For certain scalars, read the invalidValue information */
                if (!errCode) {
                    if (((NVT_TYPE_SIGNED_CHAR <= nodeType) && (nodeType <= NVT_TYPE_UNSIGNED_LONG)) || (nodeType == NVT_TYPE_SIGNED_QUAD)
                        || (nodeType == NVT_TYPE_UNSIGNED_QUAD))
                        {
                        if (pInfo->majorFmtVer >= 3) {
                            TBoolByte invalidValuePresent = F;
                            errCode = LdrfFileReadBytes(pInfo, &invalidValuePresent, sizeof(invalidValuePresent));

                            if ((!errCode) && (invalidValuePresent)) {
                                TLong invalidValue;
                                errCode = LdrfFileReadLong(pInfo, &invalidValue);

                                if ((!errCode) && (!(unsignedQuadFakingFlag)) && (!(doubleFakingFlag))) {
                                    errCode = LdrfSetScalarInvalidValue(pTypeTree, invalidValue);
                                }
                            }
                        }
                    }
                }

                /* For INT 64 types, read the invalidValue information using the int 64 info */
                if (!errCode) {
                    if ((nodeType == NVT_TYPE_SIGNED_INT64) || (nodeType == NVT_TYPE_UNSIGNED_INT64))
                        {
                        if (pInfo->majorFmtVer >= 3) {
                            TBoolByte invalidValuePresent = F;
                            errCode = LdrfFileReadBytes(pInfo, &invalidValuePresent, sizeof(invalidValuePresent));

                            if ((!errCode) && (invalidValuePresent)) {
                                __DRF_INT64 invalidValue;
                                errCode = LdrfFileReadInt64(pInfo, &invalidValue);

                                if ((!errCode) && (!(unsignedInt64FakingFlag)) && (!(signedInt64FakingFlag))) {
                                    errCode = LdrfSetScalar64InvalidValue(pTypeTree, invalidValue);
                                }
                            }
                        }
                    }
                }

                /* Set the details */
                if (!errCode) {
                    if (unsignedQuadFakingFlag) {
                        /* If compatibility mode for unsigned quad, we
                         * came through the above to stay in sync in processing
                         * the file, but now will output the type into the
                         * type tree as if it is a signed quad. */
                        unsignedQuadFakingFlag = FALSE;
                        errCode = LdrfSetScalarDetails(pTypeTree, 0x80000000, 0x7FFFFFFF, 1, 0, 0);
                    } else if (doubleFakingFlag) {
                        /* If compatibility mode for double float, we
                         * came through the above to stay in sync in processing
                         * the file, but now will output the type into the
                         * type tree as if it is a unsigned short */
                        doubleFakingFlag = FALSE;
                        errCode = LdrfSetScalarDetails(pTypeTree, 0, 255, 1, 0, 0);
                    } else if (signedInt64FakingFlag) {
                        /* If compatibility mode for signed int64, we
                         * came through the above to stay in sync in processing
                         * the file, but now will output the type into the
                         * type tree as if it is a unsigned short */
                        signedInt64FakingFlag = FALSE;
                        errCode = LdrfSetScalarDetails(pTypeTree, 0, 255, 1, 0, 0);
                    } else if (unsignedInt64FakingFlag) {
                        /* If compatibility mode for un signed int64, we
                         * came through the above to stay in sync in processing
                         * the file, but now will output the type into the
                         * type tree as if it is a unsigned short */
                        unsignedInt64FakingFlag = FALSE;
                        errCode = LdrfSetScalarDetails(pTypeTree, 0, 255, 1, 0, 0);
                    } else if (nodeType == NVT_TYPE_ENUM) {
                        errCode = LdrfSetEnumDetails(pTypeTree, enumScope, enumIndex, minValid, maxValid);
                    } else if (nodeType == NVT_TYPE_BITF) {
                        errCode = LdrfSetBitfieldDetails(pTypeTree, bitfSize, bitfOffset, bitfSigned, minValid, maxValid, scaleA, scaleB, scaleC);
                    } else if (nodeType == NVT_TYPE_FLOAT) {
                        errCode = LdrfSetFloatDetails(pTypeTree, fMinValid, fMaxValid);
                    } else if (nodeType == NVT_TYPE_DOUBLE_FLOAT) {
                        errCode = LdrfSetDoubleFloatDetails(pTypeTree, dMinValid, dMaxValid);
                    } else if (nodeType == NVT_TYPE_SIGNED_INT64 || nodeType == NVT_TYPE_UNSIGNED_INT64) {
                        errCode = LdrfSetScalar64Details(pTypeTree, i64MinValid, i64MaxValid, scaleA, scaleB, scaleC);
                    } else {
                        errCode = LdrfSetScalarDetails(pTypeTree, minValid, maxValid, scaleA, scaleB, scaleC);
                    }
                }
            }

            if (errCode == LDRF_ERR_INCOMPLETE) {
                errCode = LDRF_ERR_NONE;
            } else {
                break;
            }
        } /* end of: if (errCode == LDRF_ERR_INCOMPLETE) */
    }     /* end of: while (!errCode) */


    /* Cleanup actions if there was an error */
    if (errCode) {
        LdrfFreeTypeTree(pTypeTree);
        pTypeTree = NULL;
    }


    /* Wrapup actions */
    if (!errCode) {
        if (pTypeTree) {
            if (containsUnsignedQuadType) {
                pTypeTree->containsDataTypeUnsignedQuad = TRUE;
            }

            if (containsDoubleFloatType) {
                pTypeTree->containsDataTypeDoubleFloat = TRUE;
            }

            if (containsUnsignedInt64Type) {
                pTypeTree->containsDataTypeUnsignedInt64 = TRUE;
            }

            if (containsSignedInt64Type) {
                pTypeTree->containsDataTypeSignedInt64 = TRUE;
            }
        }
        * ppTypeTree = pTypeTree;
        if (pFlags) {
            * pFlags = flags;
        }
    }

    return errCode;
} /* end of: LdrfNVTRead */

TLdrfErrCodes LdrfNVTCompare(PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int* pRC) {
    /* Local variables */
    LPSTR pEntry2 = NULL;
    TLdrfErrCodes errCode;
    unsigned int amountToSkip = sizeof(TUShort) /* skip rec-size */
    + sizeof(TUShort);                          /* skip type-size */

    if (pInfo->majorFmtVer > 2) {
        /* Major Fmt Ver 3 or higher permits the NVT to be
         * marked as obsolete. In those format-versions, the NVT
         * must be skipped (it comes after the rec-size, but before
         * the type-size. */
        amountToSkip += sizeof(TBoolByte);
        if (pInfo->majorFmtVer >= 6) {
            /* Major Fmt Ver 6 or higher has 2 bytes worth of flags following the
               type-size. */
            amountToSkip += sizeof(TUShort);
        }  
    }

    /* Skip to the name string */
    errCode = LdrfFileSeek(pInfo, entry2Addr + OFFSET_TYPE_FILE_REC_SIZE + amountToSkip);

    if (!errCode) {
        errCode = LdrfFileReadLString(pInfo, &pEntry2);
    }

    if (!errCode) {
        * pRC = strcmp(pEntry1, pEntry2);
    }
    FREE(pEntry2);

    return errCode;
} /* end of: LdrfNVTCompare */

LDRF_FN LdrfGetCPT(PLdrfFileInfo pInfo, TUShort index, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin, PUByteArray* ppMax,
    PUByteArray* ppInit, PUShort pByteArrayLen)
{
    PUByteArray* ppInvalid = NULL;
    return LdrfGetCPTEx(pInfo, index, ppTypeTree, pInheritable, ppMin, ppMax, ppInit, ppInvalid, pByteArrayLen);
}

LDRF_FN LdrfGetCPTEx(PLdrfFileInfo pInfo, TUShort index, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin, PUByteArray* ppMax,
    PUByteArray* ppInit, PUByteArray* ppInvalid, PUShort pByteArrayLen)
{
    return LdrfGetCPTEx2(pInfo, index, ppTypeTree, pInheritable, ppMin, ppMax, ppInit, 
                         ppInvalid, pByteArrayLen, NULL);
}

LDRF_FN LdrfGetCPTEx2(PLdrfFileInfo pInfo, TUShort index, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin, PUByteArray* ppMax,
    PUByteArray* ppInit, PUByteArray* ppInvalid, PUShort pByteArrayLen, PUShort pFlags)
    {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateCPT(pInfo, index);

    if ((!errCode) || (errCode == LDRF_ERR_EMPTY_RECORD)) {
        if (!ppTypeTree) {
            errCode = LDRF_ERR_PARAM;
        } else if (errCode == LDRF_ERR_EMPTY_RECORD) {
            if (!(pInfo->emptyRecordAware)) {
                PLdrfTypeTree pTypeTree = NULL;
                PUByteArray pBA = NEW_ARRAY(TUByte, 1);
                char nameBuf[20];

                if (!pBA) {
                    errCode = LDRF_ERR_SYS;
                } else {
                    strcpy(nameBuf, (pInfo->userFlag) ? "UCPT" : "SCPT");
                    strcpy(nameBuf + 4, "empty");
                    _itoa(index, nameBuf + 9, 10);
                    strcat(nameBuf, "~");
                    errCode = LdrfNewTypeTreeNode(NVT_TYPE_UNSIGNED_SHORT, nameBuf, 0, 0, 0, 0, 0, 0, &pTypeTree);

                    if (errCode == LDRF_ERR_INCOMPLETE) {
                        errCode = LdrfSetScalarDetails(pTypeTree, 0, 255, 1, 0, 0);
                    }
                }

                if (!errCode) {
                    * ppTypeTree = pTypeTree;

                    if (pInheritable) {
                        * pInheritable = F;
                    }

                    if (ppMin) {
                        * ppMin = NULL;
                    }

                    if (ppMax) {
                        * ppMax = NULL;
                    }

                    if (ppInvalid) {
                        * ppInvalid = NULL;
                    }

                    if (ppInit) {
                        * ppInit = pBA;
                        pBA[0] = 0;
                    } else {
                        errCode = LdrfFreeByteArray(pBA);
                    }

                    if (pByteArrayLen) {
                        * pByteArrayLen = 1;
                    }
                } else {
                    LdrfFreeTypeTree(pTypeTree);
                    LdrfFreeByteArray(pBA);
                }
            }
        } else {
            errCode = LdrfCPTRead(pInfo, pInfo->u.type.pDirCPTIndex[index], NULL, ppTypeTree, pInheritable, ppMin, ppMax, ppInit, ppInvalid, pByteArrayLen, pFlags);
        }
    }

    return errCode;
} /* end of: LdrfGetCPTEx2 */

LDRF_FN LdrfValidateCPT(PLdrfFileInfo pInfo, TUShort index) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if (index < 1) {
            errCode = LDRF_ERR_PARAM;
        } else if (index > pInfo->u.type.numCPTs) {
            errCode = LDRF_ERR_NOT_FOUND;
        } else if (!(pInfo->u.type.pDirCPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.type.pDirCPTIndex[index])) {
            errCode = LDRF_ERR_EMPTY_RECORD;
        }
    }

    return errCode;
} /* end of: LdrfValidateCPT */

LDRF_FN LdrfFindEmptyCPT(PLdrfFileInfo pInfo, PUShort pIndex) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    TUShort index = 0;

    if (!errCode) {
        if (!(pInfo->u.type.numCPTs)) {
            index = 1;
        } else if (!(pInfo->u.type.pDirCPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else {
            errCode = LDRF_ERR_NOT_FOUND;

            for ( index = 1; index <= pInfo->u.type.numCPTs; ++index ) {
                if (!(pInfo->u.type.pDirCPTIndex[index])) {
                    errCode = LDRF_ERR_NONE;
                    break;
                }
            }

            if (errCode) {
                if (pInfo->u.type.numCPTs < 65535UL) {
                    index = pInfo->u.type.numCPTs + 1;
                    errCode = LDRF_ERR_NONE;
                }
            }
        }
    }

    if (pIndex) {
        * pIndex = index;
    }

    return errCode;
} /* end of: LdrfFindEmptyCPT */

LDRF_FN LdrfGetCPTByName(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin,
    PUByteArray* ppMax, PUByteArray* ppInit, PUShort pByteArrayLen)
{
    PUByteArray* ppInvalid = NULL;
    return LdrfGetCPTByNameEx(pInfo, name, pIndex, ppTypeTree, pInheritable, ppMin, ppMax, ppInit, ppInvalid, pByteArrayLen);
}

LDRF_FN LdrfGetCPTByNameEx(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin,
    PUByteArray* ppMax, PUByteArray* ppInit, PUByteArray* ppInvalid, PUShort pByteArrayLen)
{
    return LdrfGetCPTByNameEx2(pInfo, name, pIndex, ppTypeTree, pInheritable, ppMin,
                               ppMax, ppInit, ppInvalid, pByteArrayLen, NULL);
}

LDRF_FN LdrfGetCPTByNameEx2(PLdrfFileInfo pInfo, LPCSTR name, PUShort pIndex, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin,
    PUByteArray* ppMax, PUByteArray* ppInit, PUByteArray* ppInvalid, PUShort pByteArrayLen,
    PUShort pFlags)
{
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);
    TULong entry;
    TUShort len = strlen(name);

    if (!errCode) {
        if ((!ppTypeTree) || (len <= PREFIX_LEN)) {
            errCode = LDRF_ERR_PARAM;
        } else {
            LPCSTR nameSuffix = name + PREFIX_LEN;

            if ((name[0] != 'U') && (pInfo->userFlag)) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else if ((name[0] != 'S') && (!(pInfo->userFlag))) {
                errCode = LDRF_ERR_NOT_FOUND;
            } else {
                errCode =
                    LdrfFindInSortedDir(pInfo, pInfo->u.type.pDirCPTName, pInfo->u.type.numCPTs - pInfo->u.type.numEmptyCPTs, nameSuffix, &entry,
                        LdrfNVTCompare);
            }

            if (!errCode) {
                errCode = LdrfCPTRead(pInfo, pInfo->u.type.pDirCPTName[entry], pIndex, ppTypeTree, pInheritable, ppMin, ppMax, ppInit, ppInvalid, pByteArrayLen, pFlags);
            }
        }
    }

    return errCode;
} /* end of: LdrfGetCPTByNameEx2 */

LDRF_FN LdrfGetCPTObsolete(PLdrfFileInfo pInfo, TUShort index, PBoolByte pObsolete) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if (((index == 0) || (index > pInfo->u.type.numCPTs)) || (!pObsolete)) {
            errCode = LDRF_ERR_PARAM;
        } else if (pInfo->majorFmtVer < 3) {
            * pObsolete = F;
        } else if (!(pInfo->u.type.pDirCPTIndex)) {
            errCode = LDRF_ERR_INTERNAL;
        } else if (!(pInfo->u.type.pDirCPTIndex[index])) {
            if (pInfo->emptyRecordAware) {
                errCode = LDRF_ERR_EMPTY_RECORD;
            } else {
                * pObsolete = T;
            }
        } else {
            /* Skip index and rec-size */
            errCode = LdrfFileSeek(pInfo, pInfo->u.type.pDirCPTIndex[index] + OFFSET_TYPE_FILE_REC_SIZE + sizeof(TUShort)); /* sizeof rec-size */

            if (!errCode) {
                errCode = LdrfFileReadBytes(pInfo, pObsolete, sizeof(*pObsolete));
            }
        }
    }

    return errCode;
} /* end of: LdrfGetCPTObsolete */

LDRF_FN LdrfFreeByteArray(PUByteArray pByteArray) {

    FREE(pByteArray);

    return LDRF_ERR_NONE;
} /* end of: LdrfFreeByteArray */

TLdrfErrCodes LdrfCPTRead(PLdrfFileInfo pInfo, TULong startAddress, PUShort pIndex, PLdrfTypeTree* ppTypeTree, PBool pInheritable, PUByteArray* ppMin,
    PUByteArray* ppMax, PUByteArray* ppInit, PUByteArray* ppInvalid, PUShort pByteArrayLen, PUShort pFlags)
    {
    /* Local variables */
    TLdrfErrCodes errCode;
    LPCSTR prefix;
    TUByte byteArrayControl;
    TUShort typeSize;

    if (((ppMin) || (ppMax) || (ppInit) || (ppInvalid)) && (!pByteArrayLen)) {
        errCode = LDRF_ERR_PARAM;
    } else {
        if (pInfo->userFlag) {
            prefix = "UCPT_";
        } else {
            prefix = "SCPT_";
        }
        errCode = LdrfNVTRead(pInfo, prefix, startAddress, pIndex, ppTypeTree, pFlags);
    }

    if ((!errCode) && ((pByteArrayLen) || (ppMin) || (ppMax) || (ppInit) || (ppInvalid) || (pInheritable))) {
        errCode = LdrfScanTypeTree(*ppTypeTree, &typeSize, NULL);

        if (!errCode) {
            if (pByteArrayLen) {
                * pByteArrayLen = typeSize;
            }
            errCode = LdrfFileReadBytes(pInfo, &byteArrayControl, sizeof(byteArrayControl));

            if (!errCode) {
                if (pInheritable) {
                    * pInheritable = (byteArrayControl &IN_BIT) ? T : F;
                }
            }
        }
    }

    if ((!errCode) && ((ppMin) || (ppMax) || (ppInit) || (ppInvalid))) {
        if (ppMin) {
            if (!(byteArrayControl &MN_BIT)) {
                * ppMin = NULL;
            } else {
                errCode = LdrfReadByteArray(pInfo, ppMin, typeSize);
            }
        } else if (byteArrayControl & MN_BIT) {
            errCode = LdrfFileWhere(pInfo);

            if (!errCode) {
                errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + typeSize);
            }
        }

        if (!errCode) {
            if (ppMax) {
                if (!(byteArrayControl &MX_BIT)) {
                    * ppMax = NULL;
                } else {
                    errCode = LdrfReadByteArray(pInfo, ppMax, typeSize);
                }
            } else if ((ppInit) && (byteArrayControl &MX_BIT)) {
                errCode = LdrfFileWhere(pInfo);

                if (!errCode) {
                    errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + typeSize);
                }
            }
        }

        if ((!errCode) && (ppInit)) {
            errCode = LdrfReadByteArray(pInfo, ppInit, typeSize);
        }

        if (!errCode) {
            if (ppInvalid) {
                if (!(byteArrayControl &IV_BIT)) {
                    * ppInvalid = NULL;
                } else {
                    errCode = LdrfReadByteArray(pInfo, ppInvalid, typeSize);
                }
            } else if ((ppInit) && (byteArrayControl &IV_BIT)) {
                errCode = LdrfFileWhere(pInfo);

                if (!errCode) {
                    errCode = LdrfFileSeek(pInfo, pInfo->currentAddress + typeSize);
                }
            }
        }
    }

    return errCode;
} /* end of: LdrfCPTRead */

TLdrfErrCodes LdrfReadByteArray(PLdrfFileInfo pInfo, PUByteArray* ppBA, TUShort size) {
    /* Local variables */
    TLdrfErrCodes errCode = LDRF_ERR_SYS;
    PUByteArray pBA = NEW_ARRAY(TUByte, size);

    if (pBA) {
        * ppBA = pBA;
        errCode = LdrfFileReadBytes(pInfo, pBA, size);
    }

    return errCode;
} /* end of: LdrfReadByteArray */

LDRF_FN LdrfResolveAllTypeTreeRefs(PLdrfFileInfo pCatalogInfo, PUByteArray pProgID, PLdrfTypeTree pTypeTree, PUShort pTypeSize) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pCatalogInfo, LDRF_CATALOG);
    TNVTType nodeType;
    TBool hasRefs;
    char fileName[_MAX_PATH + 1];
    TUShort pathLen, typeSize;

    if (!errCode) {
        if ((!pProgID) && (!pTypeTree)) {
            errCode = LDRF_ERR_PARAM;
        }
    }

    if (!errCode) {
        errCode = LdrfScanTypeTree(pTypeTree, &typeSize, &hasRefs);

        // While there is more to read and there are references in the tree
        while ((!errCode) && (hasRefs)) {
            errCode = LdrfReadTypeTreeNode(pTypeTree, &nodeType, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

            if (errCode == LDRF_ERR_INCOMPLETE) {
                errCode = LDRF_ERR_NONE;
            } else if (errCode == LDRF_ERR_NO_MORE_NODES) {
                errCode = LDRF_ERR_NONE;
                break;
            }

            // Check to see if the node type is a reference
            if (nodeType == NVT_TYPE_REFERENCE) {
                PLdrfFileInfo pTypeFileInfo = NULL;
                PLdrfTypeTree pGraftTree;
                TUByte typeScope;
                TUShort typeIndex;

                // Get the reference details
                errCode = LdrfGetReferenceDetails(pTypeTree, &typeScope, &typeIndex);

                // Get the tree to be grafted if "get detail" succeeded
                if ((errCode == LDRF_ERR_INCOMPLETE) || (errCode == LDRF_ERR_NONE)) {
                    // search the catalog based on the reference matching mode
                    pathLen = _MAX_PATH;
                    errCode = LdrfSearchCatalog(pCatalogInfo, pProgID, typeScope, LDRF_TYPE, 0, fileName, &pathLen, NULL, NULL);
                }

                if (!errCode) {
                    // Next, open the type file that we found
                    errCode = LdrfOpenFile(fileName, LDRF_TYPE, 0, T, &pTypeFileInfo);
                }

				if (!errCode
				&& pCatalogInfo->extendedSNVTIDaware
				&& pTypeFileInfo->scope == 0) {
					errCode = LdrfEnableExtendedSNVTID(pTypeFileInfo);
				}

                if (!errCode) {
                    // Then, get the type tree for the index for the reference
                    errCode = LdrfGetNVT(pTypeFileInfo, typeIndex, &pGraftTree);
                }

                if (!errCode) {
                    // Graft the tree - this consumes the tree being grafted
                    errCode = LdrfGraftReference(pTypeTree, pGraftTree);
                }

                if (pTypeFileInfo) {
                    TLdrfErrCodes ec2 = LdrfCloseFile(pTypeFileInfo);

                    if (!errCode) {
                        errCode = ec2;
                    }
                }

                if (!errCode) {
                    // See if the new tree still has a reference;
                    // this will cause the scan of the tree to restart
                    errCode = LdrfScanTypeTree(pTypeTree, &typeSize, &hasRefs);
                }
            }
        }

        if (errCode == LDRF_ERR_INCOMPLETE) {
            errCode = LDRF_ERR_NONE;
        }
    }

    if ((!errCode) && (pTypeSize)) {
        * pTypeSize = typeSize;
    }

    return errCode;
} /* end of: LdrfResolveAllTypeTreeRefs */

LDRF_FN LdrfExtendedDataTypeAware(PLdrfFileInfo pInfo, TNVTType dataType) {
    /* Local variables */
    TLdrfErrCodes errCode = LdrfValidateFileInfoPointer(pInfo, LDRF_TYPE);

    if (!errCode) {
        if (dataType == NVT_TYPE_UNSIGNED_QUAD) {
            pInfo->u.type.dataTypeUnsignedQuadAware = TRUE;
        } else if (dataType == NVT_TYPE_DOUBLE_FLOAT) {
            pInfo->u.type.dataTypeDoubleFloatAware = TRUE;
        } else if (dataType == NVT_TYPE_SIGNED_INT64) {
            pInfo->u.type.dataTypeSignedInt64Aware = TRUE;
        } else if (dataType == NVT_TYPE_UNSIGNED_INT64) {
            pInfo->u.type.dataTypeUnsignedInt64Aware = TRUE;
        } else {
            errCode = LDRF_ERR_PARAM;
        }
    }

    return errCode;
} /* end of: LdrfExtendedDataTypeAware */
