/*
 * LcaDrfI.h - Internal version of the LcaDrf.h include file
 *
 * Copyright (C) 1996 - 2012 Echelon Corporation.  All rights reserved.
 *
 * Use of this source code is subject to the terms of the Echelon Example Software
 * License Agreement which is available at www.echelon.com/license/examplesoftware/.
 *
 * Implementation by RLE
 */


#ifndef _LCADRFI
#define _LCADRFI



#define LCADRF_INTERNAL


#ifdef _LCADRFIW
    #include "lcadrfw.h"
#else
    #include "lcadrf.h"
#endif


#ifdef __cplusplus
extern "C" {
#endif


#include "cacheutl.h"


#ifndef EOS
#define EOS '\0'
#endif


#define MAX_LANG_RESOURCE_STRING 2048
#define MAX_FILE_INFO_STRING 256
#define MAX_NC_NAME 64


#define VALIDATION_CODE_INVALID		0xDEAD
#define VALIDATION_CODE_FILE_INFO	0xF11E
#define VALIDATION_CODE_TYPE_TREE	0x19EE
#define VALIDATION_CODE_STRING_SERVICE	0x5E25


#define NUM_PROG_ID_BYTES 8
#define NUM_SCOPES	  7
#define PREFIX_LEN	  4


#define BITF_SIGNED_MASK	0x80
#define BITF_OFFSET_MASK	0x70
#define BITF_OFFSET_SHIFT	4
#define BITF_SIZE_MASK		0x0F


/* In the file info structure, there are file-specific fields that contain the
 * number of items in a directory, and the (current) capacity of the directory.
 * The directory's maximum capacity is given below. */

/* For language resource files */

#define MAX_STRINGS_CAPACITY  1000000

/* For type files */

#define MAX_NVT_CAPACITY_SCOPE0   32768	// excluding 255, EXCLUDE_NVT_SCOPE0
#define MAX_NVT_CAPACITY	4096
#define MAX_CPT_CAPACITY	65000
#define MAX_ENUM_SET_CAPACITY	65000

#define	EXCLUDE_NVT_SCOPE0	255
#define LEGACY_NVT_CAPACITY_SCOPE0	250	

/* For FPT files */

#define MAX_FPT_CAPACITY	65000

/* The DRF TYP file contains a special encoding for range and scale values.
   The encoding is contained in a single byte, and is explained in the diagram below.

   +--+--+--+--+--+--+--+--+
   |LM|LN|LZ|UM|UZ|SA|SB|SC|
   +--+--+--+--+--+--+--+--+

    where the decoding is as follows:
	If the LM bit is set, the lower validation (minValid) is the minimum possible
				according to the type of the field.
	    Else, if the LN bit is set, the lower validation is -1,
	    Else, if the LZ bit is set, the lower validation is 0,
	    Else, the lower validation data follows the encoding byte as a specific value.
	If the UM bit is set, the upper validation (maxValid) is the maximum possible
				according to the type of the field.
	    Else, if the UZ bit is set, the upper validation is 0,
	    Else, the upper validation data follows the above data as a specific value.
	If the SA bit is NOT set, the Scale A value defaults to 1 (no scaling),
	    Else the Scale A data follows the above data as a specific value.
	If the SB bit is NOT set, the Scale B value defaults to 0 (no scaling).
	    Else the Scale B data follows the above data as a specific value.
	If the SC bit is NOT set, the Scale C value defaults to 0 (no scaling).
	    Else the Scale C data follows the above data as a specific value.
*/

#define  LM_BIT  0x80
#define  LN_BIT  0x40
#define  LZ_BIT  0x20
#define  UM_BIT  0x10
#define  UZ_BIT  0x08
#define  SA_BIT  0x04
#define  SB_BIT  0x02
#define  SC_BIT  0x01


/* Likewise, the DRF TYP file contains a special encoding for CPT byte arrays following
   the CPT basic type.	There are up to four byte arrays, the MIN, MAX and INVALID validation
   override arrays (each separately optional) and the INIT array (mandatory).
   The encoding is contained in a single byte, and is explained in the diagram below.

   Also, any CPT type may be inheritable (that is, it can inherit the type of an NV in
   certain circumstances).  An inheritable CPT is indicated in the CPT control byte, too.

   +--+--+--+--+--+--+--+--+
   |IN|--|--|--|IV|--|MN|MX|
   +--+--+--+--+--+--+--+--+

    where the decoding is as follows:
	If the IN bit is set, the CPT is inheritable.
    If the IV bit is set, the invalid validation override byte array is present.
	If the MN bit is set, the minimum validation override byte array is present.
	If the MX bit is set, the maximum validation override byte array is present.
*/

#define  IN_BIT  0x80
    /* ... */
#define  IV_BIT  0x08
#define  MN_BIT  0x02
#define  MX_BIT  0x01



/* The DRF FPT file contains a special encoding for range and default values.
   The encoding is contained in a single byte, and is explained in the diagram below.

   +--+--+--+--+--+--+--+--+
   |--|--|--|--|IV|DF|MN|MX|
   +--+--+--+--+--+--+--+--+

    where the decoding is as follows:
    If the IV bit is set, the invalid validation override byte array is present.
	If the DF bit is set, the default value byte array is present.
	If the MN bit is set, the minimum validation override byte array is present.
	If the MX bit is set, the maximum validation override byte array is present.

*/

#define  DF_BIT  0x04



/* Forward reference typedefs */

typedef  struct sLdrfTypeNode	TLdrfTypeNode;
typedef  struct sEnumSet	TEnumSet;
typedef  struct sHashTbl	THashTbl;
typedef  struct sHashTblEntry	THashTblEntry;

typedef  TLdrfTypeNode *	PLdrfTypeNode;
typedef  TEnumSet *		PEnumSet;
typedef  THashTbl *		PHashTbl;
typedef  THashTblEntry *	PHashTblEntry;



/* In the file info structure below, there are file-specific fields that contain the
 * number of items in a directory, and the (current) capacity of the directory.  The
 * directory's maximum capacity is given below. */

/* For catalog files */

#define MAX_DIREC_CAPACITY	  250
#define MAX_FILE_CAPACITY	25000

/* For all files */

#define CAPACITY_INCREMENT	   10



/* The file info structure */

struct sLdrfFileInfo {
    TULong	  validator;
//    FILE *	    pFile;
    PCacheControl pCache;
    TLdrfFileType fileType;
    PLdrfFileInfo pNextValid;
    LPSTR	  pDesc;
    LPSTR	  pCreator;
    LPSTR	  pURL;
    TULong	  resDescIndex, resCreIndex, currentAddress, dataAddress, dataSize, headerSize;
    TUShort	  majorDataVer, year, headerCRC;
    TUByte	  resDescScope, resCreScope, majorFmtVer, minorFmtVer, minorDataVer;
    TUByte	  month, day, hour, minute, second;
    TUByte	  scope;
    TUByte	  refID[NUM_PROG_ID_BYTES]; /* (matches progID) */
    TBool	  userFlag, editAllowed, needsHdrInfo, needsVerInfo,
			updateHeader, updateData, emptyRecordAware;
	TBool	  extendedSNVTIDaware;		// supporting larger than 250 SNVT-IDs
    union uFileVariant {
	struct sCatalogFile {
	    TUByte   stale;
	    TUShort  numDirec;
	    TUShort  numLangFiles;
	    TUShort  numTypeFiles;
	    TUShort  numFPTFiles;
	    TUShort  numFormatFiles;
	    TUShort  numEmptyDirec;	    // This field is not used yet
	    TUShort  numEmptyLangFiles;     // This field is not used yet
	    TUShort  numEmptyTypeFiles;     // This field is not used yet
	    TUShort  numEmptyFPTFiles;	    // This field is not used yet
	    TUShort  numEmptyFormatFiles;   // This field is not used yet
	    TUShort  capacityDirec;
	    TUShort  capacityLangFiles;
	    TUShort  capacityTypeFiles;
	    TUShort  capacityFPTFiles;
	    TUShort  capacityFormatFiles;
	    TULong  *pDirDirec;
	    TULong  *pDirLangFiles;
	    TULong  *pDirTypeFiles;
	    TULong  *pDirFPTFiles;
	    TULong  *pDirFormatFiles;
	} cat;
	struct sLangFile {
	    TBool    needsLangFileInfo;
	    TULong   locale;
	    TULong   numStrings;
	    TULong   numEmptyStrings;
	    TULong   capacityStrings;
	    TULong  *pDirStrings;
	    PHashTbl pHashTbl;
	} lang;
	struct sTypFile {
	    TBool    needsTypeFileInfo;
	    TBool    dataTypeUnsignedQuadAware;
	    TBool    dataTypeDoubleFloatAware;
        TBool    dataTypeSignedInt64Aware;
        TBool    dataTypeUnsignedInt64Aware;
	    TUShort  resDeps[NUM_SCOPES];
	    TUShort  typDeps[NUM_SCOPES];
	    TUShort  numNVTs;
	    TUShort  numCPTs;
	    TUShort  numEnumSets;
	    TUShort  numEmptyNVTs;
	    TUShort  numEmptyCPTs;
	    TUShort  numEmptyEnumSets;
	    TUShort  capacityNVT;
	    TUShort  capacityCPT;
	    TUShort  capacityEnumSets;
	    TULong  *pDirNVTIndex;
	    TULong  *pDirNVTName;
	    TULong  *pDirCPTIndex;
	    TULong  *pDirCPTName;
	    TULong  *pDirEnumIndex;
	    TULong  *pDirEnumTag;
	    TULong  *pDirEnumFile;
	    PEnumSet pEnumSet;
	} type;
	struct sFPTFile {
	    TBool    needsFPTFileInfo;
	    TBool    fptSetInProgress;
	    TUShort  fptIndexBeingSet;
	    TUShort  nextNVIndexToSet;
	    TUShort  nextCPIndexToSet;
	    TUShort  resDeps[NUM_SCOPES];
	    TUShort  typDeps[NUM_SCOPES];
	    TUShort  numFPTs;
	    TUShort  numEmptyFPTs;
	    TUShort  capacityFPT;
	    TULong  *pDirFPTIndex;
	    TULong  *pDirFPTName;
	    TULong  *pDirFPTKey;
	} fpt;
	struct sFormatFile {
	    int dummy;
	} format;
    } u;
};


/* The DRF file header is defined as a fixed part followed by a variable part.	The variable
 * part varies based on the type of DRF file, and the number of things it contains.  The CRC
 * checking must know the size of the header (the header always starts at 0).  Furthermore,
 * the two-byte CRC sits in the midst of the fixed part of the header.	It can't be included
 * in the CRC checking, since it is not known at the time of the CRC computation and
 * therefore not a part of that CRC computation.  The constants below give the number of
 * bytes of the fixed header portion which precede the CRC. */

#define SIZE_HDR_PREFIX     (sizeof(TULong))
#define SIZE_DESC_STRING    (MAX_FILE_INFO_STRING - SIZE_HDR_PREFIX)
#define SIZE_CRE_STRING     MAX_FILE_INFO_STRING
#define SIZE_URL_STRING     MAX_FILE_INFO_STRING
#define SIZE_RES_CODES	    (2*(sizeof(TUByte)+3))
#define SIZE_VER_AND_TIME   ((2*sizeof(TUShort))+(8*sizeof(TUByte)))

#define SIZE_HDR_SIZE_S     (sizeof(TUShort))
#define SIZE_HDR_SIZE_L     (sizeof(TULong))

#define SIZE_FIRST_PART_HDR_S  (SIZE_HDR_PREFIX + SIZE_DESC_STRING + SIZE_CRE_STRING \
				    + SIZE_URL_STRING + SIZE_RES_CODES + SIZE_VER_AND_TIME \
				    + SIZE_HDR_SIZE_S)

#define SIZE_FIRST_PART_HDR_L  (SIZE_HDR_PREFIX + SIZE_DESC_STRING + SIZE_CRE_STRING \
				    + SIZE_URL_STRING + SIZE_RES_CODES + SIZE_VER_AND_TIME \
				    + SIZE_HDR_SIZE_L)

#define SIZE_HDR_CRC	    (sizeof(TUShort))
#define SIZE_DATA_ADDR_SZ   (2*sizeof(TULong))

#define SIZE_FIXED_HDR_S    (SIZE_FIRST_PART_HDR_S + SIZE_HDR_CRC + SIZE_DATA_ADDR_SZ)
#define SIZE_FIXED_HDR_L    (SIZE_FIRST_PART_HDR_L + SIZE_HDR_CRC + SIZE_DATA_ADDR_SZ)

/* Additional stuff specific to the header for catalog files */

#define SIZE_CAT_STALE	    (sizeof(TUByte))
#define SIZE_CAT_COUNTS     (5*sizeof(TUShort))

#define SIZE_CATALOG_FILE_HDR (SIZE_CAT_STALE+SIZE_CAT_COUNTS)

/* Additional stuff specific to the header for language resource files */

#define SIZE_LANG_HDR_SCOPE ((NUM_PROG_ID_BYTES*sizeof(TUByte))+sizeof(TUByte))
#define SIZE_LANG_LOCALE    (sizeof(TULong))
#define SIZE_LANG_COUNTS    (3)

#define SIZE_LANG_FILE_HDR  (SIZE_LANG_HDR_SCOPE + SIZE_LANG_LOCALE + SIZE_LANG_COUNTS)
// V3 includes "empty counts" in addition to regular counts
#define SIZE_LANG_FILE_HDR_V3 (SIZE_LANG_FILE_HDR + SIZE_LANG_COUNTS)


/* Additional stuff specific to the header for type files */

#define SIZE_TYPE_HDR_SCOPE ((NUM_PROG_ID_BYTES*sizeof(TUByte))+sizeof(TUByte))
#define SIZE_TYPE_DEPS	    (2*(NUM_SCOPES*sizeof(TUShort)))
#define SIZE_TYPE_COUNTS    (3*sizeof(TUShort))

#define SIZE_TYPE_FILE_HDR  (SIZE_TYPE_HDR_SCOPE + SIZE_TYPE_DEPS + SIZE_TYPE_COUNTS)
// V4 includes "empty counts" in addition to regular counts
#define SIZE_TYPE_FILE_HDR_V4 (SIZE_TYPE_FILE_HDR + SIZE_TYPE_COUNTS)

/* Additional stuff specific to the header for FPT files */

#define SIZE_FPT_HDR_SCOPE  ((NUM_PROG_ID_BYTES*sizeof(TUByte))+sizeof(TUByte))
#define SIZE_FPT_DEPS	    (2*(NUM_SCOPES*sizeof(TUShort)))
#define SIZE_FPT_COUNTS     (sizeof(TUShort))

#define SIZE_FPT_FILE_HDR   (SIZE_FPT_HDR_SCOPE + SIZE_FPT_DEPS + SIZE_FPT_COUNTS)
// V4 includes "empty counts" in addition to regular counts
#define SIZE_FPT_FILE_HDR_V4 (SIZE_FPT_FILE_HDR + SIZE_FPT_COUNTS)

/* The data portion of the file begins with a CRC value, prior to any record */

#define SIZE_DATA_CRC	    (sizeof(TUShort))

/* Data records in the catalog file have a 16-bit index followed by a 16-bit size */

#define OFFSET_CATALOG_FILE_REC_SIZE 2

/* Data records in the language resource file have a 24-bit index followed by a 16-bit size */

#define OFFSET_LANG_FILE_REC_SIZE 3

/* Data records in the type file have a 16-bit index followed by a 16-bit size */

#define OFFSET_TYPE_FILE_REC_SIZE 2

/* Data records in the FPT file have a 16-bit index followed by a 16-bit size */

#define OFFSET_FPT_FILE_REC_SIZE 2



/* The enum set structure, which hangs off the file info structure for type files */

#define MAX_ENUMS_IN_SET 255

struct sEnumSet {
    LPSTR   pEnumTag;
    LPSTR   pEnumFile;
    TULong  address;
    TUShort index;
    TUShort tagDirEntry;
    TUShort fileDirEntry;
    TUShort numEnums;
    TBool   needsToBeWritten;
    struct {
	TByte  eValue;
	LPSTR  eString;
	TUByte eResScope;
	TULong eResIndex;
    } e [MAX_ENUMS_IN_SET];
};



#ifdef LCADRF_READ_ONLY
struct sHashTbl {
    int dummy;
};
#endif


/* The type tree structure, type nodes structure, and associated data */

typedef enum eTypeTreeState {
    NVT_STATE_NONE = 0,
    NVT_STATE_NEW,
    NVT_STATE_PARTIAL,
    NVT_STATE_SET_DETAILS,
    NVT_STATE_SCAN,
    NVT_STATE_SCAN_FIND
} TTypeTreeState;


struct sLdrfTypeTree {
    TULong	    validator;
    PLdrfTypeTree   pNextValid;
    PLdrfTypeNode   pNodes;
    PLdrfTypeNode   pCurrent;
    TUShort	    arrayTypeOffsetAdder;
    TBool	    refCounter;
    TBool	    containsDataTypeUnsignedQuad;
    TBool	    containsDataTypeDoubleFloat;
    TBool	    containsDataTypeUnsignedInt64;
    TBool	    containsDataTypeSignedInt64;
    TTypeTreeState  state;
};


struct sLdrfTypeNode {
    TNVTType	    nodeType;
    TUShort	    typeSize;
    TUShort	    typeOffset;
    LPCSTR	    name;
    TUByte	    resNmScope, resCmtScope, resUntScope;
    TULong	    resNmIndex, resCmtIndex, resUntIndex;
    PLdrfTypeNode   pParent;
    PLdrfTypeNode   pFieldLink;
    union uTypeVariant {
	struct sScalar {
	    TLong   minValid, maxValid, invalidVal;
	    TShort  scaleA, scaleB, scaleC;
	    TBool   invalidValuePresent;
	} NVTScalar;
	struct sScalar64 {
	    __drf_int64   minValid, maxValid, invalidVal;
	    TShort  scaleA, scaleB, scaleC;
	    TBool   invalidValuePresent;
	} NVTScalar64;
	struct sFloat {
	    float   minValid, maxValid;
	} NVTFloat;
	struct sDouble {
	    double  minValid, maxValid;
	} NVTDouble;
	struct sBitf {
	    TUByte  bitfSize, bitfOffset;
	    TBool   bitfSigned;
	    TLong   minValid, maxValid;
	    TShort  scaleA, scaleB, scaleC;
	} NVTBitf;
	struct sEnum {
	    TUByte  enumScope;
	    TUShort enumIndex;
	    TLong   minValid, maxValid;
	} NVTEnum;
	struct sArray {
	    TUShort numElements;
	    TUShort currentCount;
	    PLdrfTypeNode pElement;
	} NVTArray;
	struct sStructUnion {
	    TUShort numFields;
	    TUShort currentCount;
	    PLdrfTypeNode pFields;
	    PLdrfTypeNode pCurrent;
	    TUByte  bitfBitsUsed;
	} NVTStructUnion;
	struct sReference {
	    TUByte  scope;
	    TUShort index;
	    PLdrfTypeNode pBackLink;
	} NVTReference;
    } u;
};




/* * * * Prototypes from drfbase.c * * * */

	/* INITIALIZATION FUNCTIONS */

extern	void
LdrfInitialize (void);

	/* MEMORY ALLOCATION FUNCTIONS */

#define NEW(typeName) \
	((typeName *)LdrfAllocFn(sizeof(typeName)))

#define NEW_STRING(chars) \
	((char *)LdrfAllocFn((chars)+1))

#define NEW_ARRAY(typeName, numElements) \
	((typeName *)LdrfAllocFn(sizeof(typeName) * (numElements)))

extern	void *
LdrfAllocFn (TULong numBytes);

#define FREE(typeName) \
	(LdrfFreeFn((void **)&typeName))

extern	void
LdrfFreeFn  (void ** ppMem);

	/* DATA TRANSFER FUNCTIONS */

extern	TLdrfErrCodes
LdrfTransferStringOut (LPSTR pStrOut, PUShort pLen, LPCSTR pStr);

	/* FILE INFO STRUCTURE MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfNewFileInfo (PLdrfFileInfo * ppInfo, TLdrfFileType fileType);

extern	TLdrfErrCodes
LdrfFindFileInfo (PLdrfFileInfo pInfo, PLdrfFileInfo * ppPrevInfo);

extern	TLdrfErrCodes
LdrfFreeFileInfo (PLdrfFileInfo * ppInfo);

extern	TLdrfErrCodes
LdrfValidateFileInfoPointer (PLdrfFileInfo pInfo, TLdrfFileType fileType);

	/* LOW-LEVEL FILE I/O FUNCTIONS */

extern	TLdrfErrCodes
LdrfFileSeek (PLdrfFileInfo pInfo, long seekSet);

extern	TLdrfErrCodes
LdrfFileWhere (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfFileReadBytes (PLdrfFileInfo pInfo, void * pBytes, TULong numBytes);

extern void 
LdrfCopyAndSwapScalar(void *pTo, const void *pFrom, TULong numBytes);

extern	TLdrfErrCodes
LdrfFileReadBigEndianBytes (PLdrfFileInfo pInfo, void * pBytes, TULong numBytes);

extern	TLdrfErrCodes
LdrfFileReadString (PLdrfFileInfo pInfo, LPSTR pS, TUShort fixedLen);

extern	TLdrfErrCodes
LdrfFileReadLString (PLdrfFileInfo pInfo, LPSTR * ppS);

extern	TLdrfErrCodes
LdrfFileReadLong (PLdrfFileInfo pInfo, TULong * pLong);

extern	TLdrfErrCodes
LdrfFileReadInt64 (PLdrfFileInfo pInfo, __DRF_PINT64 pInt64);

extern	TLdrfErrCodes
LdrfFileRead24 (PLdrfFileInfo pInfo, TULong * pLong);

extern	TLdrfErrCodes
LdrfFileReadShort (PLdrfFileInfo pInfo, TUShort * pShort);

extern	TLdrfErrCodes
LdrfFileWriteBytes (PLdrfFileInfo pInfo, const void * pBytes, TULong numBytes);

extern	TLdrfErrCodes
LdrfFileWriteBigEndianBytes (PLdrfFileInfo pInfo, const void * pBytes, TULong numBytes);

extern	TLdrfErrCodes
LdrfFileWriteString (PLdrfFileInfo pInfo, LPCSTR pS, TUShort fixedLen);

extern	TLdrfErrCodes
LdrfFileWriteLString (PLdrfFileInfo pInfo, LPCSTR pS, TUByte len);

extern	TLdrfErrCodes
LdrfFileWriteLong (PLdrfFileInfo pInfo, const TULong * pLong);

extern	TLdrfErrCodes
LdrfFileWriteInt64 (PLdrfFileInfo pInfo, const __DRF_PINT64 pInt64);

extern	TLdrfErrCodes
LdrfFileWrite24 (PLdrfFileInfo pInfo, const TULong * pLong);

extern	TLdrfErrCodes
LdrfFileWriteShort (PLdrfFileInfo pInfo, const TUShort * pShort);

	/* FILE SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfGetFileType (LPCSTR filePath, TLdrfFileType * pFileType);

extern	TLdrfErrCodes
LdrfUpdateDataSize (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfUpdateRecSize (PLdrfFileInfo pInfo, TULong startAddress, TUShort offsetRecSize);

extern	TLdrfErrCodes
LdrfMoveFileRec (PLdrfFileInfo pInfo, TULong oldAddress, TULong newAddress,
		 TUShort offsetRecSize);

	/* FILE DIRECTORY MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfReadDirectory (PLdrfFileInfo pInfo, TULong minimumSize,
		   TULong numItems, PULong * ppDir);

extern	TLdrfErrCodes
LdrfWriteDirectory (PLdrfFileInfo pInfo, TULong numItems, PULong pDir);

typedef TLdrfErrCodes TLdrfCompareFn(PLdrfFileInfo pInfo, LPCSTR pEntry1,
				     TULong entry2Addr, int * pRC);

typedef TLdrfCompareFn * PLdrfCompareFn;

extern	TLdrfErrCodes
LdrfFindInUnsortedDir (PLdrfFileInfo pInfo, TULong * pDir, TULong numEntries,
		       LPCSTR pName, PULong pEntry, PLdrfCompareFn fn);

extern	TLdrfErrCodes
LdrfFindInSortedDir (PLdrfFileInfo pInfo, TULong * pDir, TULong numEntries,
		     LPCSTR pName, PULong pEntry, PLdrfCompareFn fn);

extern	TLdrfErrCodes
LdrfSearchDirForMinEntry (PULong pDir, TULong numEntries,
			  PULong pMinEntry, PULong pMinValue);

extern	TLdrfErrCodes
LdrfIncreaseCapacity (PULong pCapacity, TULong capacityMax, PULong pHdrSizeIncrement,
		      PULong * ppDir1, PULong * ppDir2, PULong * ppDir3);

extern TLdrfErrCodes
LdrfIncreaseDirectory (PULong * ppDir, TULong oldCapacity, TULong newCapacity);

extern	TLdrfErrCodes
LdrfCatalogFindMaxDataAddress (PLdrfFileInfo pInfo,
			       PULong pDir, TUShort numEntries,
			       PULong pUpperDataAddress);

	/* TYPE TREE MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfNewTypeTree (PLdrfTypeTree * ppTypeTree);

extern	TLdrfErrCodes
LdrfFindTypeTree (PLdrfTypeTree pTypeTree, PLdrfTypeTree * ppPrevTree);

extern	TLdrfErrCodes
LdrfValidateTypeTreePointer (PLdrfTypeTree pTypeTree);

extern	void
LdrfFreeTypeTreePtr (PLdrfTypeTree * ppTypeTree);

	/* WINDOWS-SPECIFIC FUNCTIONS - portable implementation must supply
	 * definitions for certain windows system functions */

extern	TLdrfErrCodes
LdrfReadRegistryKeyString (HKEY rootKey, LPCTSTR subKey, LPCTSTR valueName,
			   TCHAR * buf, DWORD * pBufLen);



/* * * * Prototypes from drfgen.c * * * */

	/* CRC SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfComputeHeaderCRC (PLdrfFileInfo pInfo, PUShort pCRC);

extern	TLdrfErrCodes
LdrfComputeDataCRC (PLdrfFileInfo pInfo, PUShort pCRC);

extern	TLdrfErrCodes
LdrfComputeFileCRC (PLdrfFileInfo pInfo, TULong startAddr, TULong size, PUShort pCRC);

extern	TUShort
LdrfComputeCRC (PUByteArray pBlock, TULong size, TUShort startCRC);

extern	TLdrfErrCodes
LdrfUpdateHeaderCRC (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfUpdateDataCRC (PLdrfFileInfo pInfo);

	/* FILE HEADER SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfReadFileHeader (PLdrfFileInfo pInfo, TBool checkCRC);

extern	TLdrfErrCodes
LdrfInitFileHeader (PLdrfFileInfo pInfo, TUByte majFmt);

extern	TLdrfErrCodes
LdrfWriteFileHeader (PLdrfFileInfo pInfo);



/* * * * Prototypes from drfcat.c * * * */

	/* CATALOG FILE HEADER MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfReadCatalogHeader (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfInitCatalogHeader (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfWriteCatalogHeader (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfFreeCatalogFileInfo (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfExpandCatalogFileHeader (PLdrfFileInfo pInfo, TULong headerSizeIncrement);

extern	TLdrfErrCodes
LdrfRelocateCatalogFileData (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfIncreaseCapacityCatDirec (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfIncreaseCapacityCatLangFiles (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfIncreaseCapacityCatTypeFiles (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfIncreaseCapacityCatFPTFiles (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfIncreaseCapacityCatFormatFiles (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfDirecWrite (PLdrfFileInfo pInfo, TULong startAddress,
		TUShort index, LPSTR pDirecName);

extern	TLdrfErrCodes
LdrfDirecCompare (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);

extern	TLdrfErrCodes
LdrfSelectDirForFileType (PLdrfFileInfo pInfo, TLdrfFileType fileType,
			  TULong ** ppDir, PUShort pNum);

extern	TLdrfErrCodes
LdrfCatalogPurgeFiles (PLdrfFileInfo pInfo, PULong pDir, PUShort pNumEntries,
		       TBool match, TUShort matchDirecIndex);

extern	TLdrfErrCodes
LdrfCatalogAddFileInternal (PLdrfFileInfo pInfo, LPCSTR newFile, TUShort direcIndex);

extern	TLdrfErrCodes
LdrfScanFormatFile (LPCSTR newFile, PUShort pMajorDataVer, PUByte pMinorDataVer,
		    PUByte pScope, PUByteArray refID);

extern	TLdrfErrCodes
LdrfGetCatalogFileInfo (PLdrfFileInfo pInfo, TULong startAddress,
			PUShort pIndex, PUShort pRecSize,  TLdrfFileType * pFileType,
			PUByte	pScope, PUByteArray refId, PUShort pDirecIndex,
			PUShort pMajorDataVer, PUByte pMinorDataVer, PULong pLocale);

extern	TLdrfErrCodes
LdrfCatalogFileWrite (PLdrfFileInfo pInfo, TULong startAddress, TUShort index,
		      TLdrfFileType fileType, TUByte scope, PUByteArray refID,
		      TUShort direcIndex, TUShort majorDataVer, TUByte minorDataVer,
		      TULong locale, LPCSTR pFileName);

extern	TLdrfErrCodes
LdrfCatalogFileCompare (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);



/* * * * Prototypes from drflang.c * * * */

	/* INITIALIZATION FUNCTIONS */

extern	void
LdrfLangInitialize (void);


	/* LANGUAGE RESOURCE FILE HEADER MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfReadLangFileHeader (PLdrfFileInfo pInfo, TBool checkCRC);

extern	TLdrfErrCodes
LdrfFreeLangFileInfo (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfReadLangRecStart (PLdrfFileInfo pInfo, TULong startAddress,
		      PULong pIndex, PUShort pStrLen, PUShort pHashCode, PUByte pUnicode);



/* * * * Prototypes from drftype.c * * * */

	/* TYPE FILE HEADER MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfReadTypeFileHeader (PLdrfFileInfo pInfo, TBool checkCRC);

extern	TLdrfErrCodes
LdrfFreeTypeFileInfo (PLdrfFileInfo pInfo);

	/* ENUM SET MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfAllocateEnumSetStruct (PEnumSet * ppEnumSet, TUShort index, TULong startAddress);

extern	TLdrfErrCodes
LdrfDeselectEnumSet (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfReadEnumSet (PLdrfFileInfo pInfo);

extern	TLdrfErrCodes
LdrfEnumTagCompare (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);

extern	TLdrfErrCodes
LdrfEnumFileCompare (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);

	/* NVT AND CPT MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfNVTRead (PLdrfFileInfo pInfo, LPCSTR prefix, TULong startAddress,
	     PUShort pIndex, PLdrfTypeTree * ppTypeTree, PUShort pFlags);

extern	TLdrfErrCodes
LdrfNVTCompare (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);

extern	TLdrfErrCodes
LdrfCPTRead (PLdrfFileInfo pInfo, TULong startAddress,
	     PUShort pIndex, PLdrfTypeTree * ppTypeTree, PBool pInheritable,
	     PUByteArray * ppMin, PUByteArray * ppMax, 
         PUByteArray * ppInit, PUByteArray * ppInvalid, PUShort pByteArrayLen, 
         PUShort pFlags);

extern	TLdrfErrCodes
LdrfReadByteArray (PLdrfFileInfo pInfo, PUByteArray * ppBA, TUShort size);



/* * * * Prototypes from drffpt.c * * * */

	/* FPT FILE HEADER MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfReadFPTFileHeader (PLdrfFileInfo pInfo, TBool checkCRC);

extern	TLdrfErrCodes
LdrfFreeFPTFileInfo (PLdrfFileInfo pInfo);

	/* FPT MANAGEMENT AND SUPPORT FUNCTIONS */

extern	TLdrfErrCodes
LdrfFPTRead (PLdrfFileInfo pInfo, TULong startAddress,
	     PUShort pIndex, PUShort pKey, LPSTR pName, PUShort pLength,
	     PUByte pResNmScope, PULong pResNmIndex, PUByte pResCmtScope, PULong pResCmtIndex,
	     PUByte pManNVs, PUByte pOptNVs, PUByte pManCPs, PUByte pOptCPs,
	     PUByte pPrincipalNV, PUShort pFlags);

extern	TLdrfErrCodes
LdrfFPTNameCompare (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);

extern	TLdrfErrCodes
LdrfFPTKeyCompare  (PLdrfFileInfo pInfo, LPCSTR pEntry1, TULong entry2Addr, int * pRC);

extern	TLdrfErrCodes
LdrfGetFPTCommonSetup (PLdrfFileInfo pInfo, TUShort FPTIndex,
		       PUByte pManNVs, PUByte pOptNVs,
		       PUByte pManCPs, PUByte pOptCPs,
		       PUByte pPrincipalNV);

extern	TLdrfErrCodes
LdrfGetFPTNVCommonSetup (PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort NVIndex);

extern	TLdrfErrCodes
LdrfGetFPTCPCommonSetup (PLdrfFileInfo pInfo, TUShort FPTIndex, TUShort CPIndex);

extern TLdrfErrCodes 
LdrfRedirectCatalogPath(LPCSTR specifiedPath, LPSTR redirectedPath, DWORD pathSize);

/* Scalar64 utilities */
extern void 
LdrfGetScalar64Min(TBool isSigned, __DRF_PINT64 pMinValue);

extern void 
LdrfGetScalar64Max(TBool isSigned, __DRF_PINT64 pMaxValue);

extern void 
LdrfGetScalar64Zero(__DRF_PINT64 pValue);

extern void 
LdrfGetScalar64NegativeOne(__DRF_PINT64 pValue);

#ifdef __cplusplus
} /* end of: extern "C" */
#endif


#endif
